Package: fs
Title: Cross-Platform File System Operations Based on 'libuv'
Version: 2.1.0
Authors@R: c(
    person("Jim", "Hester", role = "aut"),
    person("Hadley", "Wickham", role = "aut"),
    person("Gábor", "Csárdi", role = "aut"),
    person("Jeroen", "Ooms", , "jeroenooms@gmail.com", role = "cre"),
    person("libuv project contributors", role = "cph",
           comment = "libuv library"),
    person("Joyent, Inc. and other Node contributors", role = "cph",
           comment = "libuv library"),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: A cross-platform interface to file system operations, built
    on top of the 'libuv' C library.
License: MIT + file LICENSE
URL: https://fs.r-lib.org, https://github.com/r-lib/fs
BugReports: https://github.com/r-lib/fs/issues
Depends: R (>= 4.1)
Imports: methods
Suggests: covr, crayon, knitr, pillar (>= 1.0.0), rmarkdown, spelling,
        testthat (>= 3.0.0), tibble (>= 1.1.0), vctrs (>= 0.3.0), withr
VignetteBuilder: knitr
SystemRequirements: libuv: libuv-devel (rpm) or libuv1-dev (deb).
        Alternatively to build the vendored libuv 'cmake' is required.
        GNU make.
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-04-23
Copyright: file COPYRIGHTS
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-04-18 13:56:26 UTC; jeroen
Author: Jim Hester [aut],
  Hadley Wickham [aut],
  Gábor Csárdi [aut],
  Jeroen Ooms [cre],
  libuv project contributors [cph] (libuv library),
  Joyent, Inc. and other Node contributors [cph] (libuv library),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Jeroen Ooms <jeroenooms@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-18 15:10:02 UTC
Built: R 4.4.3; x86_64-apple-darwin20; 2026-04-18 16:57:58 UTC; unix
Archs: fs.so.dSYM

Package: fs
Title: Cross-Platform File System Operations Based on 'libuv'
Version: 1.6.6
Authors@R: c(
    person("Jim", "Hester", role = "aut"),
    person("Hadley", "Wickham", , "hadley@posit.co", role = "aut"),
    person("Gábor", "Csárdi", , "csardi.gabor@gmail.com", role = c("aut", "cre")),
    person("libuv project contributors", role = "cph",
           comment = "libuv library"),
    person("Joyent, Inc. and other Node contributors", role = "cph",
           comment = "libuv library"),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: A cross-platform interface to file system operations, built
    on top of the 'libuv' C library.
License: MIT + file LICENSE
URL: https://fs.r-lib.org, https://github.com/r-lib/fs
BugReports: https://github.com/r-lib/fs/issues
Depends: R (>= 3.6)
Imports: methods
Suggests: covr, crayon, knitr, pillar (>= 1.0.0), rmarkdown, spelling,
        testthat (>= 3.0.0), tibble (>= 1.1.0), vctrs (>= 0.3.0), withr
VignetteBuilder: knitr
ByteCompile: true
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Copyright: file COPYRIGHTS
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.2.3
SystemRequirements: GNU make
NeedsCompilation: yes
Packaged: 2025-04-12 09:39:13 UTC; gaborcsardi
Author: Jim Hester [aut],
  Hadley Wickham [aut],
  Gábor Csárdi [aut, cre],
  libuv project contributors [cph] (libuv library),
  Joyent, Inc. and other Node contributors [cph] (libuv library),
  Posit Software, PBC [cph, fnd]
Maintainer: Gábor Csárdi <csardi.gabor@gmail.com>
Repository: CRAN
Date/Publication: 2025-04-12 10:40:02 UTC
Built: R 4.4.0; amd64-portbld-midnightbsd4; 2026-07-17 17:56:54 UTC; unix

Package: fs
Title: Cross-Platform File System Operations Based on 'libuv'
Version: 1.6.0
Authors@R: c(
    person("Jim", "Hester", role = "aut"),
    person("Hadley", "Wickham", , "hadley@rstudio.com", role = "aut"),
    person("Gábor", "Csárdi", , "csardi.gabor@gmail.com", role = c("aut", "cre")),
    person("libuv project contributors", role = "cph",
           comment = "libuv library"),
    person("Joyent, Inc. and other Node contributors", role = "cph",
           comment = "libuv library"),
    person("RStudio", role = c("cph", "fnd"))
  )
Description: A cross-platform interface to file system operations, built
    on top of the 'libuv' C library.
License: MIT + file LICENSE
URL: https://fs.r-lib.org, https://github.com/r-lib/fs,
        https://fs.r-lib.org/
BugReports: https://github.com/r-lib/fs/issues
Depends: R (>= 3.4)
Imports: methods
Suggests: covr, crayon, knitr, pillar (>= 1.0.0), rmarkdown, spelling,
        testthat (>= 3.0.0), tibble (>= 1.1.0), vctrs (>= 0.3.0), withr
VignetteBuilder: knitr
ByteCompile: true
Copyright: file COPYRIGHTS
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.1.2
SystemRequirements: GNU make
Config/testthat/edition: 3
Config/Needs/website: tidyverse/tidytemplate
NeedsCompilation: yes
Packaged: 2023-01-22 22:46:59 UTC; gaborcsardi
Author: Jim Hester [aut],
  Hadley Wickham [aut],
  Gábor Csárdi [aut, cre],
  libuv project contributors [cph] (libuv library),
  Joyent, Inc. and other Node contributors [cph] (libuv library),
  RStudio [cph, fnd]
Maintainer: Gábor Csárdi <csardi.gabor@gmail.com>
Repository: CRAN
Date/Publication: 2023-01-23 10:10:07 UTC

# Generated by roxygen2: do not edit by hand

S3method("!",fs_perms)
S3method("&",fs_perms)
S3method("+",fs_path)
S3method("/",fs_path)
S3method("==",fs_perms)
S3method("[",fs_bytes)
S3method("[",fs_path)
S3method("[",fs_perms)
S3method("[[",fs_bytes)
S3method("[[",fs_path)
S3method("[[",fs_perms)
S3method("|",fs_perms)
S3method(Ops,fs_bytes)
S3method(as.character,fs_bytes)
S3method(as.character,fs_perms)
S3method(as_fs_bytes,default)
S3method(as_fs_bytes,fs_bytes)
S3method(as_fs_bytes,numeric)
S3method(as_fs_path,"NULL")
S3method(as_fs_path,character)
S3method(as_fs_perms,"NULL")
S3method(as_fs_perms,character)
S3method(as_fs_perms,fs_perms)
S3method(as_fs_perms,integer)
S3method(as_fs_perms,numeric)
S3method(as_fs_perms,octmode)
S3method(format,fs_bytes)
S3method(format,fs_perms)
S3method(max,fs_bytes)
S3method(min,fs_bytes)
S3method(print,fs_bytes)
S3method(print,fs_path)
S3method(print,fs_perms)
S3method(sum,fs_bytes)
S3method(xtfrm,fs_path)
export("path_ext<-")
export(as_fs_bytes)
export(as_fs_path)
export(as_fs_perms)
export(dir_copy)
export(dir_create)
export(dir_delete)
export(dir_exists)
export(dir_info)
export(dir_ls)
export(dir_map)
export(dir_tree)
export(dir_walk)
export(file_access)
export(file_chmod)
export(file_chown)
export(file_copy)
export(file_create)
export(file_delete)
export(file_exists)
export(file_info)
export(file_move)
export(file_show)
export(file_size)
export(file_temp)
export(file_temp_pop)
export(file_temp_push)
export(file_touch)
export(fs_bytes)
export(fs_path)
export(fs_perms)
export(group_ids)
export(is_absolute_path)
export(is_dir)
export(is_dir_empty)
export(is_file)
export(is_file_empty)
export(is_link)
export(link_copy)
export(link_create)
export(link_delete)
export(link_exists)
export(link_path)
export(path)
export(path_abs)
export(path_common)
export(path_dir)
export(path_expand)
export(path_expand_r)
export(path_ext)
export(path_ext_remove)
export(path_ext_set)
export(path_file)
export(path_filter)
export(path_has_parent)
export(path_home)
export(path_home_r)
export(path_join)
export(path_norm)
export(path_package)
export(path_real)
export(path_rel)
export(path_sanitize)
export(path_select_components)
export(path_split)
export(path_temp)
export(path_tidy)
export(path_wd)
export(user_ids)
importFrom(methods,setOldClass)
importFrom(stats,na.omit)
importFrom(stats,setNames)
importFrom(utils,browseURL)
useDynLib(fs, .registration = TRUE)

# Generated by roxygen2: do not edit by hand

S3method("!",fs_perms)
S3method("&",fs_perms)
S3method("+",fs_path)
S3method("/",fs_path)
S3method("==",fs_perms)
S3method("[",fs_bytes)
S3method("[",fs_path)
S3method("[",fs_perms)
S3method("[[",fs_bytes)
S3method("[[",fs_path)
S3method("[[",fs_perms)
S3method("|",fs_perms)
S3method(Ops,fs_bytes)
S3method(as.character,fs_bytes)
S3method(as.character,fs_perms)
S3method(as_fs_bytes,default)
S3method(as_fs_bytes,fs_bytes)
S3method(as_fs_bytes,numeric)
S3method(as_fs_path,"NULL")
S3method(as_fs_path,character)
S3method(as_fs_perms,"NULL")
S3method(as_fs_perms,character)
S3method(as_fs_perms,fs_perms)
S3method(as_fs_perms,integer)
S3method(as_fs_perms,numeric)
S3method(as_fs_perms,octmode)
S3method(format,fs_bytes)
S3method(format,fs_perms)
S3method(max,fs_bytes)
S3method(min,fs_bytes)
S3method(print,fs_bytes)
S3method(print,fs_path)
S3method(print,fs_perms)
S3method(sum,fs_bytes)
S3method(xtfrm,fs_path)
export("path_ext<-")
export(as_fs_bytes)
export(as_fs_path)
export(as_fs_perms)
export(dir_copy)
export(dir_create)
export(dir_delete)
export(dir_exists)
export(dir_info)
export(dir_ls)
export(dir_map)
export(dir_tree)
export(dir_walk)
export(file_access)
export(file_chmod)
export(file_chown)
export(file_copy)
export(file_create)
export(file_delete)
export(file_exists)
export(file_info)
export(file_move)
export(file_show)
export(file_size)
export(file_temp)
export(file_temp_pop)
export(file_temp_push)
export(file_touch)
export(fs_bytes)
export(fs_path)
export(fs_perms)
export(group_ids)
export(is_absolute_path)
export(is_dir)
export(is_file)
export(is_file_empty)
export(is_link)
export(link_copy)
export(link_create)
export(link_delete)
export(link_exists)
export(link_path)
export(path)
export(path_abs)
export(path_common)
export(path_dir)
export(path_expand)
export(path_expand_r)
export(path_ext)
export(path_ext_remove)
export(path_ext_set)
export(path_file)
export(path_filter)
export(path_has_parent)
export(path_home)
export(path_home_r)
export(path_join)
export(path_norm)
export(path_package)
export(path_real)
export(path_rel)
export(path_sanitize)
export(path_split)
export(path_temp)
export(path_tidy)
export(path_wd)
export(user_ids)
importFrom(methods,setOldClass)
importFrom(stats,na.omit)
importFrom(stats,setNames)
importFrom(utils,browseURL)
useDynLib(fs, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# fs

<!-- badges: start -->

[![lifecycle](https://img.shields.io/badge/lifecycle-maturing-blue.svg)](https://lifecycle.r-lib.org/articles/stages.html#maturing)
[![R-CMD-check](https://github.com/r-lib/fs/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/fs/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/fs/branch/main/graph/badge.svg)](https://app.codecov.io/gh/r-lib/fs?branch=main)
<!-- badges: end -->

<p align="center">
<img src="https://i.imgur.com/NAux1Xc.png" width = "75%"/>
</p>

**fs** provides a cross-platform, uniform interface to file system
operations. It shares the same back-end component as
[nodejs](https://nodejs.org), the
[libuv](http://docs.libuv.org/en/v1.x/fs.html) C library, which brings
the benefit of extensive real-world use and rigorous cross-platform
testing. The name, and some of the interface, is partially inspired by
Rust’s [fs module](https://doc.rust-lang.org/std/fs/index.html).

## Installation

You can install the released version of **fs** from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("fs")
```

And the development version from [GitHub](https://github.com/) with:

``` r
# install.packages("devtools")
devtools::install_github("r-lib/fs")
```

## Comparison vs base equivalents

**fs** functions smooth over some of the idiosyncrasies of file handling
with base R functions:

- Vectorization. All **fs** functions are vectorized, accepting multiple
  paths as input. Base functions are inconsistently vectorized.

- Predictable return values that always convey a path. All **fs**
  functions return a character vector of paths, a named integer or a
  logical vector, where the names give the paths. Base return values are
  more varied: they are often logical or contain error codes which
  require downstream processing.

- Explicit failure. If **fs** operations fail, they throw an error. Base
  functions tend to generate a warning and a system dependent error
  code. This makes it easy to miss a failure.

- UTF-8 all the things. **fs** functions always convert input paths to
  UTF-8 and return results as UTF-8. This gives you path encoding
  consistency across OSes. Base functions rely on the native system
  encoding.

- Naming convention. **fs** functions use a consistent naming
  convention. Because base R’s functions were gradually added over time
  there are a number of different conventions used (e.g. `path.expand()`
  vs `normalizePath()`; `Sys.chmod()` vs `file.access()`).

### Tidy paths

**fs** functions always return ‘tidy’ paths. Tidy paths

- Always use `/` to delimit directories
- never have multiple `/` or trailing `/`

Tidy paths are also coloured (if your terminal supports it) based on the
file permissions and file type. This colouring can be customized or
extended by setting the `LS_COLORS` environment variable, in the same
output format as [GNU
dircolors](http://www.bigsoft.co.uk/blog/index.php/2008/04/11/configuring-ls_colors).

## Usage

**fs** functions are divided into four main categories:

- `path_` for manipulating and constructing paths
- `file_` for files
- `dir_` for directories
- `link_` for links

Directories and links are special types of files, so `file_` functions
will generally also work when applied to a directory or link.

``` r
library(fs)

# Construct a path to a file with `path()`
path("foo", "bar", letters[1:3], ext = "txt")
#> foo/bar/a.txt foo/bar/b.txt foo/bar/c.txt

# list files in the current directory
dir_ls()
#> DESCRIPTION      LICENSE          LICENSE.md       MAINTENANCE.md   
#> NAMESPACE        NEWS.md          R                README.Rmd       
#> README.md        _pkgdown.yml     cleanup          codecov.yml      
#> cran-comments.md fs.Rproj         inst             man              
#> man-roxygen      revdep           src              tests            
#> vignettes

# create a new directory
tmp <- dir_create(file_temp())
tmp
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpsYVhMJ/file14eb5f12e1a8

# create new files in that directory
file_create(path(tmp, "my-file.txt"))
dir_ls(tmp)
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpsYVhMJ/file14eb5f12e1a8/my-file.txt

# remove files from the directory
file_delete(path(tmp, "my-file.txt"))
dir_ls(tmp)
#> character(0)

# remove the directory
dir_delete(tmp)
```

**fs** is designed to work well with the pipe, though because it is a
minimal-dependency infrastructure package it doesn’t provide the pipe
itself. You will need to attach
[magrittr](https://magrittr.tidyverse.org) or similar.

``` r
library(magrittr)

paths <- file_temp() %>%
  dir_create() %>%
  path(letters[1:5]) %>%
  file_create()
paths
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpsYVhMJ/file14eb51c5215df/a
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpsYVhMJ/file14eb51c5215df/b
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpsYVhMJ/file14eb51c5215df/c
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpsYVhMJ/file14eb51c5215df/d
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpsYVhMJ/file14eb51c5215df/e

paths %>% file_delete()
```

**fs** functions also work well in conjunction with other
[tidyverse](https://www.tidyverse.org/) packages, like
[dplyr](https://dplyr.tidyverse.org) and
[purrr](https://purrr.tidyverse.org).

Some examples…

``` r
suppressMessages(
  library(tidyverse))
```

Filter files by type, permission and size

``` r
dir_info("src", recurse = FALSE) %>%
  filter(type == "file", permissions == "u+r", size > "10KB") %>%
  arrange(desc(size)) %>%
  select(path, permissions, size, modification_time)
#> # A tibble: 11 × 4
#>    path          permissions        size modification_time  
#>    <fs::path>    <fs::perms> <fs::bytes> <dttm>             
#>  1 src/fs.so     rwxr-xr-x        492.6K 2023-01-22 21:52:58
#>  2 src/id.o      rw-r--r--        213.5K 2023-01-22 21:52:41
#>  3 src/dir.o     rw-r--r--        102.7K 2023-01-22 21:52:40
#>  4 src/utils.o   rw-r--r--         89.2K 2023-01-22 21:52:41
#>  5 src/path.o    rw-r--r--         81.7K 2023-01-22 21:52:41
#>  6 src/link.o    rw-r--r--         74.2K 2023-01-22 21:52:41
#>  7 src/getmode.o rw-r--r--         65.3K 2023-01-22 21:52:40
#>  8 src/file.o    rw-r--r--         48.5K 2023-01-22 21:52:40
#>  9 src/error.o   rw-r--r--         18.4K 2023-01-22 21:52:40
#> 10 src/init.o    rw-r--r--         17.4K 2023-01-22 21:52:41
#> 11 src/file.cc   rw-r--r--         11.7K 2023-01-22 21:52:29
```

Tabulate and display folder size.

``` r
dir_info("src", recurse = TRUE) %>%
  group_by(directory = path_dir(path)) %>%
  tally(wt = size, sort = TRUE)
#> # A tibble: 14 × 2
#>    directory                                n
#>    <chr>                          <fs::bytes>
#>  1 src/libuv-1.44.2                     2.85M
#>  2 src/libuv-1.44.2/src/unix            1.28M
#>  3 src                                  1.22M
#>  4 src/libuv-1.44.2/test                1.05M
#>  5 src/libuv-1.44.2/src/win           742.07K
#>  6 src/libuv-1.44.2/m4                 356.7K
#>  7 src/libuv-1.44.2/src               340.12K
#>  8 src/libuv-1.44.2/include/uv        137.44K
#>  9 src/libuv-1.44.2/img               106.71K
#> 10 src/libuv-1.44.2/include            66.23K
#> 11 src/unix                             60.2K
#> 12 src/bsd                             20.02K
#> 13 src/windows                          4.73K
#> 14 src/libuv-1.44.2/test/fixtures         453
```

Read a collection of files into one data frame.

`dir_ls()` returns a named vector, so it can be used directly with
`purrr::map_df(.id)`.

``` r
# Create separate files for each species
iris %>%
  split(.$Species) %>%
  map(select, -Species) %>%
  iwalk(~ write_tsv(.x, paste0(.y, ".tsv")))

# Show the files
iris_files <- dir_ls(glob = "*.tsv")
iris_files
#> setosa.tsv     versicolor.tsv virginica.tsv

# Read the data into a single table, including the filenames
iris_files %>%
  map_df(read_tsv, .id = "file", col_types = cols(), n_max = 2)
#> # A tibble: 6 × 5
#>   file           Sepal.Length Sepal.Width Petal.Length Petal.Width
#>   <chr>                 <dbl>       <dbl>        <dbl>       <dbl>
#> 1 setosa.tsv              5.1         3.5          1.4         0.2
#> 2 setosa.tsv              4.9         3            1.4         0.2
#> 3 versicolor.tsv          7           3.2          4.7         1.4
#> 4 versicolor.tsv          6.4         3.2          4.5         1.5
#> 5 virginica.tsv           6.3         3.3          6           2.5
#> 6 virginica.tsv           5.8         2.7          5.1         1.9

file_delete(iris_files)
```

## Feedback wanted!

We hope **fs** is a useful tool for both analysis scripts and packages.
Please open [GitHub issues](https://github.com/r-lib/fs) for any feature
requests or bugs.

In particular, we have found non-ASCII filenames in non-English locales
on Windows to be especially tricky to reproduce and handle correctly.
Feedback from users who use commonly have this situation is greatly
appreciated.

## Code of Conduct

Please note that the fs project is released with a [Contributor Code of
Conduct](https://fs.r-lib.org/dev/CODE_OF_CONDUCT.html). By contributing to
this project, you agree to abide by its terms.

<!-- README.md is generated from README.Rmd. Please edit that file -->

# fs <a href="https://fs.r-lib.org/"><img src="man/figures/logo.png" align="right" height="138" alt="fs website" /></a>

<!-- badges: start -->

[![lifecycle](https://img.shields.io/badge/lifecycle-maturing-blue.svg)](https://lifecycle.r-lib.org/articles/stages.html#maturing)
[![R-CMD-check](https://github.com/r-lib/fs/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/fs/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/fs/graph/badge.svg)](https://app.codecov.io/gh/r-lib/fs)
<!-- badges: end -->

**fs** provides a cross-platform, uniform interface to file system
operations. It shares the same back-end component as
[nodejs](https://nodejs.org), the
[libuv](https://docs.libuv.org/en/v1.x/fs.html) C library, which brings
the benefit of extensive real-world use and rigorous cross-platform
testing. The name, and some of the interface, is partially inspired by
Rust’s [fs module](https://doc.rust-lang.org/std/fs/index.html).

## Installation

You can install the released version of **fs** from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("fs")
```

And the development version from [GitHub](https://github.com/) with:

``` r
pak::pak("r-lib/fs")
```

## Comparison vs base equivalents

**fs** functions smooth over some of the idiosyncrasies of file handling
with base R functions:

-   Vectorization. All **fs** functions are vectorized, accepting
    multiple paths as input. Base functions are inconsistently
    vectorized.

-   Predictable return values that always convey a path. All **fs**
    functions return a character vector of paths, a named integer or a
    logical vector, where the names give the paths. Base return values
    are more varied: they are often logical or contain error codes which
    require downstream processing.

-   Explicit failure. If **fs** operations fail, they throw an error.
    Base functions tend to generate a warning and a system dependent
    error code. This makes it easy to miss a failure.

-   UTF-8 all the things. **fs** functions always convert input paths to
    UTF-8 and return results as UTF-8. This gives you path encoding
    consistency across OSes. Base functions rely on the native system
    encoding.

-   Naming convention. **fs** functions use a consistent naming
    convention. Because base R’s functions were gradually added over
    time there are a number of different conventions used
    (e.g. `path.expand()` vs `normalizePath()`; `Sys.chmod()` vs
    `file.access()`).

### Tidy paths

**fs** functions always return ‘tidy’ paths. Tidy paths

-   Always use `/` to delimit directories
-   never have multiple `/` or trailing `/`

Tidy paths are also coloured (if your terminal supports it) based on the
file permissions and file type. This colouring can be customized or
extended by setting the `LS_COLORS` environment variable, in the same
output format as [GNU
dircolors](https://www.bigsoft.co.uk/blog/index.php/2008/04/11/configuring-ls_colors).

## Usage

**fs** functions are divided into four main categories:

-   `path_` for manipulating and constructing paths
-   `file_` for files
-   `dir_` for directories
-   `link_` for links

Directories and links are special types of files, so `file_` functions
will generally also work when applied to a directory or link.

``` r
library(fs)

# Construct a path to a file with `path()`
path("foo", "bar", letters[1:3], ext = "txt")
#> foo/bar/a.txt foo/bar/b.txt foo/bar/c.txt

# list files in the current directory
dir_ls()
#> DESCRIPTION      LICENSE          LICENSE.md       MAINTENANCE.md   
#> NAMESPACE        NEWS.md          R                README.Rmd       
#> README.md        _pkgdown.yml     air.toml         cleanup          
#> codecov.yml      cran-comments.md fs.Rproj         inst             
#> man              man-roxygen      src              tests            
#> vignettes

# create a new directory
tmp <- dir_create(file_temp())
tmp
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpBx9X5L/file1306976eefd77

# create new files in that directory
file_create(path(tmp, "my-file.txt"))
dir_ls(tmp)
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpBx9X5L/file1306976eefd77/my-file.txt

# remove files from the directory
file_delete(path(tmp, "my-file.txt"))
dir_ls(tmp)
#> character(0)

# remove the directory
dir_delete(tmp)
```

**fs** is designed to work well with the pipe, though because it is a
minimal-dependency infrastructure package it doesn’t provide the pipe
itself. You will need to attach
[magrittr](https://magrittr.tidyverse.org) or similar.

``` r
library(magrittr)

paths <- file_temp() |>
  dir_create() |>
  path(letters[1:5]) |>
  file_create()
paths
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpBx9X5L/file13069ccfb642/a
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpBx9X5L/file13069ccfb642/b
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpBx9X5L/file13069ccfb642/c
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpBx9X5L/file13069ccfb642/d
#> /var/folders/ph/fpcmzfd16rgbbk8mxvy9m2_h0000gn/T/RtmpBx9X5L/file13069ccfb642/e

paths |> file_delete()
```

**fs** functions also work well in conjunction with other
[tidyverse](https://tidyverse.org/) packages, like
[dplyr](https://dplyr.tidyverse.org) and
[purrr](https://purrr.tidyverse.org).

Some examples…

``` r
suppressMessages(
  library(tidyverse))
```

Filter files by type, permission and size

``` r
dir_info("src", recurse = FALSE) |>
  filter(type == "file", permissions == "u+r", size > "10KB") |>
  arrange(desc(size)) |>
  select(path, permissions, size, modification_time)
#> # A tibble: 12 × 4
#>    path          permissions        size modification_time  
#>    <fs::path>    <fs::perms> <fs::bytes> <dttm>             
#>  1 src/fs.so     rwxr-xr-x        445.2K 2025-04-25 13:35:54
#>  2 src/id.o      rw-r--r--        206.1K 2025-04-25 13:35:53
#>  3 src/dir.o     rw-r--r--          111K 2025-04-25 13:35:51
#>  4 src/utils.o   rw-r--r--         94.3K 2025-04-25 13:35:54
#>  5 src/path.o    rw-r--r--         87.6K 2025-04-25 13:35:53
#>  6 src/link.o    rw-r--r--           82K 2025-04-25 13:35:53
#>  7 src/getmode.o rw-r--r--         71.1K 2025-04-25 13:35:52
#>  8 src/file.o    rw-r--r--         55.3K 2025-04-25 13:35:52
#>  9 src/error.o   rw-r--r--         27.9K 2025-04-25 13:35:51
#> 10 src/init.o    rw-r--r--         20.8K 2025-04-25 13:35:53
#> 11 src/fs.o      rw-r--r--         16.3K 2025-04-25 13:35:52
#> 12 src/file.cc   rw-r--r--         11.8K 2025-04-25 12:25:11
```

Tabulate and display folder size.

``` r
dir_info("src", recurse = TRUE) |>
  group_by(directory = path_dir(path)) |>
  tally(wt = size, sort = TRUE)
#> # A tibble: 14 × 2
#>    directory                                n
#>    <chr>                          <fs::bytes>
#>  1 src/libuv-1.44.2                     2.88M
#>  2 src/libuv-1.44.2/src/unix             1.3M
#>  3 src                                  1.23M
#>  4 src/libuv-1.44.2/test                1.05M
#>  5 src/libuv-1.44.2/src/win           742.07K
#>  6 src/libuv-1.44.2/m4                 356.7K
#>  7 src/libuv-1.44.2/src               342.89K
#>  8 src/libuv-1.44.2/include/uv        137.44K
#>  9 src/libuv-1.44.2/img               106.71K
#> 10 src/unix                            68.41K
#> 11 src/libuv-1.44.2/include            66.23K
#> 12 src/bsd                             20.09K
#> 13 src/windows                          4.75K
#> 14 src/libuv-1.44.2/test/fixtures         453
```

Read a collection of files into one data frame.

`dir_ls()` returns a named vector, so it can be used directly with
`purrr::map_df(.id)`.

``` r
# Create separate files for each species
iris |>
  (\(x) split(x, x$Species))() |>
  map(select, -Species) |>
  iwalk(\(.x, .y) write_tsv(.x, paste0(.y, ".tsv")))

# Show the files
iris_files <- dir_ls(glob = "*.tsv")
iris_files
#> setosa.tsv     versicolor.tsv virginica.tsv

# Read the data into a single table, including the filenames
iris_files |>
  map_df(read_tsv, .id = "file", col_types = cols(), n_max = 2)
#> # A tibble: 6 × 5
#>   file           Sepal.Length Sepal.Width Petal.Length Petal.Width
#>   <chr>                 <dbl>       <dbl>        <dbl>       <dbl>
#> 1 setosa.tsv              5.1         3.5          1.4         0.2
#> 2 setosa.tsv              4.9         3            1.4         0.2
#> 3 versicolor.tsv          7           3.2          4.7         1.4
#> 4 versicolor.tsv          6.4         3.2          4.5         1.5
#> 5 virginica.tsv           6.3         3.3          6           2.5
#> 6 virginica.tsv           5.8         2.7          5.1         1.9

file_delete(iris_files)
```

## Feedback wanted!

We hope **fs** is a useful tool for both analysis scripts and packages.
Please open [GitHub issues](https://github.com/r-lib/fs) for any feature
requests or bugs.

In particular, we have found non-ASCII filenames in non-English locales
on Windows to be especially tricky to reproduce and handle correctly.
Feedback from users who use commonly have this situation is greatly
appreciated.

## Code of Conduct

Please note that the fs project is released with a [Contributor Code of
Conduct](https://fs.r-lib.org/dev/CODE_OF_CONDUCT.html). By contributing
to this project, you agree to abide by its terms.

describe("path_package", {
  it("errors if the package does not exist", {
    cnd <- tryCatch(path_package("arst"), error = function(e) e)
    expect_s3_class(cnd, "EEXIST")
    expect_match(conditionMessage(cnd), "Can't find package")
  })
  it("errors if the package location is not found", {
    cnd <- tryCatch(path_package("base", "foo"), error = function(e) e)
    expect_s3_class(cnd, "EEXIST")
    expect_match(conditionMessage(cnd), "do(es)? not exist")
  })
  it("works on installed files", {
    expect_error_free(p <- path_package("base", "INDEX"))
    expect_s3_class(p, "fs_path")
  })
})

#' File paths
#'
#' @description
#' Tidy file paths, character vectors which are coloured by file type on
#' capable terminals.
#'
#' Colouring can be customized by setting the `LS_COLORS` environment variable,
#' the format is the same as that read by GNU ls / dircolors.
#'
#' Colouring of file paths can be disabled by setting `LS_COLORS` to an empty
#' string e.g. `Sys.setenv(LS_COLORS = "")`.
#' @param x vector to be coerced to a fs_path object.
#' @seealso
#' <https://geoff.greer.fm/lscolors>,
#' <https://github.com/trapd00r/LS_COLORS>,
#' <https://github.com/seebi/dircolors-solarized> for some example colour
#' settings.
#' @export
#' @name fs_path
as_fs_path <- function(x) {
  UseMethod("as_fs_path")
}

#' @export
as_fs_path.NULL <- function(x) {
  new_fs_path(character())
}

#' @export
as_fs_path.character <- function(x) {
  path_tidy(x)
}

#' @rdname fs_path
#' @export
fs_path <- as_fs_path

new_fs_path <- function(x) {
  x <- enc2utf8(x)
  class(x) <- c("fs_path", "character")
  x
}
setOldClass(c("fs_path", "character"), character())

#' @export
print.fs_path <- function(x, ..., max = getOption("max.print")) {
  x <- x[seq_len(min(length(x), max))]
  if (length(x) == 0) {
    print(character(0))
  } else {
    cat(multicol(colourise_fs_path(x, ...)), sep = "")
  }

  invisible(x)
}

#' @export
`[.fs_path` <- function(x, i, ...) {
  new_fs_path(NextMethod("["))
}

#' @export
`[[.fs_path` <- function(x, i, ...) {
  new_fs_path(NextMethod("["))
}

#' @export
`/.fs_path` <- function(e1, e2) {
  path(e1, e2)
}

#' @export
`+.fs_path` <- function(e1, e2) {
  new_fs_path(paste0(e1, e2))
}

pillar_shaft.fs_path <- function(x, ..., min_width = 10, shorten = getOption("fs.fs_path.shorten", "front")) {
  pillar::new_pillar_shaft_simple(colourise_fs_path(x), ..., min_width = min_width, shorten = shorten)
}

type_sum.fs_path <- function(x) {
  "fs::path"
}

has_color <- function() {
  requireNamespace("crayon", quietly = TRUE) && crayon::has_color()
}

# From gaborcsardi/crayon/R/utils.r
multicol <- function(x) {
  if (length(x) == 0) {
    return(character())
  }
  xs <- if (has_color()) crayon::strip_style(x) else x
  max_len <- max(nchar(xs, keepNA = FALSE)) + 1
  screen_width <- getOption("width")
  num_cols <- min(length(x), max(trunc(screen_width / max_len), 1))
  if (num_cols > 1) {
    to_add <- max_len - nchar(xs, keepNA = FALSE)
    x <- paste0(x, substring(paste0(collapse = "", rep(" ", max_len)), 1, to_add))
  }
  num_rows <- ceiling(length(x) / num_cols)
  x <- c(x, rep("", num_cols * num_rows - length(x)))
  xm <- matrix(x, ncol = num_cols, byrow = TRUE)
  paste0(apply(xm, 1, paste0, collapse = ""), "\n")
}

# from defaults of dircolors version 8.28
gnu_ls_defaults <- "rs=0:di=01;34:ln=01;36:mh=00:pi=40;33:so=01;35:do=01;35:bd=40;33;01:cd=40;33;01:or=40;31;01:mi=00:su=37;41:sg=30;43:ca=30;41:tw=30;42:ow=34;42:st=37;44:ex=01;32:*.tar=01;31:*.tgz=01;31:*.arc=01;31:*.arj=01;31:*.taz=01;31:*.lha=01;31:*.lz4=01;31:*.lzh=01;31:*.lzma=01;31:*.tlz=01;31:*.txz=01;31:*.tzo=01;31:*.t7z=01;31:*.zip=01;31:*.z=01;31:*.Z=01;31:*.dz=01;31:*.gz=01;31:*.lrz=01;31:*.lz=01;31:*.lzo=01;31:*.xz=01;31:*.zst=01;31:*.tzst=01;31:*.bz2=01;31:*.bz=01;31:*.tbz=01;31:*.tbz2=01;31:*.tz=01;31:*.deb=01;31:*.rpm=01;31:*.jar=01;31:*.war=01;31:*.ear=01;31:*.sar=01;31:*.rar=01;31:*.alz=01;31:*.ace=01;31:*.zoo=01;31:*.cpio=01;31:*.7z=01;31:*.rz=01;31:*.cab=01;31:*.wim=01;31:*.swm=01;31:*.dwm=01;31:*.esd=01;31:*.jpg=01;35:*.jpeg=01;35:*.mjpg=01;35:*.mjpeg=01;35:*.gif=01;35:*.bmp=01;35:*.pbm=01;35:*.pgm=01;35:*.ppm=01;35:*.tga=01;35:*.xbm=01;35:*.xpm=01;35:*.tif=01;35:*.tiff=01;35:*.png=01;35:*.svg=01;35:*.svgz=01;35:*.mng=01;35:*.pcx=01;35:*.mov=01;35:*.mpg=01;35:*.mpeg=01;35:*.m2v=01;35:*.mkv=01;35:*.webm=01;35:*.ogm=01;35:*.mp4=01;35:*.m4v=01;35:*.mp4v=01;35:*.vob=01;35:*.qt=01;35:*.nuv=01;35:*.wmv=01;35:*.asf=01;35:*.rm=01;35:*.rmvb=01;35:*.flc=01;35:*.avi=01;35:*.fli=01;35:*.flv=01;35:*.gl=01;35:*.dl=01;35:*.xcf=01;35:*.xwd=01;35:*.yuv=01;35:*.cgm=01;35:*.emf=01;35:*.ogv=01;35:*.ogx=01;35:*.aac=00;36:*.au=00;36:*.flac=00;36:*.m4a=00;36:*.mid=00;36:*.midi=00;36:*.mka=00;36:*.mp3=00;36:*.mpc=00;36:*.ogg=00;36:*.ra=00;36:*.wav=00;36:*.oga=00;36:*.opus=00;36:*.spx=00;36:*.xspf=00;36:*.R=32:*.r=32:*.Rmd=32"

# LS mappings are at https://github.com/wertarbyte/coreutils/blob/f70c7b785b93dd436788d34827b209453157a6f2/src/dircolors.c#L60-L75

#' @importFrom stats setNames na.omit
colourise_fs_path <- function(x, ..., colors = Sys.getenv("LS_COLORS", gnu_ls_defaults)) {
  if (length(x) == 0 || !has_color() || !nzchar(colors)) {
    return(x)
  }

  perms <- try(file_info(x)$permissions, silent = TRUE)

  if (inherits(perms, "try-error")) {
    return(x)
  }

  vals <- strsplit(colors, ":")[[1]]
  nms <- strsplit(vals, "=")

  if (!(
      length(vals) == length(nms) &&
      all(lengths(nms) == 2)
      )) {
    return(x)
  }

  map <- setNames(
    vapply(nms, `[[`, character(1), 2),
    vapply(nms, `[[`, character(1), 1))
  file_types <- map[grepl("^[*][.]", names(map))]
  names(file_types) <- sub("^[*][.]", "", names(file_types))
  res <- character(length(x))

  for (i in seq_along(x)) {
    code <- map[.Call(fs_file_code_, x[[i]], as.integer(perms[[i]]))]
    if (is.na(code)) {
      code <- file_types[na.omit(tools::file_ext(x[[i]]))]
    }
    if (length(code) > 0 && !is.na(code)) {
      res[[i]] <- paste0("\033[", code, "m", x[[i]], "\033[0m")
    } else {
      res[[i]] <- x[[i]]
    }
  }
  res
}

#' @export
xtfrm.fs_path <- function(x, ...) {
  x <- unclass(x)
  NextMethod("xterm")
}

# All functions below registered in .onLoad

vec_ptype2.fs_path.fs_path <- function(x, y, ...) {
  x
}
vec_ptype2.fs_path.character <- function(x, y, ...) {
  x
}
vec_ptype2.character.fs_path <- function(x, y, ...) {
  y
}

# Note order of class is the opposite as for ptype2
vec_cast.fs_path.fs_path <- function(x, to, ...) {
  x
}
vec_cast.fs_path.character <- function(x, to, ...) {
  as_fs_path(x)
}
vec_cast.character.fs_path <- function(x, to, ...) {
  unclass(x)
}

with_dir_tree <- function(files, code, base = tempfile()) {
  if (is.null(names(files))) {
    names(files) <- rep("", length(files))
  }
  dirs <- dirname(names(files))
  unnamed <- dirs == ""
  dirs[unnamed] <- files[unnamed]
  files[unnamed] <- list(NULL)

  dir_create(path(base, dirs))
  old_wd <- setwd(base)
  on.exit({
    unlink(base, recursive = TRUE, force = TRUE)
    setwd(old_wd)
  })

  for (i in seq_along(files)) {
    if (!is.null(files[[i]])) {
      writeLines(files[[i]], con = names(files)[[i]])
    }
  }
  force(code)
}

expect_error_free <- function(...) {
  expect_error(..., regexp = NA)
}

skip_if_not_utf8 <- function() {
  skip_if_not(l10n_info()$`UTF-8`)
}

named_fs_path <- function(x) {
  new_fs_path(stats::setNames(x, x))
}

describe("as_fs_path", {
  it("returns a new fs_path object from character inputs", {
    x <- as_fs_path("foo/bar")
    expect_s3_class(x, "fs_path")
    expect_s3_class(x, "character")
    expect_equal(length(x), 1)

    x <- as_fs_path(c("foo/bar", "foo", NA))
    expect_s3_class(x, "fs_path")
    expect_s3_class(x, "character")
    expect_equal(length(x), 3)
    expect_equal(as.character(x), c("foo/bar", "foo", NA))
  })

  it("fails with non-character inputs", {
    expect_error(as_fs_path(1), "no applicable method")
    expect_error(as_fs_path(TRUE), "no applicable method")
  })

  it("preserves the class with both subset and subset2", {
    expect_s3_class(as_fs_path("foo")[1], "fs_path")
    expect_s3_class(as_fs_path("foo")[[1]], "fs_path")
  })
})

describe("colourise_fs_path", {
  it("returns an appropriately colored fs_path object", {
    withr::local_envvar(c("LS_COLORS" = gnu_ls_defaults))
    local_reproducible_output(crayon = TRUE)

    with_dir_tree(
      list(
        "foo/bar" = "test",
        "file.zip" = "test",
        "file.R" = "test",
        "file.Rmd" = "test"
        ), {
      link_create(path_abs("foo/bar"), "baz")

      # folder
      expect_equal(colourise_fs_path("foo"), "\033[01;34mfoo\033[0m")

      # file
      expect_equal(colourise_fs_path("foo/bar"), "foo/bar")

      # symlink
      expect_equal(colourise_fs_path("baz"), "\033[01;36mbaz\033[0m")

      # extension
      expect_equal(colourise_fs_path("file.zip"), "\033[01;31mfile.zip\033[0m")

      # R
      expect_equal(colourise_fs_path("file.R"), "\033[32mfile.R\033[0m")

      # Rmd
      expect_equal(colourise_fs_path("file.Rmd"), "\033[32mfile.Rmd\033[0m")
    })
  })

  it ("returns uncolored result if LS_COLORS is malformed", {
    # This has an empty leading :
    withr::local_envvar(c("LS_COLORS" = ":di=1;38;5;39:ex=1;38;5;34:ln=1;38;5;45:*.py=1;38;5;220:*.pdf=1;38;5;202:*.tex=1;38;5;196"))

    expect_equal(colourise_fs_path("foo"), "foo")
  })
})

describe("multicol", {
  files <- c("DESCRIPTION", "LICENSE.md", "NAMESPACE", "R", "README.Rmd",
    "README.md", "_pkgdown.yml", "appveyor.yml", "codecov.yml", "doc",
    "docs", "fs.Rcheck", "fs.Rproj", "fs_0.0.0.9000.tar.gz", "man",
    "man-roxygen", "script.R", "src", "tests", "tools", "foo", "foo/bar",
    "foo/bar/baz", "foo/bar/baz/bar/baz")

  it("Uses 1 column when width is less than the max size", {
    withr::with_options(c(width = 10), {
      expect_equal(max(nchar(multicol(files), keepNA = FALSE)), max(nchar(files, keepNA = FALSE)) + 1)
    })
  })

  it("Uses 2 column when width is less than the max size * 2", {
    withr::with_options(c(width = 50), {
      expect_equal(max(nchar(multicol(files), keepNA = FALSE)), 43)
    })
  })

  it("Uses 3 column when width is less than the max size * 3", {
    withr::with_options(c(width = 70), {
      expect_equal(max(nchar(multicol(files), keepNA = FALSE)), 64)
    })
  })

  it("Uses 4 column when width is less than the max size * 4", {
    withr::with_options(c(width = 90), {
      expect_equal(max(nchar(multicol(files), keepNA = FALSE)), 85)
    })
  })

  it("works with NA values", {
    withr::with_options(c(width = 10), {
      expect_equal(multicol(NA_character_), "NA\n")
      expect_equal(multicol(c("foo", NA_character_)), "foo NA  \n")
    })
  })

  it("Ignores colors when calculating width", {
    withr::with_options(c(crayon.enabled = TRUE, width = 90), {
      expect_equal(max(nchar(multicol(files), keepNA = FALSE)), 85)
    })
  })
})

test_that("common type of character and fs_path is fs_path", {
  expect_identical(
    vctrs::vec_ptype2(character(), fs_path()),
    fs_path()[0]
  )
  expect_identical(
    vctrs::vec_ptype2(fs_path(), character()),
    fs_path()[0]
  )
})

test_that("fs_path and character are coercible", {
  expect_identical(
    vctrs::vec_cast("foo", fs_path()),
    fs_path("foo")
  )
  expect_identical(
    vctrs::vec_cast(fs_path("foo"), character()),
    "foo"
  )
  expect_identical(
    vctrs::vec_cast(fs_path("foo"), fs_path()),
    fs_path("foo")
  )
})

test_that("can concatenate fs_path", {
  expect_identical(
    vctrs::vec_c(fs_path("foo"), fs_path("bar")),
    as_fs_path(c("foo", "bar"))
  )
})

describe("as_fs_path", {
  it("returns a new fs_path object from character inputs", {
    x <- as_fs_path("foo/bar")
    expect_s3_class(x, "fs_path")
    expect_s3_class(x, "character")
    expect_equal(length(x), 1)

    x <- as_fs_path(c("foo/bar", "foo", NA))
    expect_s3_class(x, "fs_path")
    expect_s3_class(x, "character")
    expect_equal(length(x), 3)
    expect_equal(as.character(x), c("foo/bar", "foo", NA))
  })

  it("fails with non-character inputs", {
    expect_error(as_fs_path(1), "no applicable method")
    expect_error(as_fs_path(TRUE), "no applicable method")
  })

  it("preserves the class with both subset and subset2", {
    expect_s3_class(as_fs_path("foo")[1], "fs_path")
    expect_s3_class(as_fs_path("foo")[[1]], "fs_path")
  })
})

describe("colourise_fs_path", {
  it("returns an appropriately colored fs_path object", {
    withr::local_envvar(c("LS_COLORS" = gnu_ls_defaults))
    withr::local_options(c(crayon.enabled = TRUE))

    with_dir_tree(
      list(
        "foo/bar" = "test",
        "file.zip" = "test",
        "file.R" = "test",
        "file.Rmd" = "test"
        ), {
      link_create(path_abs("foo/bar"), "baz")

      # folder
      expect_equal(colourise_fs_path("foo"), "\033[01;34mfoo\033[0m")

      # file
      expect_equal(colourise_fs_path("foo/bar"), "foo/bar")

      # symlink
      expect_equal(colourise_fs_path("baz"), "\033[01;36mbaz\033[0m")

      # extension
      expect_equal(colourise_fs_path("file.zip"), "\033[01;31mfile.zip\033[0m")

      # R
      expect_equal(colourise_fs_path("file.R"), "\033[32mfile.R\033[0m")

      # Rmd
      expect_equal(colourise_fs_path("file.Rmd"), "\033[32mfile.Rmd\033[0m")
    })
  })

  it ("returns uncolored result if LS_COLORS is malformed", {
    # This has an empty leading :
    withr::local_envvar(c("LS_COLORS" = ":di=1;38;5;39:ex=1;38;5;34:ln=1;38;5;45:*.py=1;38;5;220:*.pdf=1;38;5;202:*.tex=1;38;5;196"))

    expect_equal(colourise_fs_path("foo"), "foo")
  })
})

describe("multicol", {
  files <- c("DESCRIPTION", "LICENSE.md", "NAMESPACE", "R", "README.Rmd",
    "README.md", "_pkgdown.yml", "appveyor.yml", "codecov.yml", "doc",
    "docs", "fs.Rcheck", "fs.Rproj", "fs_0.0.0.9000.tar.gz", "man",
    "man-roxygen", "script.R", "src", "tests", "tools", "foo", "foo/bar",
    "foo/bar/baz", "foo/bar/baz/bar/baz")

  it("Uses 1 column when width is less than the max size", {
    withr::with_options(c(width = 10), {
      expect_equal(max(nchar(multicol(files), keepNA = FALSE)), max(nchar(files, keepNA = FALSE)) + 1)
    })
  })

  it("Uses 2 column when width is less than the max size * 2", {
    withr::with_options(c(width = 50), {
      expect_equal(max(nchar(multicol(files), keepNA = FALSE)), 43)
    })
  })

  it("Uses 3 column when width is less than the max size * 3", {
    withr::with_options(c(width = 70), {
      expect_equal(max(nchar(multicol(files), keepNA = FALSE)), 64)
    })
  })

  it("Uses 4 column when width is less than the max size * 4", {
    withr::with_options(c(width = 90), {
      expect_equal(max(nchar(multicol(files), keepNA = FALSE)), 85)
    })
  })

  it("works with NA values", {
    withr::with_options(c(width = 10), {
      expect_equal(multicol(NA_character_), "NA\n")
      expect_equal(multicol(c("foo", NA_character_)), "foo NA  \n")
    })
  })

  it("Ignores colors when calculating width", {
    withr::with_options(c(crayon.enabled = TRUE, width = 90), {
      expect_equal(max(nchar(multicol(files), keepNA = FALSE)), 85)
    })
  })
})

test_that("common type of character and fs_path is fs_path", {
  expect_identical(
    vctrs::vec_ptype2(character(), fs_path()),
    fs_path()[0]
  )
  expect_identical(
    vctrs::vec_ptype2(fs_path(), character()),
    fs_path()[0]
  )
})

test_that("fs_path and character are coercible", {
  expect_identical(
    vctrs::vec_cast("foo", fs_path()),
    fs_path("foo")
  )
  expect_identical(
    vctrs::vec_cast(fs_path("foo"), character()),
    "foo"
  )
  expect_identical(
    vctrs::vec_cast(fs_path("foo"), fs_path()),
    fs_path("foo")
  )
})

test_that("can concatenate fs_path", {
  expect_identical(
    vctrs::vec_c(fs_path("foo"), fs_path("bar")),
    as_fs_path(c("foo", "bar"))
  )
})

access_types <- c("exists" = 0L, "read" = 4L, "write" = 2L, "execute" = 1L)

#' Query for existence and access permissions
#'
#' `file_exists(path)` is a shortcut for `file_access(x, "exists")`;
#' `dir_exists(path)` and `link_exists(path)` are similar but also check that
#' the path is a directory or link, respectively. (`file_exists(path)` returns
#' `TRUE` if `path` exists and it is a directory.)
#'
#' @template fs
#' @param mode A character vector containing one or more of 'exists', 'read',
#'   'write', 'execute'.
#' @details **Cross-compatibility warning:** There is no executable bit on
#'   Windows. Checking a file for mode 'execute' on Windows, e.g.
#'   `file_access(x, "execute")` will always return `TRUE`.
#' @return A logical vector, with names corresponding to the input `path`.
#' @export
#' @examples
#' file_access("/")
#' file_access("/", "read")
#' file_access("/", "write")
#'
#' file_exists("WOMBATS")
file_access <- function(path, mode = "exists") {
  assert_no_missing(path)
  mode <- match.arg(mode, names(access_types), several.ok = TRUE)

  old <- path_expand(path)
  mode <- sum(access_types[mode])

  .Call(fs_access_, unclass(old), as.integer(mode))
}

#' @rdname file_access
#' @export
file_exists <- function(path) {
  old <- path_expand(path)
  .Call(fs_exists_, old, path)
}

#' @rdname file_access
#' @export
dir_exists <- function(path) {
  res <- is_dir(path)

  links <- is_link(path)
  res[links] <- is_dir(link_path(path[links]))

  !is.na(res) & res
}

#' @rdname file_access
#' @export
link_exists <- function(path) {
  res <- is_link(path)
  !is.na(res) & res
}

describe("file_copy", {
  it("copies an empty file and returns the new path", {
    with_dir_tree(list("foo" = ""), {
      expect_true(file_exists("foo"))
      expect_equal(file_copy("foo", "foo2"), fs_path("foo2"))
      expect_true(file_exists("foo"))
      expect_true(file_exists("foo2"))
    })
  })

  it("copies a non-empty file and returns the new path", {
    with_dir_tree(
      list("foo" = "test"), {
      expect_true(file_exists("foo"))
      expect_equal(file_copy("foo", "foo2"), fs_path("foo2"))
      expect_true(file_exists("foo"))
      expect_true(file_exists("foo2"))
      expect_equal(readLines("foo2"), readLines("foo"))
    })
  })
  it("copies multiple files", {
    with_dir_tree(list("foo" = "test", "bar" = "test2"), {
      expect_equal(file_exists(c("foo", "bar")), c(foo = TRUE, bar = TRUE))
      expect_equal(file_copy(c("foo", "bar"), c("foo2", "bar2")),
                   fs_path(c("foo2", "bar2")))
      expect_equal(readLines("foo2"), readLines("foo"))
      expect_equal(readLines("bar2"), readLines("bar"))
    })
  })
  it("errors on missing input", {
    expect_error(file_copy(NA, "foo2"), class = "invalid_argument")
    expect_error(file_copy("foo", NA), class = "invalid_argument")
  })
  with_dir_tree(list("foo/bar" = "test"), {
    it("returns the new path and copies the file", {
      expect_true(file_exists("foo/bar"))
      expect_false(file_exists("foo/bar2"))
      expect_equal(file_copy("foo/bar", "foo/bar2"), fs_path("foo/bar2"))
      expect_true(file_exists("foo/bar"))
      expect_true(file_exists("foo/bar2"))
    })
    it("fails if the new or existing directory does not exist", {
      expect_true(file_exists("foo/bar"))
      expect_error(file_copy("baz/bar", "foo/bar3"), class = "ENOENT")
      expect_error(file_copy("foo/bar", "baz/qux"), class = "ENOENT")
    })
    it("errors on missing input", {
      expect_error(file_copy(NA, "foo"), class = "invalid_argument")
      expect_error(file_copy("foo/bar", NA), class = "invalid_argument")
    })
  })
  it("copies files in different directories", {
    with_dir_tree(list("foo/bar" = "test", "foo3/bar2" = "test2", "foo2"), {
      expect_true(file_exists("foo/bar"))
      expect_true(file_exists("foo2"))
      expect_equal(
        file_copy(c("foo/bar", "foo3/bar2"), "foo2"),
        fs_path(c("foo2/bar", "foo2/bar2"))
      )
      expect_true(file_exists("foo/bar"))
      expect_true(file_exists("foo3/bar2"))
      expect_true(file_exists("foo2/bar"))
      expect_true(file_exists("foo2/bar2"))

      expect_error(file_copy("foo/bar", c("foo2", "foo3")), class = "fs_error")
    })
  })
})

describe("link_copy", {
  it("copies links and returns the new path", {
    with_dir_tree(list("foo"), {
      link_create(path_abs("foo"), "loo")
      expect_true(dir_exists("foo"))
      expect_true(link_exists("loo"))
      expect_equal(link_copy("loo", "loo2"), fs_path("loo2"))

      expect_true(link_exists("loo2"))
      expect_equal(link_path("loo2"), link_path("loo"))
    })
  })
  it("errors on missing input", {
    expect_error(link_copy(NA, "foo2"), class = "invalid_argument")
    expect_error(link_copy("foo", NA), class = "invalid_argument")
  })
})

describe("dir_copy", {
  it("copies an empty directory and returns the new path", {
    with_dir_tree(list("foo"), {
      expect_true(dir_exists("foo"))
      expect_equal(dir_copy("foo", "foo2"), fs_path("foo2"))
      expect_true(dir_exists("foo"))
      expect_true(dir_exists("foo2"))
    })
  })
  it("copies a non-empty directory and returns the new path", {
    with_dir_tree(
      list("foo/bar" = "test",
        "foo/baz" = "test2"), {
      expect_true(dir_exists("foo"))
      expect_equal(dir_copy("foo", "foo2"), fs_path("foo2"))
      expect_true(dir_exists("foo"))
      expect_true(dir_exists("foo2"))
      expect_true(file_exists("foo2/bar"))
    })
  })
  it("copies directories into other directories", {
    with_dir_tree(
      list("foo/bar" = "test"), {
        dir_create("foo2")
        expect_equal(dir_copy("foo", "foo2"), fs_path("foo2/foo"))
        expect_true(dir_exists("foo"))
        expect_true(dir_exists("foo2"))
        expect_true(dir_exists("foo2/foo"))
        expect_true(file_exists("foo2/foo/bar"))
    })
  })
  it("copies nested directories and returns the path", {
    with_dir_tree(
      list("foo/bar/baz" = "test",
        "foo/baz/qux" = "test2"), {
      expect_true(dir_exists("foo"))
      expect_equal(dir_copy("foo", "foo2"), fs_path("foo2"))
      expect_true(dir_exists("foo"))
      expect_true(dir_exists("foo2"))
      expect_true(file_exists("foo2/bar/baz"))
    })
  })
  it("can overwrite directories", {
    with_dir_tree(
      list("foo/bar" = "test",
           "foo2/bar" = "test2"), {
      dir_copy("foo", "foo2", overwrite = TRUE)
      expect_equal(readLines("foo/bar"), readLines("foo2/bar"))
    })
  })
  it("copies absolute paths", {
    with_dir_tree(
      list("foo/bar/baz" = "test",
        "foo/baz/qux" = "test2"), {
      expect_equal(dir_copy(path_abs("foo"), path_abs("foo2")), path_abs("foo2"))
      expect_true(dir_exists("foo2"))
      expect_true(dir_exists(path_abs("foo2")))
    })
  })
  it("copies hidden files, directories and links", {
    with_dir_tree(
      list("foo/.bar/.baz" = "test"), {
        link_create(path_abs("foo/.bar"), "foo/.qux")
      expect_equal(dir_copy("foo", "foo2"), fs_path("foo2"))
      expect_true(dir_exists("foo2"))
      expect_true(dir_exists("foo2/.bar"))
      expect_true(file_exists("foo2/.bar/.baz"))
      expect_true(link_exists("foo2/.qux"))
    })
  })
  it("copies links and returns the path", {
    with_dir_tree(
      list("foo/bar/baz" = "test"), {
        link_create(path_abs("foo/bar"), "foo/foo")
        expect_equal(dir_copy("foo", "foo2"), fs_path("foo2"))
        expect_true(dir_exists("foo2"))
        expect_true(link_exists("foo2/foo"))
        expect_equal(link_path("foo2/foo"), link_path("foo/foo"))
    })
  })
  it("errors on missing input", {
    expect_error(dir_copy(NA, "foo2"), class = "invalid_argument")
    expect_error(dir_copy("foo", NA), class = "invalid_argument")
  })
  it("dir_copy() works for >1 path", {
    grandparent_before <- withr::local_tempdir("dir_copy-multiple_paths-before-")
    grandparent_after <- withr::local_tempdir("dir_copy-multiple_paths-after-")

    parents <- dir_create(path(grandparent_before, c("alfa", "beta")))
    file_create(path(parents, c("apple", "banana")))

    expect_error_free(
      dir_copy(parents, path(grandparent_after, path_file(parents)))
    )
  })
})

describe("file_access", {
  with_dir_tree(list("foo/bar" = "test"), {
    it("errors on missing input", {
      expect_error(file_access(NA), class = "invalid_argument")
    })
    it("checks for file existence", {
      expect_equal(file_access("foo", "exists"), c(foo = TRUE))
    })
    it("checks for file read and write ability", {
      skip_on_os(c("windows", "solaris"))
      if (Sys.info()[["effective_user"]] == "root") skip("root user")
      file_chmod("foo/bar", "-rw")
      expect_equal(file_access("foo/bar", "read"), c("foo/bar" = FALSE))
      expect_equal(file_access("foo/bar", "write"), c("foo/bar" = FALSE))

      file_chmod("foo/bar", "+rw")
      expect_equal(file_access("foo/bar", "read"), c("foo/bar" = TRUE))
      expect_equal(file_access("foo/bar", "write"), c("foo/bar" = TRUE))
    })
    it("checks for file execute ability", {
      skip_on_os(c("windows", "solaris"))
      if (Sys.info()[["sysname"]] == "DragonFly" &&
          Sys.info()[["effective_user"]] == "root") {
        skip("file_access() is always executable for root on DragonFlyBSD")
      }
      file_chmod("foo/bar", "-x")
      expect_equal(file_access("foo/bar", "execute"), c("foo/bar" = FALSE))
      file_chmod("foo/bar", "+x")
      expect_equal(file_access("foo/bar", "execute"), c("foo/bar" = TRUE))
    })
  })
})

with_dir_tree(list("foo/bar"  = "test"), {
  link_create(path_abs("foo"), "loo")

  describe("file_exists", {
    it("returns true for files that exist, false for those that do not", {
      expect_equal(file_exists("foo"), c(foo = TRUE))
      expect_equal(file_exists("missing"), c(missing = FALSE))
      expect_equal(
        file_exists(c("foo", "missing", "foo/bar", "loo")),
        c(foo = TRUE, missing = FALSE, "foo/bar" = TRUE, "loo" = TRUE))
    })
    it("returns FALSE on missing input", {
      expect_identical(file_exists(NA_character_), structure(names = NA, FALSE))
    })
    it ("expands ~", {
      withr::with_envvar(c("R_FS_HOME" = getwd()), {
        expect_identical(file_exists("~/foo"), c("~/foo" = TRUE))
        expect_identical(file_exists("~/missing"), c("~/missing" = FALSE))
      })
    })
  })

  describe("dir_exists", {
    it("returns true for dir that exist, false for those that do not or are not directories", {
      expect_equal(dir_exists("foo"), c(foo = TRUE))
      expect_equal(dir_exists("missing"), c(missing = FALSE))
      expect_equal(
        dir_exists(c("foo", "missing", "foo/bar")),
        c(foo = TRUE, missing = FALSE, "foo/bar" = FALSE))
    })
    it("returns true for links to directories, like -d in bash", {
        expect_equal(dir_exists("loo"), c(loo = TRUE))
    })
    it("returns FALSE on missing input", {
      expect_identical(dir_exists(NA_character_), structure(names = NA, FALSE))
    })
  })

  describe("link_exists", {
    it("returns true for link that exist, false for those that do not or are not symlinks", {
      expect_equal(link_exists("loo"), c(loo = TRUE))
      expect_equal(link_exists("missing"), c(missing = FALSE))
      expect_equal(
        link_exists(c("foo", "missing", "foo/bar", "loo")),
        c(foo = FALSE, missing = FALSE, "foo/bar" = FALSE, "loo" = TRUE))
    })
    it("returns FALSE on missing input", {
      expect_identical(link_exists(NA_character_), structure(names = NA, FALSE))
    })
  })
})

describe("file_access", {
  with_dir_tree(list("foo/bar" = "test"), {
    it("errors on missing input", {
      expect_error(file_access(NA), class = "invalid_argument")
    })
    it("checks for file existence", {
      expect_equal(file_access("foo", "exists"), c(foo = TRUE))
    })
    it("checks for file read and write ability", {
      skip_on_os(c("windows", "solaris"))
      if (Sys.info()[["effective_user"]] == "root") skip("root user")
      file_chmod("foo/bar", "-rw")
      expect_equal(file_access("foo/bar", "read"), c("foo/bar" = FALSE))
      expect_equal(file_access("foo/bar", "write"), c("foo/bar" = FALSE))

      file_chmod("foo/bar", "+rw")
      expect_equal(file_access("foo/bar", "read"), c("foo/bar" = TRUE))
      expect_equal(file_access("foo/bar", "write"), c("foo/bar" = TRUE))
    })
    it("checks for file execute ability", {
      skip_on_os(c("windows", "solaris"))
      file_chmod("foo/bar", "-x")
      expect_equal(file_access("foo/bar", "execute"), c("foo/bar" = FALSE))
      file_chmod("foo/bar", "+x")
      expect_equal(file_access("foo/bar", "execute"), c("foo/bar" = TRUE))
    })
  })
})

with_dir_tree(list("foo/bar"  = "test"), {
  link_create(path_abs("foo"), "loo")

  describe("file_exists", {
    it("returns true for files that exist, false for those that do not", {
      expect_equal(file_exists("foo"), c(foo = TRUE))
      expect_equal(file_exists("missing"), c(missing = FALSE))
      expect_equal(
        file_exists(c("foo", "missing", "foo/bar", "loo")),
        c(foo = TRUE, missing = FALSE, "foo/bar" = TRUE, "loo" = TRUE))
    })
    it("returns FALSE on missing input", {
      expect_identical(file_exists(NA_character_), structure(names = NA, FALSE))
    })
    it ("expands ~", {
      withr::with_envvar(c("R_FS_HOME" = getwd()), {
        expect_identical(file_exists("~/foo"), c("~/foo" = TRUE))
        expect_identical(file_exists("~/missing"), c("~/missing" = FALSE))
      })
    })
  })

  describe("dir_exists", {
    it("returns true for dir that exist, false for those that do not or are not directories", {
      expect_equal(dir_exists("foo"), c(foo = TRUE))
      expect_equal(dir_exists("missing"), c(missing = FALSE))
      expect_equal(
        dir_exists(c("foo", "missing", "foo/bar")),
        c(foo = TRUE, missing = FALSE, "foo/bar" = FALSE))
    })
    it("returns true for links to directories, like -d in bash", {
        expect_equal(dir_exists("loo"), c(loo = TRUE))
    })
    it("returns FALSE on missing input", {
      expect_identical(dir_exists(NA_character_), structure(names = NA, FALSE))
    })
  })

  describe("link_exists", {
    it("returns true for link that exist, false for those that do not or are not symlinks", {
      expect_equal(link_exists("loo"), c(loo = TRUE))
      expect_equal(link_exists("missing"), c(missing = FALSE))
      expect_equal(
        link_exists(c("foo", "missing", "foo/bar", "loo")),
        c(foo = FALSE, missing = FALSE, "foo/bar" = FALSE, "loo" = TRUE))
    })
    it("returns FALSE on missing input", {
      expect_identical(link_exists(NA_character_), structure(names = NA, FALSE))
    })
  })
})

#' Print contents of directories in a tree-like format
#'
#' @param path A path to print the tree from
#' @inheritParams dir_ls
#' @inheritDotParams dir_ls
#'
#' @export
dir_tree <- function(path = ".", recurse = TRUE, ...) {
  files <- dir_ls(path, recurse = recurse, ...)
  by_dir <- split(files, path_dir(files))

  ch <- box_chars()

  get_coloured_name <- function(x) {
    coloured <- colourise_fs_path(x)
    sub(x, path_file(x), coloured, fixed = TRUE)
  }

  print_leaf <- function(x, indent) {
    leafs <- by_dir[[x]]
    for (i in seq_along(leafs)) {
      if (i == length(leafs)) {
        cat(indent, pc(ch$l, ch$h, ch$h, " "), get_coloured_name(leafs[[i]]), "\n", sep = "")
        print_leaf(leafs[[i]], paste0(indent, "    "))
      } else {
        cat(indent, pc(ch$j, ch$h, ch$h, " "), get_coloured_name(leafs[[i]]), "\n", sep = "")
        print_leaf(leafs[[i]], paste0(indent, pc(ch$v, "   ")))
      }
    }
  }

  cat(colourise_fs_path(path), "\n", sep = "")
  print_leaf(path_expand(path), "")

  invisible(files)
}

pc <- function(...) {
  paste0(..., collapse = "")
}

# These are derived from https://github.com/r-lib/cli/blob/e9acc82b0d20fa5c64dd529400b622c0338374ed/R/tree.R#L111
box_chars <- function() {
  if (is_utf8_output()) {
    list(
      "h" = "\u2500",                   # horizontal
      "v" = "\u2502",                   # vertical
      "l" = "\u2514",
      "j" = "\u251C"
    )
  } else {
    list(
      "h" = "-",                        # horizontal
      "v" = "|",                        # vertical
      "l" = "\\",
      "j" = "+"
    )
  }
}

is_latex_output <- function() {
  if (!("knitr" %in% loadedNamespaces())) return(FALSE)
  get("is_latex_output", asNamespace("knitr"))()
}

is_utf8_output <- function() {
  opt <- getOption("cli.unicode", NULL)
  if (! is.null(opt)) {
    isTRUE(opt)
  } else {
    l10n_info()$`UTF-8` && !is_latex_output()
  }
}

test_that("dir_tree() works", {
  withr::with_options(c("cli.unicode" = FALSE), {
    with_dir_tree(list("foo/1" = ""), {
      expect_snapshot(dir_tree("foo"))
    })
  })

  withr::with_options(c("cli.unicode" = FALSE), {
    with_dir_tree(list("foo/1" = "", "foo/2" = "", "foo/3" = "", "foo/4" = ""), {
      expect_snapshot(dir_tree("foo"))
    })
  })

  withr::with_options(c("cli.unicode" = FALSE), {
    with_dir_tree(list("foo/bar/1" = "", "foo/bar/baz/2" = "", "foo/qux/3" = "", "foo/4" = ""), {
      expect_snapshot(dir_tree("foo"))
    })
  })
})

context("test-tree.R")

test_that("dir_tree() works", {
  withr::with_options(c("cli.unicode" = FALSE), {
    reference <- fs::path_abs(test_path("ref-tree-1"))
    with_dir_tree(list("foo/1" = ""), {
      expect_known_output(dir_tree("foo"), reference)
    })
  })

  withr::with_options(c("cli.unicode" = FALSE), {
    reference <- fs::path_abs(test_path("ref-tree-2"))
    with_dir_tree(list("foo/1" = "", "foo/2" = "", "foo/3" = "", "foo/4" = ""), {
      expect_known_output(dir_tree("foo"), reference)
    })
  })

  withr::with_options(c("cli.unicode" = FALSE), {
    reference <- fs::path_abs(test_path("ref-tree-3"))
    with_dir_tree(list("foo/bar/1" = "", "foo/bar/baz/2" = "", "foo/qux/3" = "", "foo/4" = ""), {
      expect_known_output(dir_tree("foo"), reference)
    })
  })
})

#' Print contents of directories in a tree-like format
#'
#' @param path A path to print the tree from
#' @inheritParams dir_ls
#' @param ... Additional arguments passed to [dir_ls].
#'
#' @export
dir_tree <- function(path = ".", recurse = TRUE, ...) {
  files <- dir_ls(path, recurse = recurse, ...)
  by_dir <- split(files, path_dir(files))

  ch <- box_chars()

  get_coloured_name <- function(x) {
    coloured <- colourise_fs_path(x)
    sub(x, path_file(x), coloured, fixed = TRUE)
  }

  print_leaf <- function(x, indent) {
    leafs <- by_dir[[x]]
    for (i in seq_along(leafs)) {
      if (i == length(leafs)) {
        cat(indent, pc(ch$l, ch$h, ch$h, " "), get_coloured_name(leafs[[i]]), "\n", sep = "")
        print_leaf(leafs[[i]], paste0(indent, "    "))
      } else {
        cat(indent, pc(ch$j, ch$h, ch$h, " "), get_coloured_name(leafs[[i]]), "\n", sep = "")
        print_leaf(leafs[[i]], paste0(indent, pc(ch$v, "   ")))
      }
    }
  }

  cat(colourise_fs_path(path), "\n", sep = "")
  print_leaf(path_expand(path), "")

  invisible(files)
}

pc <- function(...) {
  paste0(..., collapse = "")
}

# These are derived from https://github.com/r-lib/cli/blob/e9acc82b0d20fa5c64dd529400b622c0338374ed/R/tree.R#L111
box_chars <- function() {
  if (is_utf8_output()) {
    list(
      "h" = "\u2500",                   # horizontal
      "v" = "\u2502",                   # vertical
      "l" = "\u2514",
      "j" = "\u251C"
    )
  } else {
    list(
      "h" = "-",                        # horizontal
      "v" = "|",                        # vertical
      "l" = "\\",
      "j" = "+"
    )
  }
}

is_latex_output <- function() {
  if (!("knitr" %in% loadedNamespaces())) return(FALSE)
  get("is_latex_output", asNamespace("knitr"))()
}

is_utf8_output <- function() {
  opt <- getOption("cli.unicode", NULL)
  if (! is.null(opt)) {
    isTRUE(opt)
  } else {
    l10n_info()$`UTF-8` && !is_latex_output()
  }
}

# nocov start
.onLoad <- function(...) {
  register_s3_method("pillar", "pillar_shaft", "fs_path")
  register_s3_method("pillar", "type_sum", "fs_path")
  register_s3_method("testthat", "compare", "fs_path")
  register_s3_method("vctrs", "vec_ptype2", "fs_path.fs_path")
  register_s3_method("vctrs", "vec_ptype2", "character.fs_path")
  register_s3_method("vctrs", "vec_ptype2", "fs_path.character")
  register_s3_method("vctrs", "vec_cast", "fs_path.fs_path")
  register_s3_method("vctrs", "vec_cast", "character.fs_path")
  register_s3_method("vctrs", "vec_cast", "fs_path.character")

  register_s3_method("pillar", "pillar_shaft", "fs_bytes")
  register_s3_method("pillar", "type_sum", "fs_bytes")
  register_s3_method("vctrs", "vec_ptype2", "fs_bytes.fs_bytes")
  register_s3_method("vctrs", "vec_ptype2", "double.fs_bytes")
  register_s3_method("vctrs", "vec_ptype2", "fs_bytes.double")
  register_s3_method("vctrs", "vec_cast", "fs_bytes.fs_bytes")
  register_s3_method("vctrs", "vec_cast", "double.fs_bytes")
  register_s3_method("vctrs", "vec_cast", "fs_bytes.double")
  register_s3_method("vctrs", "vec_ptype2", "integer.fs_bytes")
  register_s3_method("vctrs", "vec_ptype2", "fs_bytes.integer")
  register_s3_method("vctrs", "vec_cast", "integer.fs_bytes")
  register_s3_method("vctrs", "vec_cast", "fs_bytes.integer")

  register_s3_method("pillar", "pillar_shaft", "fs_perms")
  register_s3_method("pillar", "type_sum", "fs_perms")
  register_s3_method("testthat", "compare", "fs_perms")
  register_s3_method("vctrs", "vec_ptype2", "fs_perms.fs_perms")
  register_s3_method("vctrs", "vec_ptype2", "double.fs_perms")
  register_s3_method("vctrs", "vec_ptype2", "fs_perms.double")
  register_s3_method("vctrs", "vec_cast", "fs_perms.fs_perms")
  register_s3_method("vctrs", "vec_cast", "double.fs_perms")
  register_s3_method("vctrs", "vec_cast", "fs_perms.double")
  register_s3_method("vctrs", "vec_ptype2", "integer.fs_perms")
  register_s3_method("vctrs", "vec_ptype2", "fs_perms.integer")
  register_s3_method("vctrs", "vec_cast", "integer.fs_perms")
  register_s3_method("vctrs", "vec_cast", "fs_perms.integer")
  register_s3_method("vctrs", "vec_ptype2", "character.fs_perms")
  register_s3_method("vctrs", "vec_ptype2", "fs_perms.character")
  register_s3_method("vctrs", "vec_cast", "character.fs_perms")
  register_s3_method("vctrs", "vec_cast", "fs_perms.character")

  invisible()
}

register_s3_method <- function(pkg, generic, class, fun = NULL) {
  stopifnot(is.character(pkg), length(pkg) == 1)
  stopifnot(is.character(generic), length(generic) == 1)
  stopifnot(is.character(class), length(class) == 1)

  if (is.null(fun)) {
    fun <- get(paste0(generic, ".", class), envir = parent.frame())
  } else {
    stopifnot(is.function(fun))
  }

  if (pkg %in% loadedNamespaces()) {
    registerS3method(generic, class, fun, envir = asNamespace(pkg))
  }

  # Always register hook in case package is later unloaded & reloaded
  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      registerS3method(generic, class, fun, envir = asNamespace(pkg))
    }
  )
}
# nocov end

units <- c('B' = 1, 'K' = 1024, 'M' = 1024 ^ 2, 'G' = 1024 ^ 3, 'T' = 1024 ^ 4, 'P' = 1024 ^ 5, 'E' = 1024 ^ 6, 'Z' = 1024 ^ 7, 'Y' = 1024 ^ 8)

#' Human readable file sizes
#'
#' Construct, manipulate and display vectors of file sizes. These are numeric
#' vectors, so you can compare them numerically, but they can also be compared
#' to human readable values such as '10MB'.
#'
#' @param x A numeric or character vector. Character representations can use
#'   shorthand sizes (see examples).
#' @examples
#' fs_bytes("1")
#' fs_bytes("1K")
#' fs_bytes("1Kb")
#' fs_bytes("1Kib")
#' fs_bytes("1MB")
#'
#' fs_bytes("1KB") < "1MB"
#'
#' sum(fs_bytes(c("1MB", "5MB", "500KB")))
#' @name fs_bytes
#' @export
as_fs_bytes <- function(x) {
  UseMethod("as_fs_bytes")
}

#' @export
#' @rdname fs_bytes
fs_bytes <- as_fs_bytes

new_fs_bytes <- function(x) {
  class(x) <- c("fs_bytes", "numeric")
  x
}
setOldClass(c("fs_bytes", "numeric"), numeric())

#' @export
as_fs_bytes.default <- function(x = numeric()) {
  if (length(x) == 0) {
    return(new_fs_bytes(numeric()))
  }
  x <- as.character(x)
  m <- captures(x, regexpr("^(?<size>[[:digit:].]+)\\s*(?<unit>[KMGTPEZY]?)i?[Bb]?$", x, perl = TRUE))
  m$unit[m$unit == ""] <- "B"
  new_fs_bytes(unname(as.numeric(m$size) * units[m$unit]))
}

#' @export
as_fs_bytes.fs_bytes <- function(x) {
  return(x)
}

#' @export
as_fs_bytes.numeric <- function(x) {
  new_fs_bytes(x)
}

# Adapted from https://github.com/gaborcsardi/prettyunits
# Aims to be consistent with ls -lh, so uses 1024 KiB units, 3 or less digits etc.
#' @export
format.fs_bytes <- function(x, scientific = FALSE, digits = 3, drop0trailing = TRUE, ...) {
  bytes <- unclass(x)

  exponent <- pmin(floor(log(bytes, 1024)), length(units) - 1)
  res <- round(bytes / 1024 ^ exponent, 2)
  unit <- ifelse (exponent == 0, "", names(units)[exponent + 1])

  ## Zero bytes
  res[bytes == 0] <- 0
  unit[bytes == 0] <- ""

  ## NA and NaN bytes
  res[is.na(bytes)] <- NA_real_
  res[is.nan(bytes)] <- NaN
  unit[is.na(bytes)] <- ""            # Includes NaN as well

  res <- format(res, scientific = scientific, digits = digits, drop0trailing = drop0trailing, ...)

  paste0(res, unit)
}

#' @export
as.character.fs_bytes <- format.fs_bytes

#' @export
print.fs_bytes <- function(x, ...) {
  cat(format.fs_bytes(x, ...))
}

#' @export
sum.fs_bytes <- function(x, ...) {
  new_fs_bytes(NextMethod())
}

#' @export
min.fs_bytes <- function(x, ...) {
  new_fs_bytes(NextMethod())
}

#' @export
max.fs_bytes <- function(x, ...) {
  new_fs_bytes(NextMethod())
}

#' @export
`[.fs_bytes` <- function(x, i, ...) {
  new_fs_bytes(NextMethod("["))
}

#' @export
`[[.fs_bytes` <- function(x, i, ...) {
  new_fs_bytes(NextMethod("[["))
}

#' @export
# Adapted from Ops.numeric_version
Ops.fs_bytes <- function (e1, e2) {
  if (nargs() == 1L) {
    stop(sprintf("unary '%s' not defined for \"fs_bytes\" objects", .Generic),
      call. = FALSE)
  }

  boolean <- switch(.Generic,
    `+` = TRUE,
    `-` = TRUE,
    `*` = TRUE,
    `/` = TRUE,
    `^` = TRUE,
    `<` = TRUE,
    `>` = TRUE,
    `==` = TRUE,
    `!=` = TRUE,
    `<=` = TRUE,
    `>=` = TRUE,
  FALSE)
  if (!boolean) {
    stop(sprintf("'%s' not defined for \"fs_bytes\" objects", .Generic),
      call. = FALSE)
  }
  e1 <- as_fs_bytes(e1)
  e2 <- as_fs_bytes(e2)
  NextMethod(.Generic)
}

pillar_shaft.fs_bytes <- function(x, ...) {
  pillar::new_pillar_shaft_simple(format.fs_bytes(x), align = "right", ...)
}

type_sum.fs_bytes <- function(x) {
  "fs::bytes"
}

# All functions below registered in .onLoad

vec_ptype2.fs_bytes.fs_bytes <- function(x, y, ...) {
  x
}
vec_ptype2.fs_bytes.double <- function(x, y, ...) {
  x
}
vec_ptype2.double.fs_bytes <- function(x, y, ...) {
  y
}

# Note order of class is the opposite as for ptype2
vec_cast.fs_bytes.fs_bytes <- function(x, to, ...) {
  x
}
vec_cast.fs_bytes.double <- function(x, to, ...) {
  as_fs_bytes(x)
}
vec_cast.double.fs_bytes <- function(x, to, ...) {
  unclass(x)
}

vec_ptype2.fs_bytes.integer <- function(x, y, ...) {
  x
}
vec_ptype2.integer.fs_bytes <- function(x, y, ...) {
  y
}

vec_cast.fs_bytes.integer <- function(x, to, ...) {
  as_fs_bytes(x)
}
vec_cast.integer.fs_bytes <- function(x, to, ...) {
  unclass(x)
}

describe("as_fs_bytes", {
  it("accepts numeric input unchanged", {
    expect_equal(unclass(as_fs_bytes(123L)), 123L)
    expect_equal(unclass(as_fs_bytes(123)), 123)
  })
  it("accepts fs_byte input unchanged", {
    x <- as_fs_bytes(123)
    expect_equal(as_fs_bytes(x), x)
  })
  it("coerces character input", {
    expect_equal(unclass(as_fs_bytes("1")), 1)
    expect_equal(unclass(as_fs_bytes("1K")), 1024)
    expect_equal(unclass(as_fs_bytes("1M")), 1024 * 1024)
    expect_equal(unclass(as_fs_bytes("10M")), 10 * 1024 * 1024)
    expect_equal(unclass(as_fs_bytes("1G")), 1024 * 1024 * 1024)
  })
})

describe("format.fs_bytes", {
  it("formats bytes under 1024 as whole numbers", {
    expect_equal(format(fs_bytes(0)), "0")
    expect_equal(format(fs_bytes(1)), "1")
    expect_equal(format(fs_bytes(1023)), "1023")
  })
  it("formats bytes 1024 and up as abbreviated numbers", {
    expect_equal(format(fs_bytes(1024)), "1K")
    expect_equal(format(fs_bytes(1025)), "1K")
    expect_equal(format(fs_bytes(2^16)), "64K")
    expect_equal(format(fs_bytes(2^24)), "16M")
    expect_equal(format(fs_bytes(2^24 + 555555)), "16.5M")
    expect_equal(format(fs_bytes(2^32)), "4G")
    expect_equal(format(fs_bytes(2^48)), "256T")
    expect_equal(format(fs_bytes(2^64)), "16E")
  })
  it("handles NA and NaN", {
    expect_equal(format(fs_bytes(NA)), "NA")
    expect_equal(format(fs_bytes(NaN)), "NaN")
  })
  it("works with vectors", {
    v <- c(NA, 1, 2^13, 2^20, NaN, 2^15)
    expect_equal(
      format(fs_bytes(v), trim = TRUE),
      c("NA", "1", "8K", "1M", "NaN", "32K"))

    expect_equal(format(fs_bytes(numeric())), character())
  })
})

describe("sum.fs_bytes", {
  it("sums its input and returns a fs_byte", {
    expect_equal(sum(fs_bytes(0)), new_fs_bytes(0))
    expect_equal(sum(fs_bytes(c(1, 2))), new_fs_bytes(3))
    expect_equal(sum(fs_bytes(c(1, NA))), new_fs_bytes(NA_real_))
  })
})

describe("min.fs_bytes", {
  it("finds minimum input and returns a fs_byte", {
    expect_equal(min(fs_bytes(0)), new_fs_bytes(0))
    expect_equal(min(fs_bytes(c(1, 2))), new_fs_bytes(1))
    expect_equal(min(fs_bytes(c(1, NA))), new_fs_bytes(NA_real_))
  })
})

describe("max.fs_bytes", {
  it("finds maximum input and returns a fs_byte", {
    expect_equal(max(fs_bytes(0)), new_fs_bytes(0))
    expect_equal(max(fs_bytes(c(1, 2))), new_fs_bytes(2))
    expect_equal(max(fs_bytes(c(1, NA))), new_fs_bytes(NA_real_))
  })
})

describe("[.fs_bytes", {
  it("retains the fs_bytes class", {
    x <- fs_bytes(c(100, 200, 300))
    expect_equal(x[], x)
    expect_equal(x[1], new_fs_bytes(100))
    expect_equal(x[1:2], new_fs_bytes(c(100, 200)))
  })
})

describe("[[.fs_bytes", {
  it("retains the fs_bytes class", {
    x <- fs_bytes(c(100, 200, 300))
    expect_equal(x[[1]], new_fs_bytes(100))
  })
})

describe("Ops.fs_bytes", {
  it("errors for unary operators", {
    x <- fs_bytes(c(100, 200, 300))
    expect_error(!x, "unary '!' not defined for \"fs_bytes\" objects")
    expect_error(+x, "unary '\\+' not defined for \"fs_bytes\" objects")
    expect_error(-x, "unary '-' not defined for \"fs_bytes\" objects")
  })

  it("works with boolean comparison operators", {
    x <- fs_bytes(c(100, 200, 300))

    expect_equal(x == 100, c(TRUE, FALSE, FALSE))
    expect_equal(x != 100, c(FALSE, TRUE, TRUE))
    expect_equal(x > 100, c(FALSE, TRUE, TRUE))
    expect_equal(x >= 100, c(TRUE, TRUE, TRUE))
    expect_equal(x < 200, c(TRUE, FALSE, FALSE))
    expect_equal(x <= 200, c(TRUE, TRUE, FALSE))
  })

  it("works with arithmetic operators", {
    x <- fs_bytes(c(100, 200, 300))

    expect_equal(x + 100, fs_bytes(c(200, 300, 400)))
    expect_equal(x - 100, fs_bytes(c(0, 100, 200)))
    expect_equal(x * 100, fs_bytes(c(10000, 20000, 30000)))
    expect_equal(x / 2, fs_bytes(c(50, 100, 150)))
    expect_equal(x ^ 2, fs_bytes(c(10000, 40000, 90000)))
  })

  it("errors for other binary operators", {
    x <- fs_bytes(c(100, 200, 300))
    expect_error(x %% 2, "'%%' not defined for \"fs_bytes\" objects")
    expect_error(x %/% 2, "'%/%' not defined for \"fs_bytes\" objects")
    expect_error(x & TRUE, "'&' not defined for \"fs_bytes\" objects")
    expect_error(x | TRUE, "'|' not defined for \"fs_bytes\" objects")
  })
})

test_that("common type of double and fs_bytes is fs_bytes", {
  expect_identical(
    vctrs::vec_ptype2(double(), fs_bytes()),
    fs_bytes()[0]
  )
  expect_identical(
    vctrs::vec_ptype2(fs_bytes(), double()),
    fs_bytes()[0]
  )
})

test_that("fs_bytes and double are coercible", {
  expect_identical(
    vctrs::vec_cast(1, fs_bytes()),
    fs_bytes(1)
  )
  expect_identical(
    vctrs::vec_cast(fs_bytes(1), double()),
    1
  )
  expect_identical(
    vctrs::vec_cast(fs_bytes(1), fs_bytes()),
    fs_bytes(1)
  )
})

test_that("can concatenate fs_bytes", {
  expect_identical(
    vctrs::vec_c(fs_bytes(1), fs_bytes(2)),
    as_fs_bytes(c(1, 2))
  )
})

test_that("common type of integer and fs_bytes is fs_bytes", {
  expect_identical(
    vctrs::vec_ptype2(integer(), fs_bytes()),
    fs_bytes()[0]
  )
  expect_identical(
    vctrs::vec_ptype2(fs_bytes(), integer()),
    fs_bytes()[0]
  )
})

test_that("fs_bytes and integer are coercible", {
  expect_identical(
    vctrs::vec_cast(1L, fs_bytes()),
    fs_bytes(1L)
  )
  expect_identical(
    vctrs::vec_cast(fs_bytes(1L), integer()),
    1L
  )
})

#' Functions to test for file types
#'
#' @return A named logical vector, where the names give the paths. If the given
#'   object does not exist, `NA` is returned.
#' @seealso [file_exists()], [dir_exists()] and [link_exists()] if you want
#'   to ensure that the path also exists.
#' @template fs
#' @importFrom stats setNames
#' @inheritParams file_info
#' @export
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' dir_create("d")
#'
#' file_create("d/file.txt")
#' dir_create("d/dir")
#' link_create(path(path_abs("d"), "file.txt"), "d/link")
#'
#' paths <- dir_ls("d")
#' is_file(paths)
#' is_dir(paths)
#' is_link(paths)
#'
#' # Cleanup
#' dir_delete("d")
#' \dontshow{setwd(.old_wd)}
is_file <- function(path, follow = TRUE) {
  res <- file_info(path, follow = follow)
  setNames(!is.na(res$type) & res$type == "file", path)
}

#' @rdname is_file
#' @export
is_dir <- function(path, follow = TRUE) {
  res <- file_info(path, follow = follow)
  setNames(!is.na(res$type) & res$type == "directory", path)
}

#' @rdname is_file
#' @export
is_link <- function(path) {
  res <- file_info(path)
  setNames(!is.na(res$type) & res$type == "symlink", path)
}

#' @rdname is_file
#' @export
is_file_empty <- function(path, follow = TRUE) {
  res <- file_info(path, follow = follow)

  setNames(!is.na(res$size) & res$size == 0, res$path)
}

#' Test if a path is an absolute path
#'
#' @template fs
#' @export
#' @examples
#' is_absolute_path("/foo")
#' is_absolute_path("C:\\foo")
#' is_absolute_path("\\\\myserver\\foo\\bar")
#'
#' is_absolute_path("foo/bar")
is_absolute_path <- function(path) {
  tidy_path <- path_tidy(path)
  grepl("^~", path) | grepl("^(/+|[A-Za-z]:)", tidy_path)
}

with_dir_tree(list("foo/bar"  = "test"), {
  skip_on_os("windows")
  link_create(path_abs("foo/bar"), "foo2")
  link_create("foo/bar", "relbar")
  link_create("foo", "relfoo")

  describe("is_file", {
    it("returns true for files, false for non-files, NA if no object exists", {
      expect_true(is_file("foo/bar"))
      expect_false(is_file("foo"))
      expect_false(is_file("foo2", follow = FALSE))
      expect_true(is_file("foo2", follow = TRUE))
      expect_equal(is_file("baz"), c(baz = FALSE))
      .old_wd <- setwd("foo")
      expect_false(is_file("../relbar", follow = FALSE))
      expect_true(is_file("../relbar", follow = TRUE))
      setwd(.old_wd)
    })
  })

  describe("is_dir", {
    it("returns true for dirs, false for non-dirs, NA if no object exists", {
      expect_true(is_dir("foo"))
      expect_false(is_dir("foo/bar"))
      expect_false(is_dir("foo2", follow = FALSE))
      expect_false(is_dir("foo2", follow = TRUE))
      expect_equal(is_dir("baz"), c(baz = FALSE))
      .old_wd <- setwd("foo")
      expect_false(is_dir("../relfoo", follow = FALSE))
      expect_true(is_dir("../relfoo", follow = TRUE))
      setwd(.old_wd)
    })
  })

  describe("is_link", {
    it("returns true for links, false for non-links, NA if no object exists", {
      expect_true(is_link("foo2"))
      expect_false(is_link("foo"))
      expect_false(is_link("foo/bar"))
      expect_equal(is_link("baz"), c(baz = FALSE))
    })
  })

  describe("is_file_empty", {
    it("returns true for empty files and false for others", {
      file_create("blah")
      expect_true(is_file_empty("blah"))
      expect_false(is_file_empty("foo/bar"))
      expect_false(is_file_empty("baz"))
    })
  })
})

with_dir_tree(list("foo/bar"  = "test"), {
  describe("is_absolute_path", {
    it("detects windows absolute paths", {
      expect_true(is_absolute_path("c:\\"))
      expect_true(is_absolute_path("c:/"))
      expect_true(is_absolute_path("P:/"))
      expect_true(is_absolute_path("P:\\"))
      expect_true(is_absolute_path("\\\\server\\mountpoint\\"))
      expect_true(is_absolute_path("\\foo"))
      expect_true(is_absolute_path("\\foo\\bar"))
    })
    it("detects posix absolute paths", {
      expect_false(is_absolute_path(""))
      expect_false(is_absolute_path("foo/bar"))
      expect_false(is_absolute_path("./foo/bar"))
      expect_false(is_absolute_path("../foo/bar"))

      expect_true(is_absolute_path("/"))
      expect_true(is_absolute_path("/foo"))
      expect_true(is_absolute_path("/foo/bar"))
      expect_true(is_absolute_path("~/foo/bar"))
    })
  })
})

#' Query file metadata
#'
#' Compared to `[file.info]` the full results of a `stat(2)` system call are
#' returned and some columns are returned as S3 classes to make manipulation
#' more natural. On systems which do not support all metadata (such as Windows)
#' default values are used.
#' @template fs
#' @param follow If `TRUE`, symbolic links will be followed (recursively) and
#'   the results will be that of the final file rather than the link.
#' @inheritParams dir_ls
#' @return A data.frame with metadata for each file. Columns returned are as follows.
#'  \item{path}{The input path, as a [fs_path()] character vector.}
#'  \item{type}{The file type, as a factor of file types.}
#'  \item{size}{The file size, as a [fs_bytes()] numeric vector.}
#'  \item{permissions}{The file permissions, as a [fs_perms()] integer vector.}
#'  \item{modification_time}{The time of last data modification, as a [POSIXct] datetime.}
#'  \item{user}{The file owner name - as a character vector.}
#'  \item{group}{The file group name - as a character vector.}
#'  \item{device_id}{The file device id - as a numeric vector.}
#'  \item{hard_links}{The number of hard links to the file - as a numeric vector.}
#'  \item{special_device_id}{The special device id of the file - as a numeric vector.}
#'  \item{inode}{The inode of the file - as a numeric vector.}
#'  \item{block_size}{The optimal block for the file - as a numeric vector.}
#'  \item{blocks}{The number of blocks allocated for the file - as a numeric vector.}
#'  \item{flags}{The user defined flags for the file - as an integer vector.}
#'  \item{generation}{The generation number for the file - as a numeric vector.}
#'  \item{access_time}{The time of last access - as a [POSIXct] datetime.}
#'  \item{change_time}{The time of last file status change - as a [POSIXct] datetime.}
#'  \item{birth_time}{The time when the inode was created - as a [POSIXct] datetime.}
#' @seealso [dir_info()] to display file information for files in a given
#'   directory.
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' write.csv(mtcars, "mtcars.csv")
#' file_info("mtcars.csv")
#'
#' # Files in the working directory modified more than 20 days ago
#' files <- file_info(dir_ls())
#' files$path[difftime(Sys.time(), files$modification_time, units = "days") > 20]
#'
#' # Cleanup
#' file_delete("mtcars.csv")
#' \dontshow{setwd(.old_wd)}
#' @export
file_info <- function(path, fail = TRUE, follow = FALSE) {
  old <- path_expand(path)

  res <- .Call(fs_stat_, old, fail)

  res$path <- path_tidy(path)

  res$type <- factor(res$type, levels = file_types, labels = names(file_types))

  # TODO: convert to UTC times?
  res$access_time <- .POSIXct(res$access_time)
  res$modification_time <- .POSIXct(res$modification_time)
  res$change_time <- .POSIXct(res$change_time)
  res$birth_time <- .POSIXct(res$birth_time)

  important <- c("path", "type", "size", "permissions", "modification_time", "user", "group")
  res <- res[c(important, setdiff(names(res), important))]

  is_symlink <- !is.na(res$type) & res$type == "symlink"
  while(follow && any(is_symlink)) {
    lpath <- link_path(path[is_symlink])
    lpath <- ifelse(is_absolute_path(lpath), lpath, path(path_dir(path[is_symlink]), lpath))
    res[is_symlink, ] <- file_info(lpath, fail = fail, follow = FALSE)
    is_symlink <- !is.na(res$type) & res$type == "symlink"
  }

  res
  as_tibble(res)
}

#' @export
#' @rdname file_info
file_size <- function(path, fail = TRUE) {
  res <- file_info(path, fail)

  stats::setNames(res$size, path)
}

file_types <- c(
  "any" = -1,
  "block_device" = 0L,
  "character_device" = 1L,
  "directory" = 2L,
  "FIFO" = 3L,
  "symlink" = 4L,
  "file" = 5L,
  "socket" = 6L)

#' Change file permissions
#' @template fs
#' @param mode A character representation of the mode, in either hexidecimal or symbolic format.
#' @details **Cross-compatibility warning:** File permissions differ on Windows
#'   from POSIX systems. Windows does not use an executable bit, so attempting
#'   to change this will have no effect. Windows also does not have user
#'   groups, so only the user permissions (`u`) are relevant.
#' @export
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' file_create("foo", mode = "000")
#' file_chmod("foo", "777")
#' file_info("foo")$permissions
#'
#' file_chmod("foo", "u-x")
#' file_info("foo")$permissions
#'
#' file_chmod("foo", "a-wrx")
#' file_info("foo")$permissions
#'
#' file_chmod("foo", "u+wr")
#' file_info("foo")$permissions
#'
#' # It is also vectorized
#' files <- c("foo", file_create("bar", mode = "000"))
#' file_chmod(files, "a+rwx")
#' file_info(files)$permissions
#'
#' file_chmod(files, c("644", "600"))
#' file_info(files)$permissions
#' \dontshow{setwd(.old_wd)}
file_chmod <- function(path, mode) {
  assert_no_missing(path)
  mode <- as_fs_perms(mode, mode = file_info(path)$permissions)

  old <- path_expand(path)

  .Call(fs_chmod_, old, as.integer(mode))

  invisible(path_tidy(path))
}

#' Change owner or group of a file
#' @template fs
#' @param user_id The user id of the new owner, specified as a numeric ID or
#'   name. The R process must be privileged to change this.
#' @param group_id The group id of the new owner, specified as a numeric ID or
#'   name.
#' @export
file_chown <- function(path, user_id = NULL, group_id = NULL) {
  assert_no_missing(path)

  old <- path_expand(path)

  if (is.null(user_id)) {
    user_id <- -1
  }

  if (is.null(group_id)) {
    group_id <- -1
  }

  if (is.character(user_id)) {
    user_id <- .Call(fs_getpwnam_, user_id)
  }

  if (is.character(group_id)) {
    group_id <- .Call(fs_getgrnam_, group_id)
  }

  .Call(fs_chown_, old, as.integer(user_id), as.integer(group_id))

  invisible(path_tidy(path))
}

#' Open files or directories
#'
#' @template fs
#' @return The directories that were opened (invisibly).
#' @importFrom utils browseURL
#' @inheritParams utils::browseURL
#' @export
file_show <- function(path = ".", browser = getOption("browser")) {
  assert_no_missing(path)

  old <- path_expand(path)

  for (p in path) {
    browseURL(p, browser = browser)
  }

  invisible(path_tidy(path))
}


#' Move or rename files
#'
#' Compared to [file.rename] `file_move()` always fails if it is unable to move
#' a file, rather than signaling a Warning and returning an error code.
#' @template fs
#' @param new_path New file path. If `new_path` is existing directory, the file
#'   will be moved into that directory; otherwise it will be moved/renamed to
#'   the full path.
#'
#'   Should either be the same length as `path`, or a single directory.
#' @return The new path (invisibly).
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' file_create("foo")
#' file_move("foo", "bar")
#' file_exists(c("foo", "bar"))
#' file_delete("bar")
#' \dontshow{setwd(.old_wd)}
#' @export
file_move <- function(path, new_path) {
  assert_no_missing(path)
  assert_no_missing(new_path)

  old <- path_expand(path)
  new <- path_expand(new_path)

  is_directory <- file_exists(new) & is_dir(new)

  if (length(new) == 1 && is_directory[[1]]) {
    new <- rep(new, length(path))
  }
  assert("Length of `path` must equal length of `new_path`", length(old) == length(new))

  new[is_directory] <- path(new[is_directory], basename(old))

  .Call(fs_move_, old, new)

  invisible(path_tidy(new))
}

#' Change file access and modification times
#'
#' Unlike the touch POSIX utility this does not create the file if it does not
#' exist. Use [file_create()] to do this if needed.
#'
#' @template fs
#'
#' @param access_time,modification_time The times to set, inputs will be
#'   coerced to [POSIXct] objects.
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' file_create("foo")
#' file_touch("foo", "2018-01-01")
#' file_info("foo")[c("access_time", "modification_time", "change_time", "birth_time")]
#' \dontshow{setwd(.old_wd)}
#' @export
file_touch <- function(path, access_time = Sys.time(), modification_time = access_time) {
  assert_no_missing(path)

  access_time <- as.POSIXct(access_time)
  modification_time <- as.POSIXct(modification_time)

  path <- path_expand(path)

  .Call(fs_create_, path, 420L)
  .Call(fs_touch_, path, access_time, modification_time)

  invisible(path_tidy(path))
}

describe("file_info", {
  with_dir_tree(list(
      "foo/bar" = "test",
      "NA" = ""), {
    link_create(path_abs("foo"), "foo2")

    it("returns a correct tibble", {
      x <- file_info(c("foo", "foo/bar", "foo2", "NA"))
      expect_s3_class(x, c("tbl", "tbl_df", "data.frame"))
      expect_length(x, 18)
      expect_equal(nrow(x), 4)
      expect_equal(as.character(x$path), c("foo", "foo/bar", "foo2", "NA"))
      expect_equal(as.character(x$type), c("directory", "file", "symlink", "file"))
    })

    it("returns NA if a file does not exist", {
      x <- file_info("missing")
      expect_s3_class(x, c("tbl", "tbl_df", "data.frame"))
      expect_length(x, 18)
      expect_equal(nrow(x), 1)
      expect_equal(as.character(x$path), "missing")
      expect_equal(sum(is.na(x)), 17)
    })
    it("returns NA on NA input", {
      x <- file_info(NA_character_)
      expect_s3_class(x, c("tbl", "tbl_df", "data.frame"))
      expect_length(x, 18)
      expect_equal(nrow(x), 1)
      expect_equal(as.character(x$path), NA_character_)
      expect_equal(sum(is.na(x)), 18)
    })
    it("can be subset as a data.frame", {
      x <- file_info("foo/bar")
      class(x) <- "data.frame"
      for (col in seq_along(x)) {
        expect_true(length(x[[1, col]]) == 1)
      }
    })
    it("can be subset as a tibble", {
      x <- file_info("foo/bar")
      expect_s3_class(x, "tbl_df")
      for (col in seq_along(x)) {
        expect_true(length(x[[1, col]]) == 1)
      }
    })
  })
})

describe("file_size", {
  with_dir_tree(list(
      "foo" = "test",
      "bar" = "test2"), {
      it("returns the correct file size", {
        expect_equal(file_size("foo"), stats::setNames(file_info("foo")$size, "foo"))
        expect_equal(file_size("bar"), stats::setNames(file_info("bar")$size, "bar"))
      })
      it("returns an object of class fs_bytes, numeric (#239)", {
        expect_equal(class(file_size("foo")), c("fs_bytes", "numeric"))
        expect_equal(class(file_size("foo")[]), c("fs_bytes", "numeric"))
      })
  })
})

# Windows permissions only really allow setting read only, so we will skip
# the more thorough tests
if (!is_windows()) {
  describe("file_chmod", {
    with_dir_tree(list("foo/bar" = "test"), {
      it("returns the input path and changes permissions with symbolic input", {
        expect_equal(file_chmod("foo/bar", "u=rw,go=r"), fs_path("foo/bar"))
        expect_equal(
          as.character(file_info("foo/bar")$permissions),
          as.character(fs_perms("644"))
        )
        expect_equal(file_chmod("foo/bar", "u+x"), fs_path("foo/bar"))
        expect_equal(
          as.character(file_info("foo/bar")$permissions),
          as.character(fs_perms("744"))
        )
        expect_equal(file_chmod("foo/bar", "g-r"), fs_path("foo/bar"))
        expect_equal(
          as.character(file_info("foo/bar")$permissions),
          as.character(fs_perms("704"))
        )
        expect_equal(file_chmod("foo/bar", "o-r"), fs_path("foo/bar"))
        expect_equal(
          as.character(file_info("foo/bar")$permissions),
          as.character(fs_perms("700"))
        )
      })

      it("returns the input path and changes permissions with octal input", {
        expect_equal(file_chmod("foo/bar", "644"), fs_path("foo/bar"))
        expect_equal(
          as.character(file_info("foo/bar")$permissions),
          as.character(fs_perms("644"))
        )
        expect_equal(file_chmod("foo/bar", "744"), fs_path("foo/bar"))
        expect_equal(
          as.character(file_info("foo/bar")$permissions),
          as.character(fs_perms("744"))
        )
        expect_equal(file_chmod("foo/bar", "704"), fs_path("foo/bar"))
        expect_equal(
          as.character(file_info("foo/bar")$permissions),
          as.character(fs_perms("704"))
        )
        expect_equal(file_chmod("foo/bar", "700"), fs_path("foo/bar"))
        expect_equal(
          as.character(file_info("foo/bar")$permissions),
          as.character(fs_perms("700"))
        )
      })

      it("returns the input path and changes permissions with display input", {
        expect_equal(file_chmod("foo/bar", "rw-r--r--"), fs_path("foo/bar"))
        expect_true(file_info("foo/bar")$permissions == "644")
        expect_equal(file_chmod("foo/bar", "rwxr--r--"), fs_path("foo/bar"))
        expect_equal(
          as.character(file_info("foo/bar")$permissions),
          as.character(fs_perms("744"))
        )
        expect_equal(file_chmod("foo/bar", "rwx---r--"), fs_path("foo/bar"))
        expect_equal(
          as.character(file_info("foo/bar")$permissions),
          as.character(fs_perms("704"))
        )
        expect_equal(file_chmod("foo/bar", "rwx------"), fs_path("foo/bar"))
        expect_equal(
          as.character(file_info("foo/bar")$permissions),
          as.character(fs_perms("700"))
        )
      })

      it("errors if given an invalid mode", {
        expect_error(file_chmod("foo", "g+S"), "Invalid mode 'g\\+S'")
      })
      it("errors on missing input", {
        expect_error(file_chmod(NA, "u+x"), class = "invalid_argument")
      })

      it("is vectorized over files and permissions", {
        file_create("foo/baz")
        files <- c("foo/bar", "foo/baz")
        expect_equal(file_chmod(files, "000"), fs_path(files))
        expect_true(all(file_info(files)$permissions == c("000", "000")))

        expect_equal(file_chmod(files, "644"), fs_path(files))
        expect_true(all(file_info(files)$permissions == c("644", "644")))

        expect_equal(file_chmod(files, c("u+x", "o+x")), fs_path(files))
        expect_true(all(file_info(files)$permissions == c("744", "645")))

        expect_error(file_chmod(files, c("u+x", "o+x", "g+x")), class = "invalid_argument")
      })
    })
  })
}

describe("file_chown", {
  with_dir_tree(list("foo/bar" = "test"), {
    it("changes the ownership of a file, returns the input path", {
      skip("need elevated permissions to change uid")

      # Make everyone have write access, so we can delete this after changing ownership
      file_chmod("foo/bar", "a+w")

      # Change ownership to root
      expect_equal(file_chown("foo/bar", user_id = 0), "foo/bar")
      expect_true(file_info("foo/bar")$user == "root")
    })
    it("errors on missing input", {
      expect_error(file_chown(NA, user_id = 0), class = "invalid_argument")
    })
  })
})

describe("file_move", {
  it("renames files in the same directory", {
    with_dir_tree(list("foo" = "test"), {
      expect_true(file_exists("foo"))
      expect_equal(file_move("foo", "bar"), fs_path("bar"))
      expect_false(file_exists("foo"))
      expect_true(file_exists("bar"))
    })
  })

  it("renames files in different directories", {
    with_dir_tree(list("foo/bar" = "test", "foo2"), {
      expect_true(file_exists("foo/bar"))
      expect_true(file_exists("foo2"))
      expect_equal(file_move("foo/bar", "foo2"), fs_path("foo2/bar"))
      expect_false(file_exists("foo/bar"))
      expect_true(file_exists("foo2/bar"))

      expect_error(file_move("foo/bar", c("foo2", "foo3")), class = "fs_error")
    })
  })
  it("errors on missing input", {
    expect_error(file_move(NA, "foo2"), class = "invalid_argument")
    expect_error(file_move("foo", NA), class = "invalid_argument")
  })
})

describe("file_touch", {
  it("updates modification_time and access_time", {
    with_dir_tree("dir", {
      file_create("foo")
      now <- Sys.time()

      file_touch("foo", now)
      old <- file_info("foo")

      file_touch("foo", now + 10, now + 20)

      new <- file_info("foo")
      expect_true(old$modification_time < new$modification_time)
      expect_true(old$access_time < new$access_time)

      file_touch("foo", now - 10, now - 20)
      new2 <- file_info("foo")
      expect_true(old$modification_time > new2$modification_time)
      expect_true(old$access_time > new2$access_time)
    })
  })

  it("errors on missing input", {
    expect_error(file_touch(NA), class = "invalid_argument")
  })

  it("creates the file if it doesn't exist", {
    with_dir_tree("dir", {
      file_touch("foo")
      expect_true(file_exists("foo"))
    })
  })
})

#' Create, modify and view file permissions
#'
#' `fs_perms()` objects help one create and modify file permissions easily.
#' They support both numeric input, octal and symbolic character
#' representations. Compared to [octmode] they support symbolic representations
#' and display the mode the same format as `ls` on POSIX systems.
#'
#' @details
#' On POSIX systems the permissions are displayed as a 9 character string with
#' three sets of three characters. Each set corresponds to the permissions for
#' the user, the group and other (or default) users.
#'
#' If the first character of each set is a "r", the file is readable for those
#' users, if a "-", it is not readable.
#'
#' If the second character of each set is a "w", the file is writable for those
#' users, if a "-", it is not writable.
#'
#' The third character is more complex, and is the first of the following
#' characters which apply.
#' - 'S' If the character is part of the owner permissions and the file is not
#'   executable or the directory is not searchable by the owner, and the
#'   set-user-id bit is set.
#' - 'S' If the character is part of the group permissions and the file is not
#'   executable or the directory is not searchable by the group, and the
#'   set-group-id bit is set.
#' - 'T' If the character is part of the other permissions and the file is not
#'   executable or the directory is not searchable by others, and the 'sticky'
#'   (S_ISVTX) bit is set.
#' - 's' If the character is part of the owner permissions and the file is
#'   executable or the directory searchable by the owner, and the set-user-id bit
#'   is set.
#' - 's' If the character is part of the group permissions and the file is
#'   executable or the directory searchable by the group, and the set-group-id
#'   bit is set.
#' - 't' If the character is part of the other permissions and the file is
#'   executable or the directory searchable by others, and the ''sticky''
#'   (S_ISVTX) bit is set.
#' - 'x' The file is executable or the directory is searchable.
#' - '-' If none of the above apply.
#' Most commonly the third character is either 'x' or `-`.
#'
#' On Windows the permissions are displayed as a 3 character string where the
#' third character is only `-` or `x`.
#' @param x An object which is to be coerced to a fs_perms object. Can be an
#'   number or octal character representation, including symbolic
#'   representations.
#' @param ... Additional arguments passed to methods.
#' @examples
#' # Integer and numeric
#' fs_perms(420L)
#' fs_perms(c(511, 420))
#'
#' # Octal
#' fs_perms("777")
#' fs_perms(c("777", "644"))
#'
#' # Symbolic
#' fs_perms("a+rwx")
#' fs_perms(c("a+rwx", "u+rw,go+r"))
#'
#' # Use the `&` and `|`operators to check for certain permissions
#' (fs_perms("777") & "u+r") == "u+r"
#' @export
#' @name fs_perms
as_fs_perms <- function(x, ...) {
  UseMethod("as_fs_perms")
}

#' @export
as_fs_perms.NULL <- function(x, ...) {
  new_fs_perms(integer())
}

#' @export
#' @rdname fs_perms
fs_perms <- as_fs_perms

#' @export
print.fs_perms <- function(x, ...) {
  print(format(x, ...), quote = FALSE)

  invisible(x)
}

#' @export
format.fs_perms <- function(x, ...) {
  vapply(x, function(x) .Call(fs_strmode_, x), character(1))
}

#' @export
as.character.fs_perms <- format.fs_perms

#' @export
`[.fs_perms` <- function(x, i, ...) {
  new_fs_perms(NextMethod("["))
}

#' @export
`[[.fs_perms` <- function(x, i, ...) {
  new_fs_perms(NextMethod("[["))
}

#' @export
as_fs_perms.fs_perms <- function(x, ...) x

#' @export
as_fs_perms.character <- function(x, ..., mode = 0) {
  # matches inputs in rwxrwxrwx mode
  res <- x

  if (!is_windows()) {
    is_display_mode <- grepl("^[rwxXst-]{9}$", x)
    res[is_display_mode] <- display_mode_to_symbolic_mode_posix(res[is_display_mode])
  } else {
    is_display_mode <- grepl("^[rwxXst-]{3}$", x)
    res[is_display_mode] <- display_mode_to_symbolic_mode_windows(res[is_display_mode])
  }

  if (length(x) > 1 && length(mode) > 1 && length(x) != length(mode)) {

    msg <- sprintf(paste0(
        "`x` and `mode` must have compatible lengths:\n",
        "* `x` has length %i\n",
        "* `mode` has length %i\n"), length(x), length(mode))


    stop(fs_error(msg))
  }

  # Recycled over both res and mode
  n <- max(length(res), length(mode))
  out <- integer(n)
  for (i in seq_len(n)) {
    out[[i]] <- as.integer(
      .Call(fs_getmode_,
        res[[((i + 1) %% length(res)) + 1L]],
        as.integer(mode[[((i + 1) %% length(mode)) + 1L]])))
  }
  new_fs_perms(out)
}

display_mode_to_symbolic_mode_posix <- function(x) {
  paste0("u=", substring(x, 1, 3), ",g=", substring(x, 4,6), ",o=", substring(x, 7, 9))
}
display_mode_to_symbolic_mode_windows <- function(x) {
  paste0("u=", substring(x, 1, 3))
}

#' @export
as_fs_perms.octmode <- function(x, ...) {
  class(x) <- c("fs_perms", "integer")
  x
}

#' @export
as_fs_perms.numeric <- function(x, ...) {
  if (all(is.na(x) | x == as.integer(x))) {
    x <- as.integer(x)
    return(new_fs_perms(x))
  }
  stop("'x' cannot be coerced to class \"fs_perms\"", call. = FALSE)
}

#' @export
as_fs_perms.integer <- function(x, ...) {
  new_fs_perms(x)
}

new_fs_perms <- function(x) {
  assert("`x` must be an integer", is.integer(x))
  class(x) <- c("fs_perms", "integer")
  x
}
setOldClass(c("fs_perms", "integer"), integer())

#' @export
`!.fs_perms` <- function(a) {
  new_fs_perms(bitwNot(a))
}

#' @export
`&.fs_perms` <- function(a, b) {
  new_fs_perms(bitwAnd(a, as_fs_perms(b)))
}

#' @export
`|.fs_perms` <- function(a, b) {
  new_fs_perms(bitwOr(a, as_fs_perms(b)))
}

#' @export
`==.fs_perms` <- function(a, b) {
  b <- as_fs_perms(b)
  unclass(a & b) == unclass(b)
}

pillar_shaft.fs_perms <- function(x, ...) {
  pillar::new_pillar_shaft_simple(format.fs_perms(x), ...)
}

type_sum.fs_perms <- function(x) {
  "fs::perms"
}

# All functions below registered in .onLoad

vec_ptype2.fs_perms.fs_perms <- function(x, y, ...) {
  x
}
vec_ptype2.fs_perms.double <- function(x, y, ...) {
  x
}
vec_ptype2.double.fs_perms <- function(x, y, ...) {
  y
}

# Note order of class is the opposite as for ptype2
vec_cast.fs_perms.fs_perms <- function(x, to, ...) {
  x
}
vec_cast.fs_perms.double <- function(x, to, ...) {
  as_fs_perms(x)
}
vec_cast.double.fs_perms <- function(x, to, ...) {
  unclass(x)
}

vec_ptype2.fs_perms.integer <- function(x, y, ...) {
  x
}
vec_ptype2.integer.fs_perms <- function(x, y, ...) {
  y
}

vec_cast.fs_perms.integer <- function(x, to, ...) {
  as_fs_perms(x)
}
vec_cast.integer.fs_perms <- function(x, to, ...) {
  unclass(x)
}

vec_ptype2.fs_perms.character <- function(x, y, ...) {
  x
}
vec_ptype2.character.fs_perms <- function(x, y, ...) {
  y
}

vec_cast.fs_perms.character <- function(x, to, ...) {
  as_fs_perms(x)
}
vec_cast.character.fs_perms <- function(x, to, ...) {
  as.character(x)
}

if (!is_windows()) {
describe("as_fs_perms (POSIX)", {
  it("coerces integers", {
    expect_equal(as_fs_perms(420L), new_fs_perms(420L))
    expect_equal(as_fs_perms(NA_integer_), new_fs_perms(NA_integer_))

    expect_equal(as_fs_perms(c(511L, 420L)), new_fs_perms(c(511L, 420L)))
  })

  it("coerces doubles to integers", {
    expect_equal(as_fs_perms(420), new_fs_perms(420L))
    expect_equal(as_fs_perms(NA_real_), new_fs_perms(NA_integer_))
    expect_equal(as_fs_perms(c(511, 420)), new_fs_perms(c(511L, 420L)))

    expect_error(as_fs_perms(420.5), "'x' cannot be coerced to class \"fs_perms\"")
  })
  it("coerces octmode", {
    expect_equal(as_fs_perms(as.octmode("777")), new_fs_perms(511L))
    expect_equal(as_fs_perms(as.octmode(c("644", "777"))), new_fs_perms(c(420L, 511L)))
  })
  it("coerces characters in octal notation", {
    expect_equal(as_fs_perms("777"), new_fs_perms(511L))
    expect_equal(as_fs_perms(c("644", "777")), new_fs_perms(c(420L, 511L)))

    expect_error(as_fs_perms("777777"), "Invalid mode '777777'")
  })
  it("coerces characters in symbolic notation", {
    expect_equal(as_fs_perms("a+rwx"), new_fs_perms(511L))
    expect_equal(as_fs_perms(c("u+rw,go+r", "a+rwx")), new_fs_perms(c(420L, 511L)))
  })
  it("coerces characters in symbolic notation", {
    expect_equal(as_fs_perms("rw-r--r--"), new_fs_perms(420L))
    expect_equal(as_fs_perms(c("rw-r--r--", "rwxrwxrwx")), new_fs_perms(c(420L, 511L)))
  })
  it("preserves the class with both subset and subset2", {
    expect_s3_class(as_fs_perms("777")[1], "fs_perms")
    expect_s3_class(as_fs_perms("777")[[1]], "fs_perms")
  })
})

} else {
describe("as_fs_perms (Windows)", {
  skip_on_os(c("mac", "linux", "solaris"))
  it("coerces integers", {
    expect_equal(as_fs_perms(384L), new_fs_perms(384L))
    expect_equal(as_fs_perms(NA_integer_), new_fs_perms(NA_integer_))

    expect_equal(as_fs_perms(c(256L, 384L)), new_fs_perms(c(256L, 384L)))
  })

  it("coerces doubles to integers", {
    expect_equal(as_fs_perms(384), new_fs_perms(384L))
    expect_equal(as_fs_perms(NA_real_), new_fs_perms(NA_integer_))
    expect_equal(as_fs_perms(c(256, 384)), new_fs_perms(c(256L, 384L)))

    expect_error(as_fs_perms(420.5), "'x' cannot be coerced to class \"fs_perms\"")
  })
  it("coerces octmode", {
    expect_equal(as_fs_perms(as.octmode("600")), new_fs_perms(384L))
    expect_equal(as_fs_perms(as.octmode(c("600", "700"))), new_fs_perms(c(384L, 448L)))
    expect_equal(as_fs_perms(as.octmode(c("666", "777"))), new_fs_perms(c(438L, 511L)))
  })
  it("coerces characters in octal notation", {
    expect_equal(as_fs_perms("700"), new_fs_perms(448L))
    expect_equal(as_fs_perms(c("644", "777")), new_fs_perms(c(384L, 448L)))

    #expect_error(as_fs_perms("777777"), "Invalid mode '777777'")
  })
  it("ignores group and other octmode groups", {
    expect_equal(as_fs_perms(c("666", "777")), new_fs_perms(c(384L, 448L)))
  })
  it("coerces characters in symbolic notation", {
    expect_equal(as_fs_perms("a+rw"), new_fs_perms(384L))
    expect_equal(as_fs_perms(c("u+rw", "a+rwx")), new_fs_perms(c(384L, 448L)))
  })
  it("coerces characters in symbolic notation", {
    expect_equal(as_fs_perms("rw-"), new_fs_perms(384L))
    expect_equal(as_fs_perms(c("rw-", "rwx")), new_fs_perms(c(384L, 448L)))
  })
  it("returns 'rwx' on windows", {
    expect_equal(as.character(as_fs_perms("777")), "rwx")
  })
})
}

test_that("common type of double and fs_perms is fs_perms", {
  expect_identical(
    vctrs::vec_ptype2(double(), fs_perms()),
    fs_perms()[0]
  )
  expect_identical(
    vctrs::vec_ptype2(fs_perms(), double()),
    fs_perms()[0]
  )
})

test_that("fs_perms and double are coercible", {
  expect_identical(
    vctrs::vec_cast(1, fs_perms()),
    fs_perms(1)
  )
  expect_identical(
    vctrs::vec_cast(fs_perms(1), double()),
    1L
  )
  expect_identical(
    vctrs::vec_cast(fs_perms(1), fs_perms()),
    fs_perms(1)
  )
})

test_that("can concatenate fs_perms", {
  expect_identical(
    vctrs::vec_c(fs_perms(1), fs_perms(2)),
    as_fs_perms(c(1, 2))
  )
})

test_that("common type of integer and fs_perms is fs_perms", {
  expect_identical(
    vctrs::vec_ptype2(integer(), fs_perms()),
    fs_perms()[0]
  )
  expect_identical(
    vctrs::vec_ptype2(fs_perms(), integer()),
    fs_perms()[0]
  )
})

test_that("fs_perms and integer are coercible", {
  expect_identical(
    vctrs::vec_cast(1L, fs_perms()),
    fs_perms(1L)
  )
  expect_identical(
    vctrs::vec_cast(fs_perms(1L), integer()),
    1L
  )
})

test_that("common type of character and fs_perms is fs_perms", {
  expect_identical(
    vctrs::vec_ptype2(character(), fs_perms()),
    fs_perms()[0]
  )
  expect_identical(
    vctrs::vec_ptype2(fs_perms(), character()),
    fs_perms()[0]
  )
})

test_that("fs_perms and character are coercible", {
  expect_identical(
    vctrs::vec_cast("777", fs_perms()),
    fs_perms("777")
  )

  # This test only return rwx on Windows due to the less fine grained
  # permissions, so we skip it
  skip_on_os("windows")
  expect_identical(
    vctrs::vec_cast(fs_perms("777"), character()),
    "rwxrwxrwx"
  )
})

if (!is_windows()) {
  describe("as_fs_perms (POSIX)", {
    it("coerces integers", {
      expect_equal(as_fs_perms(420L), new_fs_perms(420L))
      expect_equal(as_fs_perms(NA_integer_), new_fs_perms(NA_integer_))

      expect_equal(as_fs_perms(c(511L, 420L)), new_fs_perms(c(511L, 420L)))
    })

    it("coerces doubles to integers", {
      expect_equal(as_fs_perms(420), new_fs_perms(420L))
      expect_equal(as_fs_perms(NA_real_), new_fs_perms(NA_integer_))
      expect_equal(as_fs_perms(c(511, 420)), new_fs_perms(c(511L, 420L)))
    })
    it("coerces octmode", {
      expect_equal(as_fs_perms(as.octmode("777")), new_fs_perms(511L))
      expect_equal(
        as_fs_perms(as.octmode(c("644", "777"))),
        new_fs_perms(c(420L, 511L))
      )
    })
    it("coerces characters in octal notation", {
      expect_equal(as_fs_perms("777"), new_fs_perms(511L))
      expect_equal(as_fs_perms(c("644", "777")), new_fs_perms(c(420L, 511L)))

      expect_snapshot(
        error = TRUE,
        as_fs_perms("777777"),
        transform = transform_error
      )
    })
    it("coerces characters in symbolic notation", {
      expect_equal(as_fs_perms("a+rwx"), new_fs_perms(511L))
      expect_equal(
        as_fs_perms(c("u+rw,go+r", "a+rwx")),
        new_fs_perms(c(420L, 511L))
      )
    })
    it("coerces characters in symbolic notation", {
      expect_equal(as_fs_perms("rw-r--r--"), new_fs_perms(420L))
      expect_equal(
        as_fs_perms(c("rw-r--r--", "rwxrwxrwx")),
        new_fs_perms(c(420L, 511L))
      )
    })
    it("preserves the class with both subset and subset2", {
      expect_s3_class(as_fs_perms("777")[1], "fs_perms")
      expect_s3_class(as_fs_perms("777")[[1]], "fs_perms")
    })
  })
} else {
  describe("as_fs_perms (Windows)", {
    skip_on_os(c("mac", "linux", "solaris"))
    it("coerces integers", {
      expect_equal(as_fs_perms(384L), new_fs_perms(384L))
      expect_equal(as_fs_perms(NA_integer_), new_fs_perms(NA_integer_))

      expect_equal(as_fs_perms(c(256L, 384L)), new_fs_perms(c(256L, 384L)))
    })

    it("coerces doubles to integers", {
      expect_equal(as_fs_perms(384), new_fs_perms(384L))
      expect_equal(as_fs_perms(NA_real_), new_fs_perms(NA_integer_))
      expect_equal(as_fs_perms(c(256, 384)), new_fs_perms(c(256L, 384L)))
    })
    it("coerces octmode", {
      expect_equal(as_fs_perms(as.octmode("600")), new_fs_perms(384L))
      expect_equal(
        as_fs_perms(as.octmode(c("600", "700"))),
        new_fs_perms(c(384L, 448L))
      )
      expect_equal(
        as_fs_perms(as.octmode(c("666", "777"))),
        new_fs_perms(c(438L, 511L))
      )
    })
    it("coerces characters in octal notation", {
      expect_equal(as_fs_perms("700"), new_fs_perms(448L))
      expect_equal(as_fs_perms(c("644", "777")), new_fs_perms(c(384L, 448L)))

      #expect_snapshot(error = TRUE, as_fs_perms("777777"))
    })
    it("ignores group and other octmode groups", {
      expect_equal(as_fs_perms(c("666", "777")), new_fs_perms(c(384L, 448L)))
    })
    it("coerces characters in symbolic notation", {
      expect_equal(as_fs_perms("a+rw"), new_fs_perms(384L))
      expect_equal(as_fs_perms(c("u+rw", "a+rwx")), new_fs_perms(c(384L, 448L)))
    })
    it("coerces characters in symbolic notation", {
      expect_equal(as_fs_perms("rw-"), new_fs_perms(384L))
      expect_equal(as_fs_perms(c("rw-", "rwx")), new_fs_perms(c(384L, 448L)))
    })
    it("returns 'rwx' on windows", {
      expect_equal(as.character(as_fs_perms("777")), "rwx")
    })
  })
}

test_that("common type of double and fs_perms is fs_perms", {
  expect_identical(
    vctrs::vec_ptype2(double(), fs_perms()),
    fs_perms()[0]
  )
  expect_identical(
    vctrs::vec_ptype2(fs_perms(), double()),
    fs_perms()[0]
  )
})

test_that("fs_perms and double are coercible", {
  expect_identical(
    vctrs::vec_cast(1, fs_perms()),
    fs_perms(1)
  )
  expect_identical(
    vctrs::vec_cast(fs_perms(1), double()),
    1L
  )
  expect_identical(
    vctrs::vec_cast(fs_perms(1), fs_perms()),
    fs_perms(1)
  )
})

test_that("can concatenate fs_perms", {
  expect_identical(
    vctrs::vec_c(fs_perms(1), fs_perms(2)),
    as_fs_perms(c(1, 2))
  )
})

test_that("common type of integer and fs_perms is fs_perms", {
  expect_identical(
    vctrs::vec_ptype2(integer(), fs_perms()),
    fs_perms()[0]
  )
  expect_identical(
    vctrs::vec_ptype2(fs_perms(), integer()),
    fs_perms()[0]
  )
})

test_that("fs_perms and integer are coercible", {
  expect_identical(
    vctrs::vec_cast(1L, fs_perms()),
    fs_perms(1L)
  )
  expect_identical(
    vctrs::vec_cast(fs_perms(1L), integer()),
    1L
  )
})

test_that("common type of character and fs_perms is fs_perms", {
  expect_identical(
    vctrs::vec_ptype2(character(), fs_perms()),
    fs_perms()[0]
  )
  expect_identical(
    vctrs::vec_ptype2(fs_perms(), character()),
    fs_perms()[0]
  )
})

test_that("fs_perms and character are coercible", {
  expect_identical(
    vctrs::vec_cast("777", fs_perms()),
    fs_perms("777")
  )

  # This test only return rwx on Windows due to the less fine grained
  # permissions, so we skip it
  skip_on_os("windows")
  expect_identical(
    vctrs::vec_cast(fs_perms("777"), character()),
    "rwxrwxrwx"
  )
})

#' @description
#' \ifelse{html}{
#' \out{
#' <a href="https://www.tidyverse.org/lifecycle/#maturing">
#'   <img src="https://img.shields.io/badge/lifecycle-maturing-blue.svg" alt="Maturing lifecycle">
#' </a>
#' }
#' }{
#' \url{https://www.tidyverse.org/lifecycle/#maturing}
#' }
#' @keywords internal
"_PACKAGE"

captures <- function(x, m) {
  assert("`x` must be a character", is.character(x))
  assert("`m` must be a match object from `regexpr()`",
    inherits(m, "integer") &&
    all(c("match.length", "capture.start", "capture.length", "capture.names") %in% names(attributes(m))))

  starts <- attr(m, "capture.start")
  strings <- substring(x, starts, starts + attr(m, "capture.length") - 1L)
  res <- data.frame(matrix(strings, ncol = NCOL(starts)), stringsAsFactors = FALSE)
  colnames(res) <- auto_name(attr(m, "capture.names"))
  res[is.na(m) | m == -1, ] <- NA_character_
  res
}

auto_name <- function(names) {
  missing <- names == ""
  if (all(!missing)) {
    return(names)
  }
  names[missing] <- seq_along(names)[missing]
  names
}

is_windows <- function() {
  # mock for tests
  if (isTRUE(Sys.getenv("FS_IS_WINDOWS", "FALSE") == "TRUE")) {
    return(TRUE)
  }

  tolower(Sys.info()[["sysname"]]) == "windows"
}

# This is needed to avoid checking the class of fs_path objects in the
# tests.
# @export
compare.fs_path <- function(x, y, ...) {
  if (identical(class(y), "character")) {
    class(x) <- NULL
  }
  names(x) <- NULL
  names(y) <- NULL
  NextMethod("compare")
}

# @export
compare.fs_perms <- function(x, y, ...) {
  if (!inherits(y, "fs_perms")) {
    y <- as.character(as_fs_perms(y))
    x <- as.character(x)
  }
  NextMethod("compare")
}

nchar <- function(x, type = "chars", allowNA = FALSE, keepNA = NA) {
  # keepNA was introduced in R 3.2.1, previous behavior was equivalent to keepNA
  # = FALSE
  if (getRversion() < "3.2.1") {
    if (!identical(keepNA, FALSE)) {
      stop("`keepNA` must be `FALSE` for R < 3.2.1", call. = FALSE)
    }
    return(base::nchar(x, type, allowNA))
  }
  base::nchar(x, type, allowNA, keepNA)
}

`%||%` <- function(x, y) if (is.null(x)) y else x

# Only use deterministic entries if we are building documentation in pkgdown.
pkgdown_tmp <- function(path) {
  if (identical(Sys.getenv("IN_PKGDOWN"), "true")) {
    file_temp_push(path)
  }
}

# This is adapted from glue::collapse
# https://github.com/tidyverse/glue/blob/cac874724d09d430036d1bdeba77982e953f29a2/R/glue.R#L140-L161
collapse <- function(x, sep = "", width = Inf, last = "") {
  if (length(x) == 0) {
    return(character())
  }
  if (any(is.na(x))) {
    return(NA_character_)
  }

  if (nzchar(last) && length(x) > 1) {
    res <- collapse(x[seq(1, length(x) - 1)], sep = sep, width = Inf)
    return(collapse(paste0(res, last, x[length(x)]), width = width))
  }
  x <- paste0(x, collapse = sep)
  if (width < Inf) {
    x_width <- nchar(x, "width", keepNA = FALSE)
    too_wide <- x_width > width
    if (too_wide) {
      x <- paste0(substr(x, 1, width - 3), "...")
    }
  }
  x
}

assert_no_missing <- function(x) {
  nme <- as.character(substitute(x))
  idx <- which(is.na(x))
  if (length(idx) > 0) {
    number <- prettyNum(length(idx), big.mark = ",")
    remaining_width <- getOption("width") - nchar(number, keepNA = FALSE) - 29
    indexes <- collapse(idx, width = remaining_width, sep = ", ", last = " and ")
    msg <- sprintf(
"`%s` must not have missing values
  * NAs found at %s locations: %s",
    nme,
    number,
    indexes)

    stop(fs_error(msg))
  }
}

assert <- function(msg, ..., class = "invalid_argument") {
  tests <- unlist(list(...))

  if (!all(tests)) {
    stop(fs_error(msg, class = class))
  }
}

fs_error <- function(msg, class = "invalid_argument") {
  structure(class = c(class, "fs_error", "error", "condition"), list(message = msg))
}

lengths <- function(x) {
  vapply(x, length, integer(1))
}

as_tibble <- function(x) {
  if (getOption("fs.use_tibble", TRUE) && is_installed("tibble")) {
    tibble::as_tibble(x)
  } else {
    x
  }
}

is_installed <- function(pkg) {
  isTRUE(requireNamespace(pkg, quietly = TRUE))
}

mkdirp <- function(x) {
  dir.create(x, showWarnings = FALSE, recursive = TRUE)
}

#' Read the value of a symbolic link
#'
#' @return A tidy path to the object the link points to.
#' @template fs
#' @export
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' file_create("foo")
#' link_create(path_abs("foo"), "bar")
#' link_path("bar")
#'
#' # Cleanup
#' file_delete(c("foo", "bar"))
#' \dontshow{setwd(.old_wd)}
link_path <- function(path) {
  assert_no_missing(path)

  old <- path_expand(path)

  path_tidy(.Call(fs_readlink_, old))
}

env <- new.env(parent = emptyenv())
env$temp_names <- character()

#' Create names for temporary files
#'
#' `file_temp()` returns the name which can be used as a temporary file.
#'
#' `file_temp_push()` can be used to supply deterministic entries in the
#' temporary file stack. This can be useful for reproducibility in like example
#' documentation and vignettes.
#'
#' `file_temp_pop()` can be used to explicitly remove an entry from the
#' internal stack, however generally this is done instead by calling
#' `file_temp()`.
#'
#' `path_temp()` constructs a path within the session temporary directory.
#' @param pattern A character vector with the non-random portion of the name.
#' @param tmp_dir The directory the file will be created in.
#' @param ext The file extension of the temporary file.
#' @param ... Additional paths appended to the temporary directory by `path()`.
#' @template fs
#' @export
#' @examples
#' \dontshow{file_temp_push("/tmp/filedd461c46df20")}
#'
#' path_temp()
#' path_temp("does-not-exist")
#'
#' file_temp()
#' file_temp(ext = "png")
#' file_temp("image", ext = "png")
#'
#'
#' # You can make the temp file paths deterministic
#' file_temp_push(letters)
#' file_temp()
#' file_temp()
#'
#' # Or explicitly remove values
#' while (!is.null(file_temp_pop())) next
#' file_temp_pop()
file_temp <- function(pattern = "file", tmp_dir = tempdir(), ext = "") {
  assert_no_missing(tmp_dir)

  has_extension <- nzchar(ext)

  # ensure the extension starts with _one_ .
  ext[has_extension] <- paste0(".", sub("^[.]", "", ext[has_extension]))

  path_tidy(file_temp_pop() %||% tempfile(pattern, tmp_dir, ext))
}

#' @export
#' @rdname file_temp
file_temp_push <- function(path) {
  assert_no_missing(path)

  old <- path_expand(path)

  env$temp_names <- c(env$temp_names, old)

  invisible(path_tidy(path))
}


#' @export
#' @rdname file_temp
file_temp_pop <- function() {
  n <- length(env$temp_names)
  if (n == 0) {
    return(NULL)
  }

  out <- env$temp_names[[1]]
  env$temp_names <- env$temp_names[-1]
  path_tidy(out)
}

#' @export
#' @rdname file_temp
path_temp <- function(...) {
  path(tempdir(), ...)
}

test_that("file_temp() returns unique temporary files", {
  prev <- file_temp()
  for(i in seq_len(10)) {
    expect_true(file_temp() != prev)
  }
})

test_that("file_temp() extension applied correctly", {
  tmp_no_ext <- file_temp()
  tmp_has_ext <- file_temp(ext = "pdf")
  expect_equal(path_ext(tmp_no_ext), "")
  expect_equal(path_ext(tmp_has_ext), "pdf")

  exts = c("xlsx", "doc", "")
  tmpfile_multiple <- tempfile(fileext = paste0(c(".", ".", ""), exts))
  file_temp_multiple <- file_temp(ext = exts)
  expect_equal(path_ext(file_temp_multiple), path_ext(tmpfile_multiple))

  expect_equal(
    path_ext(file_temp(ext = "pdf")),
    path_ext(file_temp(ext = ".pdf"))
  )
})

test_that("file_temp() can use have deterministic results if desired", {
  prev <- file_temp()
  file_temp_push(c("foo", "bar", "baz", "bar"))
  expect_equal(file_temp(), fs_path("foo"))
  expect_equal(file_temp(), fs_path("bar"))
  expect_equal(file_temp(), fs_path("baz"))

  file_temp_pop()
  x <- file_temp()
  expect_true(x != "bar" && x != prev)
})

test_that("file_temp() errors if given NA input", {
  expect_error(file_temp(tmp_dir = NA), class = "invalid_argument")
})

#' Lookup Users and Groups on a system
#'
#' These functions use the GETPWENT(3) and GETGRENT(3) system calls to query
#' users and groups respectively.
#'
#' @return They return their results in a `data.frame`. On windows both
#'   functions return an empty `data.frame` because windows does not have user
#'   or group ids.
#' @name id
#' @export
#' @examples
#' # list first 6 groups
#' head(group_ids())
#'
#' # list first 6 users
#' head(user_ids())
group_ids <- function() {
  res <- .Call(fs_groups_)
  res <- unique(res[order(res$group_id), ])
  row.names(res) <- NULL
  res
}


#' @rdname id
#' @export
user_ids <- function() {
  res <- .Call(fs_users_)
  res <- unique(res[order(res$user_id), ])
  row.names(res) <- NULL
  res
}

if (is_windows()) {
  describe("group_ids()", {
    it("returns the group ids", {
      g <- group_ids()
      expect_s3_class(g, "data.frame")
      expect_true(nrow(g) == 0)
      expect_equal(colnames(g),  c("group_id", "group_name"))
    })
  })
  describe("user_ids()", {
    it("returns the group ids", {
      u <- user_ids()
      expect_s3_class(u, "data.frame")
      expect_true(nrow(u) == 0)
      expect_equal(colnames(u),  c("user_id", "user_name"))
    })
  })
} else {
  describe("group_ids()", {
    it("returns the group ids", {
      g <- group_ids()
      expect_s3_class(g, "data.frame")
      expect_true(nrow(g) > 0)
      expect_equal(colnames(g),  c("group_id", "group_name"))
    })
  })
  describe("user_ids()", {
    it("returns the group ids", {
      u <- user_ids()
      expect_s3_class(u, "data.frame")
      expect_true(nrow(u) > 0)
      expect_equal(colnames(u),  c("user_id", "user_name"))
    })
  })
}

#' Create files, directories, or links
#'
#' The functions `file_create()` and `dir_create()` ensure that `path` exists;
#' if it already exists it will be left unchanged. That means that compared to
#' [file.create()], `file_create()` will not truncate an existing file, and
#' compared to [dir.create()], `dir_create()` will silently ignore existing
#' directories.
#'
#' @param path A character vector of one or more paths. For `link_create()`, 
#'   this is the target.
#' @param mode If file/directory is created, what mode should it have?
#'
#'   Links do not have mode; they inherit the mode of the file they link to.
#' @param recurse should intermediate directories be created if they do not
#'   exist?
#' @param recursive (Deprecated) If `TRUE` recurse fully.
#' @param ... Additional arguments passed to [path()]
#' @return The path to the created object (invisibly).
#' @name create
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' file_create("foo")
#' is_file("foo")
#' # dir_create applied to the same path will fail
#' try(dir_create("foo"))
#'
#' dir_create("bar")
#' is_dir("bar")
#' # file_create applied to the same path will fail
#' try(file_create("bar"))
#'
#' # Cleanup
#' file_delete("foo")
#' dir_delete("bar")
#' \dontshow{setwd(.old_wd)}
#' @export
file_create <- function(path, ..., mode = "u=rw,go=r") {
  assert_no_missing(path)
  assert("`mode` must be of length 1", length(mode) == 1)

  mode <- as_fs_perms(mode)
  new <- path_expand(path(path, ...))

  .Call(fs_create_, new, as.integer(mode))
  invisible(path_tidy(new))
}

#' @export
#' @rdname create
dir_create <- function(path, ..., mode = "u=rwx,go=rx", recurse = TRUE, recursive) {
  assert_no_missing(path)
  assert("`mode` must be of length 1", length(mode) == 1)

  if (!missing(recursive)) {
    recurse <- recursive
    warning("`recursive` is deprecated, please use `recurse` instead", immediate. = TRUE, call. = FALSE)
  }

  mode <- as_fs_perms(mode)
  new <- path_expand(path(path, ...))

  if (!isTRUE(recurse)) {
    .Call(fs_mkdir_, new, as.integer(mode))
    return(invisible(path_tidy(new)))
  }

  paths <- path_split(new)
  for (p in paths) {
    if (length(p) == 1) {
      .Call(fs_mkdir_, p, as.integer(mode))
    } else {
      p_paths <- Reduce(get("path", mode = "function"), p, accumulate = TRUE)
      if (is_absolute_path(p[[1]])) {
        p_paths <- p_paths[-1]
      }
      .Call(fs_mkdir_, p_paths, as.integer(mode))
    }
  }

  invisible(path_tidy(new))
}

#' @export
#' @rdname create
#' @param new_path The path where the link should be created.
#' @param symbolic Boolean value determining if the link should be a symbolic
#'   (the default) or hard link.
link_create <- function(path, new_path, symbolic = TRUE) {
  assert_no_missing(path)
  assert_no_missing(new_path)
  assert("Length of `path` must equal length of `new_path`", length(path) == length(new_path))

  old <- path_expand(path)
  new <- path_expand(new_path)

  if (isTRUE(symbolic)) {
    .Call(fs_link_create_symbolic_, old, new)
  } else {
    .Call(fs_link_create_hard_, old, new)
  }

  invisible(path_tidy(new_path))
}

describe("file_delete", {
  with_dir_tree(list("foo/bar" = "test"), {
    it("returns the input path and deletes the file", {
      expect_true(file_exists("foo/bar"))
      expect_equal(file_delete("foo/bar"), fs_path("foo/bar"))
      expect_false(file_exists("foo/bar"))
    })
    it("can delete directories and files", {
      dir_create("baz")

      expect_true(dir_exists("baz"))
      expect_equal(file_delete("baz"), fs_path("baz"))
      expect_false(dir_exists("baz"))

      dir_create("baz")
      file_create(c("baz/bar", "faz"))

      expect_true(dir_exists("baz"))
      expect_true(file_exists("faz"))

      expect_equal(file_delete(c("baz", "faz")), fs_path(c("baz", "faz")))

      expect_false(dir_exists("baz"))
      expect_false(file_exists("faz"))
    })
    it("errors on missing input", {
      expect_error(file_delete(NA), class = "invalid_argument")
    })
  })
})

describe("dir_delete", {
  it("deletes an empty directory and returns the path", {
    with_dir_tree(list("foo"), {
      expect_true(dir_exists("foo"))
      expect_equal(dir_delete("foo"), fs_path("foo"))
      expect_false(dir_exists("foo"))
    })
  })
  it("deletes a non-empty directory and returns the path", {
    with_dir_tree(
      list("foo/bar" = "test",
        "foo/baz" = "test2"), {
      expect_true(dir_exists("foo"))
      expect_equal(dir_delete("foo"), fs_path("foo"))
      expect_false(dir_exists("foo"))
    })
  })
  it("deletes nested directories and returns the path", {
    with_dir_tree(
      list("foo/bar/baz" = "test",
        "foo/baz/qux" = "test2"), {
      expect_true(dir_exists("foo"))
      expect_equal(dir_delete("foo"), fs_path("foo"))
      expect_false(dir_exists("foo"))
    })
  })
  it("deletes a non-empty directory with hidden files and directories and returns the path", {
    with_dir_tree(
      list("foo/bar" = "test",
        "foo/baz" = "test2",
        "foo/.blah" = "foo",
        ".dir"), {
      expect_true(dir_exists("foo"))
      expect_equal(dir_delete("foo"), fs_path("foo"))
      expect_false(dir_exists("foo"))
    })
  })
  it("errors on missing input", {
    expect_error(dir_delete(NA), class = "invalid_argument")
  })
})

describe("link_delete", {

  skip_on_os("windows")

  with_dir_tree(list("foo/bar" = "test"), {
    link_create("foo/bar", "loo")
    it("returns the input path and deletes the file", {
      expect_true(file_exists("loo"))
      expect_equal(link_delete("loo"), fs_path("loo"))
      expect_false(file_exists("loo"))
    })
    it("errors on missing input", {
      expect_error(link_delete(NA), class = "invalid_argument")
    })
    it("errors on non-link input", {
      expect_error(link_delete("foo"), class = "invalid_argument")
      expect_error(link_delete("foo/bar"), class = "invalid_argument")
    })
  })
})

describe("link_path", {
  with_dir_tree("foo", {
    link_create(path_abs("foo"), "loo")
    it("fails if given a non-link", {
      expect_error(link_path("foo"), "Failed to read link", class = "fs_error")
    })
    it("returns the link path if given a link", {
      expect_equal(link_path("loo"), path_abs("foo"))
    })
    it("fails silently if a link already exists and it points to the same location", {
      expect_error_free(link_create(path_abs("foo"), "loo"))
    })
    it("fails if a link already exists and it does not point to the same location", {
      dir_create("boo")
      expect_error(link_create(path_abs("boo"), "loo"), class = "EEXIST")
    })
  })
  it("errors on missing input", {
    expect_error(link_path(NA), class = "invalid_argument")
  })
})

test_that("file_create works with new and existing files", {
  x1 <- file_create(tempfile())

  expect_true(file_exists(x1))
  expect_error(file_create(x1), NA)
  expect_error(file_create(NA), class = "invalid_argument")

  unlink(x1)
})

test_that("dir_create works with new and existing files", {
  x1 <- dir_create(tempfile())

  expect_true(file_exists(x1))
  expect_error(dir_create(x1), NA)
  expect_error(dir_create(NA), class = "invalid_argument")

  unlink(x1, recursive = TRUE)
})

test_that("file_create works with multiple path arguments", {
  mkdirp(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  x1 <- file_create(tmp, "foo")

  expect_equal(x1, path(tmp, "foo"))
  expect_true(file_exists(x1))
})

test_that("dir_create works with multiple path arguments", {
  x1 <- dir_create(tempdir(), "foo", "bar")
  x2 <- dir_create(tempdir(), "foo", "baz", recurse = FALSE)

  expect_equal(x1, path(tempdir(), "foo", "bar"))
  expect_equal(x2, path(tempdir(), "foo", "baz"))
  expect_true(dir_exists(x1))
  expect_true(dir_exists(x2))

  unlink(x1)
  unlink(x2)
})

test_that("dir_create sets the mode properly", {
  skip_on_cran()
  skip_on_os("windows")
  x1 <- dir_create(tempfile(), mode = "775")

  expect_true(file_exists(x1))
  expect_equal(as.character(file_info(x1)$permissions), "rwxrwxr-x")

  unlink(x1, recursive = TRUE)
})

test_that("dir_create works with absolute paths and recurse = FALSE (#204)", {
  x1 <- dir_create(path_abs(tempfile()), recurse = FALSE)

  expect_true(file_exists(x1))

  unlink(x1, recursive = TRUE)
})

test_that("dir_create fails silently if the directory or link exists and fails if a file exists", {
  with_dir_tree(c("foo/bar" = ""), {
    expect_error_free(dir_create("foo"))
    expect_error(dir_create("foo/bar"), class = "EEXIST")
  })
})

test_that("dir_create fails with EACCES if it cannot create the directory", {
  skip_on_os("windows")
  skip_on_cran()
  if (Sys.info()[["effective_user"]] == "root") skip("root user")

  with_dir_tree("foo", {
    # Set current directly as read-only
    file_chmod(".", "u-w")
    expect_error(dir_create("bar"), class = "EACCES")
  })
})

test_that("link_create does not modify existing links", {

  # Windows does not currently support file based symlinks
  skip_on_os("windows")

  x <- dir_create(tempfile())

  dir_create(path(x, "dir1"))
  dir_create(path(x, "dir2"))

  link_create(path(x, "dir1"), path(x, "link"))
  expect_error(link_create(path(x, "dir2"), path(x, "link")), "file already exists", class = "EEXIST")
  expect_equal(link_path(path(x, "link")), path(x, "dir1"))

  expect_error(link_create(NA), class = "invalid_argument")
  expect_error(link_create(path(x, "dir2"), NA), class = "invalid_argument")
})

context("test-create.R")

test_that("file_create works with new and existing files", {
  x1 <- file_create(tempfile())

  expect_true(file_exists(x1))
  expect_error(file_create(x1), NA)
  expect_error(file_create(NA), class = "invalid_argument")
})

test_that("dir_create works with new and existing files", {
  x1 <- dir_create(tempfile())

  expect_true(file_exists(x1))
  expect_error(dir_create(x1), NA)
  expect_error(dir_create(NA), class = "invalid_argument")
})

test_that("dir_create fails silently if the directory or link exists and fails if a file exists", {
  with_dir_tree(c("foo/bar" = ""), {
    expect_error_free(dir_create("foo"))
    expect_error(dir_create("foo/bar"), class = "EEXIST")
  })
})

test_that("dir_create fails with EACCES if it cannot create the directory", {
  skip_on_os("windows")
  skip_on_cran()

  with_dir_tree("foo", {
    # Set current directly as read-only
    file_chmod(".", "u-w")
    expect_error(dir_create("bar"), class = "EACCES")
  })
})

test_that("link_create does not modify existing links", {

  # Windows does not currently support file based symlinks
  skip_on_os("windows")

  x <- dir_create(tempfile())

  dir_create(path(x, "dir1"))
  dir_create(path(x, "dir2"))

  link_create(path(x, "dir1"), path(x, "link"))
  expect_error(link_create(path(x, "dir2"), path(x, "link")), "file already exists", class = "EEXIST")
  expect_equal(link_path(path(x, "link")), path(x, "dir1"))

  expect_error(link_create(NA), class = "invalid_argument")
  expect_error(link_create(path(x, "dir2"), NA), class = "invalid_argument")
})

#' @useDynLib fs, .registration = TRUE
#' @importFrom methods setOldClass
NULL

# nocov start
.onUnload <- function(libpath) {
  .Call(fs_cleanup_)
  library.dynam.unload("fs", libpath)
}
# nocov end

#' Path computations
#'
#' All functions apart from `path_real()` are purely path computations, so the
#' files in question do not need to exist on the filesystem.
#' @template fs
#' @name path_math
#' @return The new path(s) in an `fs_path` object, which is a character vector
#'   that also has class `fs_path`. Except `path_split()`, which returns a list
#'   of character vectors of path components.
#' @seealso [path_expand()] for expansion of user's home directory.
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' dir_create("a")
#' file_create("a/b")
#' link_create(path_abs("a"), "c")
#'
#' # Realize the path
#' path_real("c/b")
#'
#' # Split a path
#' parts <- path_split("a/b")
#' parts
#'
#' # Join it together
#' path_join(parts)
#'
#' # Find the absolute path
#' path_abs("..")
#'
#' # Normalize a path
#' path_norm("a/../b\\c/.")
#'
#' # Compute a relative path
#' path_rel("/foo/abc", "/foo/bar/baz")
#'
#' # Find the common path between multiple paths
#' path_common(c("/foo/bar/baz", "/foo/bar/abc", "/foo/xyz/123"))
#'
#' # Cleanup
#' dir_delete("a")
#' link_delete("c")
#' \dontshow{setwd(.old_wd)}
NULL

#' Construct path to a file or directory
#'
#' `path()` constructs a relative path, `path_wd()` constructs an absolute path
#' from the current working directory.
#'
#' @param ... character vectors, if any values are NA, the result will also be
#'   NA. The paths follow the recycling rules used in the tibble package,
#'   namely that only length 1 arguments are recycled.
#' @param ext An optional extension to append to the generated path.
#' @export
#' @seealso [path_home()], [path_package()] for functions to construct paths
#'   relative to the home and package directories respectively.
#' @examples
#' path("foo", "bar", "baz", ext = "zip")
#'
#' path("foo", letters[1:3], ext = "txt")
path <- function(..., ext = "") {
  args <- list(...)
  assert_recyclable(args)

  path_tidy(.Call(fs_path_, lapply(args, function(x) enc2utf8(as.character(x))), ext))
}

assert_recyclable <- function(x) {
  if (length(x) == 0) {
    return()
  }
  len <- vapply(x, length, integer(1))
  max_len <- max(len)
  different <- which(len != 0 & len != max_len)
  assert(
    "Arguments must have consistent lengths, only values of length one are recycled.",
    all(len[different] == 1)
  )
}

#' @rdname path
#' @export
path_wd <- function(..., ext = "") {
  path(getwd(), ..., ext = ext)
}

#' @describeIn path_math returns the canonical path, eliminating any symbolic
#'   links and the special references `~`, `~user`, `.`, and `..`, , i.e. it
#'   calls `path_expand()` (literally) and `path_norm()` (effectively).
#' @export
path_real <- function(path) {
  path <- enc2utf8(as.character(path))
  old <- path_expand(path)

  is_missing <- is.na(path)
  old[!is_missing] <- .Call(fs_realize_, old[!is_missing])

  path_tidy(old)
}

#' Tidy paths
#'
#' untidy paths are all different, tidy paths are all the same.
#' Tidy paths always use `/` to delimit directories, never have
#' multiple `/` or trailing `/` and have colourised output based on the file
#' type.
#'
#' @return An `fs_path` object, which is a character vector that also has class
#'   `fs_path`
#' @template fs
#' @export
path_tidy <- function(path) {
  path <- enc2utf8(as.character(path))
  new_fs_path(.Call(fs_tidy_, path))
}


#' @describeIn path_math splits paths into parts.
#' @export
# TODO: examples
path_split <- function(path) {
  path <- path_tidy(path)

  # Split drive / UNC parts
  # split keep unc paths together, but keep root paths as first part.
  # //foo => '//foo' 'bar'
  # /foo/bar => '/' 'foo' 'bar'
  strsplit(path, "^(?=/)(?!//)|(?<!^)(?<!^/)/", perl = TRUE)
}

#' @describeIn path_math joins parts together. The inverse of [path_split()].
#' See [path()] to concatenate vectorized strings into a path.
#' @param parts A character vector or a list of character vectors, corresponding
#'   to split paths.
#' @export
path_join <- function(parts) {
  if (length(parts) == 0) {
    return(path_tidy(""))
  }
  if (!is.list(parts)) {
    return(path_tidy(.Call(fs_path_, as.list(enc2utf8(as.character(parts))), "")))
  }
  path_tidy(vapply(parts, path_join, character(1)))
}

#' @describeIn path_math returns a normalized, absolute version of a path.
#' @export
path_abs <- function(path, start = ".") {
  if (!is_absolute_path(start)) {
    start <- path_norm(path(getwd(), start))
  }
  is_abs <- is_absolute_path(path)
  path[is_abs] <- path_norm(path[is_abs])
  path[!is_abs] <- path_norm(path(start, path[!is_abs]))
  path_tidy(path)
}


#' @describeIn path_math eliminates `.` references and rationalizes up-level
#'   `..` references, so `A/./B` and `A/foo/../B` both become `A/B`, but `../B`
#'   is not changed. If one of the paths is a symbolic link, this may change the
#'   meaning of the path, so consider using `path_real()` instead.
#' @export
path_norm <- function(path) {
  non_missing <- !is.na(path)

  parts <- path_split(path)
  path_norm_one <- function(p) {
    p <- p[p != "."]
    double_dots <- p == ".."
    if (any(double_dots)) {
      res <- character(length(p))
      size <- 0
      is_abs <- is_absolute_path(p[[1]])
      for (i in seq_along(p)) {
        if (p[[i]] != ".." || (! is_abs && size == 0) || (size > 0 && res[[size]] == "..")) {
          res[[size <- size + 1]] <- p[[i]]
        } else if (size > 0) {
          size <- size - 1
        }
      }
      res <- res[seq(1, size)]
      if (is_abs && res[[1]] != p[[1]]) {
        res <- c(p[[1]], res)
      }
      p <- res
    }
    if (length(p) == 0) {
      return(path_tidy("."))
    }
    path_join(p)
  }
  parts[non_missing] <- vapply(parts[non_missing], path_norm_one, character(1))
  path_tidy(parts)
}

#' @describeIn path_math computes the path relative to the `start` path,
#'   which can be either an absolute or relative path.
#' @export
#' @param start A starting directory to compute the path relative to.
# This implementation is partially derived from
# https://github.com/python/cpython/blob/9c99fd163d5ca9bcc0b7ddd0d1e3b8717a63237c/Lib/posixpath.py#L446
path_rel <- function(path, start = ".") {
  start <- path_abs(path_expand(start))
  path <- path_abs(path_expand(path))

  path_rel_one <- function(p) {
    common <- path_common(c(start, p))
    starts <- path_split(start)[[1]]
    paths <- path_split(p)[[1]]

    i <- length(path_split(common)[[1]])
    double_dot_part <- rep("..", (length(starts) - i))
    if (i + 1 <= length(paths)) {
      path_part <- paths[seq(i + 1, length(paths))]
    } else {
      path_part <- character()
    }
    rels <- c(double_dot_part, path_part)
    if (length(rels) == 0) {
      return(path_tidy("."))
    }

    path_join(rels)
  }
  if (is.na(start)) {
    return(path_tidy(NA_character_))
  }

  is_missing <- is.na(path)

  path[!is_missing] <- vapply(path[!is_missing], path_rel_one, character(1))

  path_tidy(path)
}

#' Finding the User Home Directory
#'
#' * `path_expand()` performs tilde expansion on a path, replacing instances of
#' `~` or `~user` with the user's home directory.
#' * `path_home()` constructs a path within the expanded users home directory,
#' calling it with _no_ arguments can be useful to verify what fs considers the
#' home directory.
#' * `path_expand_r()` and `path_home_r()` are equivalents which always use R's
#' definition of the home directory.
#' @details
#' `path_expand()` differs from [base::path.expand()] in the interpretation of
#' the home directory of Windows. In particular `path_expand()` uses the path
#' set in the `USERPROFILE` environment variable and, if unset, then uses
#' `HOMEDRIVE`/`HOMEPATH`.
#'
#' In contrast [base::path.expand()] first checks for `R_USER` then `HOME`,
#' which in the default configuration of R on Windows are both set to the user's
#' document directory, e.g. `C:\\Users\\username\\Documents`.
#' [base::path.expand()] also does not support `~otheruser` syntax on Windows,
#' whereas `path_expand()` does support this syntax on all systems.
#'
#' This definition makes fs more consistent with the definition of home
#' directory used on Windows in other languages, such as
#' [python](https://docs.python.org/3/library/os.path.html#os.path.expanduser)
#' and [rust](https://doc.rust-lang.org/std/env/fn.home_dir.html#windows). This
#' is also more compatible with external tools such as git and ssh, both of
#' which put user-level files in `USERPROFILE` by default. It also allows you to
#' write portable paths, such as `~/Desktop` that points to the Desktop location
#' on Windows, macOS and (most) Linux systems.
#'
#' Users can set the `R_FS_HOME` environment variable to override the
#' definitions on any platform.
#' @seealso [R for Windows FAQ - 2.14](https://cran.r-project.org/bin/windows/base/rw-FAQ.html#What-are-HOME-and-working-directories_003f)
#' for behavior of [base::path.expand()].
#' @param ... Additional paths appended to the home directory by [path()].
#' @inheritParams path_math
#' @export
#' @examples
#' # Expand a path
#' path_expand("~/bin")
#'
#' # You can use `path_home()` without arguments to see what is being used as
#' # the home diretory.
#' path_home()
#' path_home("R")
#'
#' # This will likely differ from the above on Windows
#' path_home_r()
path_expand <- function(path) {
  path <- enc2utf8(path)

  # We use the windows implementation if R_FS_HOME is set or if on windows
  path_tidy(.Call(fs_expand_, path, Sys.getenv("R_FS_HOME") != "" || is_windows()))
}

#' @rdname path_expand
#' @export
path_expand_r <- function(path) {
  path <- enc2utf8(path)

  # Unconditionally use R_ExpandFileName
  path_tidy(.Call(fs_expand_, path, FALSE))
}

#' @rdname path_expand
#' @export
path_home <- function(...) {
  path(path_expand("~/"), ...)
}

#' @rdname path_expand
#' @export
path_home_r <- function(...) {
  path(path_expand_r("~/"), ...)
}

#' Manipulate file paths
#'
#' `path_file()` returns the filename portion of the path, `path_dir()` returns
#' the directory portion. `path_ext()` returns the last extension (if any) for a
#' path. `path_ext_remove()` removes the last extension and returns the rest of
#' the path. `path_ext_set()` replaces the extension with a new extension. If
#' there is no existing extension the new extension is appended.
#'
#' Note because these are not full file paths they return regular character
#' vectors, not `fs_path()` objects.
#' @template fs
#' @param ext,value The new file extension.
#' @seealso [base::basename()], [base::dirname()]
#' @export
#' @examples
#' path_file("dir/file.zip")
#'
#' path_dir("dir/file.zip")
#'
#' path_ext("dir/file.zip")
#'
#' path_ext("file.tar.gz")
#'
#' path_ext_remove("file.tar.gz")
#'
#' # Only one level of extension is removed
#' path_ext_set(path_ext_remove("file.tar.gz"), "zip")
#' @export
path_file <- function(path) {
  is_missing <- is.na(path)
  path[!is_missing] <- basename(path[!is_missing])
  as.character(path)
}

#' @rdname path_file
#' @export
path_dir <- function(path) {
  is_missing <- is.na(path)
  path[!is_missing] <- dirname(path[!is_missing])
  as.character(path_tidy(path))
}

#' @rdname path_file
#' @export
path_ext <- function(path) {
  if (length(path) == 0) {
    return(character())
  }

  res <- captures(path_file(path), regexpr("(?<!^|[.]|/)[.]([^.]+)$", path_file(path), perl = TRUE))[[1]]
  res[!is.na(path) & is.na(res)] <- ""
  res
}

#' @rdname path_file
#' @export
path_ext_remove <- function(path) {
  dir <- path_dir(path)
  file <- sub("(?<!^|[.]|/)[.][^.]+$", "", path_file(path), perl = TRUE)

  na <- is.na(path)
  no_dir <- dir == "." | dir == ""

  path[!na & no_dir] <- path_tidy(file[!na & no_dir])

  path[!na & !no_dir] <- path(dir[!na & !no_dir], file[!na & !no_dir])

  path
}

#' @rdname path_file
#' @export
path_ext_set <- function(path, ext) {

  if (!(length(ext) == length(path) || length(ext) == 1)) {
    assert_recyclable(list(path, ext))
  }

  # Remove a leading . if present
  ext <- sub("^[.]", "", ext)

  has_ext <- nzchar(ext)
  to_set <- !is.na(path) & has_ext

  if (length(ext) == 1) {
    ext <- rep(ext, sum(to_set))
  }

  path[to_set] <- paste0(
    path_ext_remove(path[to_set]), ".", ext
  )

  path_tidy(path)
}

#' @rdname path_file
#' @export
`path_ext<-` <- function(path, value) {
  path_ext_set(path, value)
}

#' @describeIn path_math finds the common parts of two (or more) paths.
#' @export
path_common <- function(path) {

  is_missing <- is.na(path)

  if (any(is_missing)) {
    return(path_tidy(NA))
  }

  is_abs <- is_absolute_path(path)

  # We must either have all absolute paths, or all relative paths.
  if (!(all(is_abs) || all(!is_abs))) {
    stop(fs_error("Can't mix absolute and relative paths"))
  }

  path <- path_norm(path)
  path <- sort(path)

  # remove . entries from the split paths
  parts <- lapply(path_split(path), function(x) x[x != "."])
  s1 <- parts[[1]]
  s2 <- parts[[length(parts)]]
  common <- s1
  for (i in seq_along(s1)) {
    if (s1[[i]] != s2[[i]]) {
      if (i == 1) {
        common <- ""
      } else {
        common <- s1[seq(1, i - 1)]
      }
      break;
    }
  }
  path_join(common)
}

#' Filter paths
#'
#' @template fs
#' @param glob A wildcard aka globbing pattern (e.g. `*.csv`) passed on to [grep()] to filter paths.
#' @param regexp A regular expression (e.g. `[.]csv$`) passed on to [grep()] to filter paths.
#' @param invert If `TRUE` return files which do _not_ match
#' @param ... Additional arguments passed to [grep].
#' @export
#' @examples
#' path_filter(c("foo", "boo", "bar"), glob = "*oo")
#' path_filter(c("foo", "boo", "bar"), glob = "*oo", invert = TRUE)
#'
#' path_filter(c("foo", "boo", "bar"), regexp = "b.r")
path_filter <- function(path, glob = NULL, regexp = NULL, invert = FALSE, ...) {
  if (!is.null(glob)) {
    if (!is.null(regexp)) {
      stop(fs_error("`glob` and `regexp` cannot both be set."))
    }
    regexp <- utils::glob2rx(glob)
  }
  if (!is.null(regexp)) {
    path <- grep(x = path, pattern = regexp, value = TRUE, invert = isTRUE(invert), ...)
  }
  setNames(path_tidy(path), path)
}

#' @describeIn path_math determine if a path has a given parent.
#' @param parent The parent path.
#' @export
path_has_parent <- function(path, parent) {
  path <- path_abs(path)
  parent <- path_abs(parent)

  res <- logical(length(path))

  assert_recyclable(list(path, parent))
  if (length(path) == 1) {
    path <- rep(path, length(parent))
  }

  if (length(parent) == 1) {
    parent <- rep(parent, length(path))
  }

  for (i in seq_along(path)) {
    res[[i]] <- identical(as.character(path_common(c(path[[i]], parent[[i]]))), as.character(parent[[i]]))
  }
  res
}

describe("path", {
  it("returns paths UTF-8 encoded", {
    skip_on_os("solaris")
    expect_equal(Encoding(path("föö")), "UTF-8")
  })

  it("returns paths UTF-8 encoded", {
    skip_on_os("solaris")
    expect_equal(Encoding(path("\U4F60\U597D.R")), "UTF-8")
  })

  it("returns empty strings for empty inputs", {
    expect_equal(path(""), fs_path(""))
    expect_equal(path(character()), fs_path(character()))
    expect_equal(path("foo", character(), "bar"), fs_path(character()))
  })

  it("propagates NA strings", {
    expect_equal(path(NA_character_), fs_path(NA_character_))
    expect_equal(path("foo", NA_character_), fs_path(NA_character_))
    expect_equal(
      path(c("foo", "bar"), c("baz", NA_character_)),
      fs_path(c("foo/baz", NA_character_))
    )
  })

  it("appends a ext if provided", {
    expect_equal(path("foo", ext = "bar"), fs_path("foo.bar"))
    expect_equal(
      path(c("foo", "baz"), ext = "bar"),
      fs_path(c("foo.bar", "baz.bar"))
    )
    expect_equal(
      path(c("foo", "baz", NA_character_), ext = "bar"),
      fs_path(c("foo.bar", "baz.bar", NA_character_))
    )
  })

  it("does not double paths", {
    expect_equal(path("/", "foo"), fs_path("/foo"))
    expect_equal(path("\\", "foo"), fs_path("/foo"))
    expect_equal(path("", "foo"), fs_path("/foo"))
    expect_equal(path("foo/", "bar"), fs_path("foo/bar"))
    expect_equal(path("foo\\", "bar"), fs_path("foo/bar"))
    expect_equal(path("foo//", "bar"), fs_path("foo/bar"))

    # This could be a UNC path, so we keep the doubled path.
    expect_equal(path("//", "foo"), fs_path("//foo"))
  })

  it("errors on paths which are too long", {
    expect_error(path(paste(rep("a", 100000), collapse = "")), "less than PATH_MAX")

    expect_error(do.call(path, as.list(rep("a", 100000))), "less than PATH_MAX")
  })

  it("follows recycling rules", {
    expect_equal(path("foo", character()), fs_path(character()))
    expect_equal(path("foo", "bar"), fs_path("foo/bar"))
    expect_equal(
      path("foo", c("bar", "baz")),
      fs_path(c("foo/bar", "foo/baz"))
    )
    expect_equal(
      path(c("foo", "qux"), c("bar", "baz")),
      fs_path(c("foo/bar", "qux/baz"))
    )

    expect_error(path(c("foo", "qux", "foo2"), c("bar", "baz")), "Arguments must have consistent lengths", class = "invalid_argument")
  })
})

describe("path_wd", {
  it("returns an absolute path from the working directory", {
    x <- path_wd()
    expect_true(is_absolute_path(x))

    expect_equal(x, path(getwd()))
  })
})

describe("path_real", {
  it("returns the real path for symbolic links", {
    with_dir_tree(list("foo/bar" = "test"), {
      link_create(path_real("foo"), "foo2")
      expect_equal(path_real("foo2"), path_real("foo"))
    })
  })

  it ("works with indirect symlinks", {
    skip_on_os("windows")

    with_dir_tree("foo", {
      wd <- path_wd()
      link_create(path("..", path_file(wd)), "self")
      link_create("self", "link")
      expect_equal(path_real("link"), wd)
    })
  })

  it ("works with parent symlinks", {
    skip_on_os("windows")

    # If there are symlinks in the parents of relative paths we need to resolve
    # them.
    # e.g. working directory is /usr/doc with 'doc' being a symlink to
    # /usr/share/doc. If we call `path_real("a")` we should return
    # /usr/share/doc/a.
    with_dir_tree("y", {
      wd <- path_wd()
      link_create("y", "k")
      withr::with_dir("k", {
        file_create("a")
        expect_equal(path_real("a"), path(wd, "/y/a"))
        unlink("a")
      })
    })
  })

  it ("resolves paths before normalizing", {

    skip_on_os("windows")

    # if we have the following hierarchy: a/k/y
    # and a symbolic link 'link-y' pointing to 'y' in directory 'a',
    # then `path_real("link-y/..")` should return 'k', not 'a'.
    with_dir_tree("a/k/y", {
      link_create("a/k/y", "link-y")

      expect_equal(path_real("link-y/.."), path_real("a/k"))
    })
  })

  it ("resolves paths before normalizing", {

    skip_on_os("windows")

    # if we have the following hierarchy: a/k/y
    # and a symbolic link 'link-y' pointing to 'y' in directory 'a',
    # then `path_real("link-y/..")` should return 'k', not 'a'.
    with_dir_tree("k", {
      wd <- path_wd()
      link_create(wd, "link")

      withr::with_dir(path_dir(wd), {
        base <- path_file(wd)
        expect_equal(path_real(path(base, "link")), wd)
        expect_equal(path_real(path(base, "link/k")), path(wd, "k"))
      })
    })
  })

  it("propagates NAs", {
    with_dir_tree(list("foo/bar" = "test"), {
      link_create(path_real("foo"), fs_path("foo2"))
      expect_equal(path_real(NA_character_), fs_path(NA_character_))
      expect_equal(
        path_real(c("foo2", NA_character_)),
        fs_path(c(path_real("foo"), NA_character_))
      )
    })
  })

  it("converts inputs to character if required", {
    with_dir_tree(list("foo/bar" = "test"), {
      link_create(path_real("foo"), "2")
      expect_equal(path_real(NA), fs_path(NA_character_)) ## NA (logical) coerced to NA_character_
      expect_equal(path_real(2), path_real("2"))
    })
  })

})

describe("path_split", {
  it("returns the path split", {
    expect_equal(path_split("foo/bar")[[1]], c("foo", "bar"))
    expect_equal(path_split(c("foo/bar", "foo/baz")), list(c("foo", "bar"), c("foo", "baz")))
  })

  it("does not split the root path", {
    expect_equal(path_split("/usr/bin")[[1]], c("/", "usr", "bin"))
    expect_equal(path_split("c:/usr/bin")[[1]], c("C:", "usr", "bin"))
    expect_equal(path_split("X:/usr/bin")[[1]], c("X:", "usr", "bin"))
    expect_equal(path_split("//server/usr/bin")[[1]], c("//server", "usr", "bin"))
    expect_equal(path_split("\\\\server\\usr\\bin")[[1]], c("//server", "usr", "bin"))
  })
})

describe("path_tidy", {
  it("does not expand ~", {
    expect_equal(path_tidy("~/foo"), fs_path("~/foo"))
    expect_equal(path_tidy("~/foo/"), fs_path("~/foo"))
    expect_equal(path_tidy("~//foo"), fs_path("~/foo"))
    expect_equal(path_tidy("~//foo"), fs_path("~/foo"))
    expect_equal(path_tidy("~\\foo\\"), fs_path("~/foo"))
  })

  it("always uses / for delimiting, never multiple / or trailing /", {
    expect_equal(path_tidy("foo/bar/baz"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("foo/bar/baz/"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("foo//bar//baz"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("foo//bar//baz//"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("foo\\bar\\baz"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("foo\\\\bar\\\\baz"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("//foo\\\\bar\\\\baz"), fs_path("//foo/bar/baz"))
    expect_equal(path_tidy("foo\\\\bar\\\\baz\\"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("foo\\\\bar\\\\baz\\\\"), fs_path("foo/bar/baz"))
  })

  it("always capitalizes and appends windows root paths with /", {
    expect_equal(path_tidy("C:"), fs_path("C:/"))
    expect_equal(path_tidy("c:"), fs_path("C:/"))
    expect_equal(path_tidy("X:"), fs_path("X:/"))
    expect_equal(path_tidy("x:"), fs_path("X:/"))
    expect_equal(path_tidy("c:/"), fs_path("C:/"))
    expect_equal(path_tidy("c://"), fs_path("C:/"))
    expect_equal(path_tidy("c:\\"), fs_path("C:/"))
    expect_equal(path_tidy("c:\\\\"), fs_path("C:/"))
  })

  it("passes NA along", {
    expect_equal(path_tidy(NA_character_), fs_path(NA_character_))
    expect_equal(
      path_tidy(c("foo/bar", NA_character_)),
      fs_path(c("foo/bar", NA_character_))
    )
  })

  it("works for non-UTF8 path", {
    x <- "folder//fa\xE7ile.txt/"
    Encoding(x) <- "latin1"
    out <- fs::path_tidy(x)
    expect_equal(Encoding(out), "UTF-8")
    expect_equal(out, fs_path("folder/façile.txt"))
  })

  it("converts inputs to character if required", {
      expect_equal(path_tidy(c("foo/bar", NA)), fs_path(c("foo/bar", NA_character_)))
      expect_equal(path_tidy(1), fs_path("1"))
  })
})

describe("path_temp", {
  it("returned tidies tempdir()", {
    expect_equal(path_temp(), path_tidy(tempdir()))
  })
})

describe("path_ext", {
  it ("returns 0 length outputs for 0 length inputs", {
    expect_equal(path_ext(character()), character())
  })

  it ("returns the path extension, or \"\" if one does not exist", {
    expect_equal(path_ext("foo.bar"), "bar")
    expect_equal(path_ext("foo.boo.bar"), "bar")
    expect_equal(path_ext("foo.boo.biff.bar"), "bar")
    expect_equal(path_ext(".csh.rc"), "rc")
    expect_equal(path_ext("nodots"), "")
    expect_equal(path_ext(".cshrc"), "")
    expect_equal(path_ext("...manydots"), "")
    expect_equal(path_ext("...manydots.ext"), "ext")
    expect_equal(path_ext("manydots..."), "")
    expect_equal(path_ext("manydots...ext"), "ext")
    expect_equal(path_ext("."), "")
    expect_equal(path_ext(".."), "")
    expect_equal(path_ext("........"), "")
    expect_equal(path_ext(""), "")
    expect_equal(path_ext(".bar"), "")
    expect_equal(path_ext("foo/.bar"), "")
    expect_equal(path_ext(c("foo.bar", NA_character_)), c("bar", NA_character_))
    expect_equal(path_ext("foo.bar/baz"), "")
  })
  it ("works with non-ASCII inputs", {
    skip_if_not_utf8()

    expect_equal(path_ext("f\U00F6\U00F6.txt"), "txt")
    expect_equal(path_ext("f\U00F6\U00F6.t\U00E4r"), "t\U00E4r")
  })
  it("returns a normal character vector", {
    expect_equal(class(path_dir("foo.sh")), "character")
  })
})

describe("path_ext_remove", {
  it ("removes the path extension", {
    expect_equal(path_ext_remove("foo.bar"), "foo")
    expect_equal(path_ext_remove("foo.boo.bar"), "foo.boo")
    expect_equal(path_ext_remove("foo.boo.biff.bar"), "foo.boo.biff")
    expect_equal(path_ext_remove(".csh.rc"), ".csh")
    expect_equal(path_ext_remove("nodots"), "nodots")
    expect_equal(path_ext_remove(".cshrc"), ".cshrc")
    expect_equal(path_ext_remove("...manydots"), "...manydots")
    expect_equal(path_ext_remove("...manydots.ext"), "...manydots")
    expect_equal(path_ext_remove("manydots..."), "manydots...")
    expect_equal(path_ext_remove("manydots...ext"), "manydots")
    expect_equal(path_ext_remove("."), ".")
    expect_equal(path_ext_remove(".."), "..")
    expect_equal(path_ext_remove("........"), "........")
    expect_equal(path_ext_remove(""), "")
    expect_equal(path_ext_remove(NA_character_), NA_character_)
    expect_equal(path_ext_remove(c("foo.bar", NA_character_)), c("foo", NA_character_))
    expect_equal(path_ext_remove(".bar"), ".bar")
    expect_equal(path_ext_remove("foo/.bar"), "foo/.bar")
    expect_equal(path_ext_remove("foo.bar/abc.123"), "foo.bar/abc")
    expect_equal(path_ext_remove("foo.bar/abc"), "foo.bar/abc")
  })
  it ("works with non-ASCII inputs", {
    skip_if_not_utf8()

    expect_equal(path_ext_remove("f\U00F6\U00F6.txt"), "f\U00F6\U00F6")
    expect_equal(path_ext_remove("f\U00F6\U00F6.t\U00E4r"), "f\U00F6\U00F6")
  })
})

describe("path_ext_set", {
  it ("replaces the path extension", {
    expect_equal(path_ext_set("foo.bar", "baz"), fs_path("foo.baz"))
    expect_equal(path_ext_set("foo.boo.bar", "baz"), fs_path("foo.boo.baz"))
    expect_equal(
      path_ext_set("foo.boo.biff.bar", "baz"),
      fs_path("foo.boo.biff.baz")
    )
    expect_equal(path_ext_set(".csh.rc", "gz"), fs_path(".csh.gz"))
    expect_equal(path_ext_set("nodots", "bar"), fs_path("nodots.bar"))
    expect_equal(path_ext_set(".cshrc", "bar"), fs_path(".cshrc.bar"))
    expect_equal(
      path_ext_set("...manydots", "bar"),
      fs_path("...manydots.bar")
    )
    expect_equal(
      path_ext_set("...manydots.ext", "bar"),
      fs_path("...manydots.bar")
    )
    expect_equal(path_ext_set(".", "bar"), fs_path("..bar"))
    expect_equal(path_ext_set("..", "bar"), fs_path("...bar"))
    expect_equal(path_ext_set("........", "bar"), fs_path(".........bar"))
    expect_equal(path_ext_set("", "bar"), fs_path(".bar"))
    expect_equal(path_ext_set(NA_character_, "bar"), fs_path(NA_character_))
    expect_equal(
      path_ext_set(c("foo", NA_character_), "bar"),
      fs_path(c("foo.bar", NA_character_))
    )
    expect_equal(path_ext_set(".bar", "baz"), fs_path(".bar.baz"))
    expect_equal(path_ext_set("foo/.bar", "baz"), fs_path("foo/.bar.baz"))
    expect_equal(path_ext_set("foo", ""), fs_path("foo"))
  })
  it ("works the same with and without a leading . for ext", {
    expect_equal(path_ext_set("foo", "bar"), fs_path("foo.bar"))
    expect_equal(path_ext_set("foo", ".bar"), fs_path("foo.bar"))
  })
  it ("only removes a leading . from the extension", {
    expect_equal(path_ext_set("foo", "b.ar"), fs_path("foo.b.ar"))
    expect_equal(path_ext_set("foo", ".b.ar"), fs_path("foo.b.ar"))
  })
  it ("works with multiple paths (#205)", {
    multiple_paths <- c("a", "b")
    expect_equal(
      path_ext_set(multiple_paths, "csv"),
      fs_path(c("a.csv", "b.csv"))
    )
  })
  it ("works with multiple extensions (#250)", {
    multiple_paths <- c("a", "b")
    multiple_exts <- c("csv", "tsv")
    expect_equal(
      path_ext_set(multiple_paths, multiple_exts),
      fs_path(c("a.csv", "b.tsv"))
    )

    expect_error(path_ext_set(multiple_paths, c(multiple_exts, "xls")), class = "fs_error", "consistent lengths")
  })
  it ("works with non-ASCII inputs", {
    skip_if_not_utf8()

    expect_equal(
      path_ext_set("f\U00F6\U00F6.txt", "bar"),
      fs_path("f\U00F6\U00F6.bar")
    )
    expect_equal(
      path_ext_set("f\U00F6\U00F6.t\U00E4r", "bar"),
      fs_path("f\U00F6\U00F6.bar")
    )
  })
})

describe("path_ext<-", {
  it ("replaces the path extension", {
    x <- "...manydots"
    path_ext(x) <- "bar"
    expect_equal(x, fs_path("...manydots.bar"))
  })
})

# test cases derived from https://github.com/python/cpython/blob/6f0eb93183519024cb360162bdd81b9faec97ba6/Lib/test/test_posixpath.py#L276

describe("path_norm", {
  it ("works with POSIX paths", {
    expect_equal(path_norm(""), fs_path("."))
    expect_equal(path_norm(""), fs_path("."))
    expect_equal(path_norm(".."), fs_path(".."))
    expect_equal(path_norm("/../../.."), fs_path("/"))
    expect_equal(path_norm("/../.././.."), fs_path("/"))
    expect_equal(path_norm("../.././.."), fs_path("../../.."))
    expect_equal(path_norm("/"), fs_path("/"))
    expect_equal(path_norm("//"), fs_path("/"))
    expect_equal(path_norm("///"), fs_path("/"))
    expect_equal(path_norm("//foo"), fs_path("//foo"))
    expect_equal(path_norm("//foo/.//bar//"), fs_path("//foo/bar"))
    expect_equal(path_norm("/foo/.//bar//.//..//.//baz"), fs_path("/foo/baz"))
    expect_equal(path_norm("/..//./foo/.//bar"), fs_path("/foo/bar"))
  })

  it ("works with POSIX paths", {
    expect_equal(path_norm("A//////././//.//B"), fs_path("A/B"))
    expect_equal(path_norm("A/./B"), fs_path("A/B"))
    expect_equal(path_norm("A/foo/../B"), fs_path("A/B"))
    expect_equal(path_norm("C:A//B"), fs_path("C:A/B"))
    expect_equal(path_norm("D:A/./B"), fs_path("D:A/B"))
    expect_equal(path_norm("e:A/foo/../B"), fs_path("E:A/B"))

    expect_equal(path_norm("C:///A//B"), fs_path("C:/A/B"))
    expect_equal(path_norm("D:///A/./B"), fs_path("D:/A/B"))
    expect_equal(path_norm("e:///A/foo/../B"), fs_path("E:/A/B"))

    expect_equal(path_norm(".."), fs_path(".."))
    expect_equal(path_norm("."), fs_path("."))
    expect_equal(path_norm(""), fs_path("."))
    expect_equal(path_norm("/"), fs_path("/"))
    expect_equal(path_norm("c:/"), fs_path("C:/"))
    expect_equal(path_norm("/../.././.."), fs_path("/"))
    expect_equal(path_norm("c:/../../.."), fs_path("C:/"))
    expect_equal(path_norm("../.././.."), fs_path("../../.."))
    expect_equal(path_norm("C:////a/b"), fs_path("C:/a/b"))
    expect_equal(
      path_norm("//machine/share//a/b"),
      fs_path("//machine/share/a/b")
    )

    expect_equal(path_norm("\\\\?\\D:/XY\\Z"), fs_path("//?/D:/XY/Z"))
  })

  it ("works with missing values", {
    expect_equal(path_norm(NA), fs_path(NA_character_))
    expect_equal(path_norm(c("foo", NA)), fs_path(c("foo", NA)))
    expect_equal(
      path_norm(c(NA, NA)),
      fs_path(c(NA_character_, NA_character_))
    )  })
})

# Test cases derived from https://github.com/python/cpython/blob/6f0eb93183519024cb360162bdd81b9faec97ba6/Lib/test/test_posixpath.py

describe("path_common", {
  it ("finds the common path", {
    expect_error(path_common(c("/usr", "usr")), "Can't mix", class = "fs_error")
    expect_error(path_common(c("usr", "/usr")), "Can't mix", class = "fs_error")

    expect_equal(path_common(c("/usr/local")), fs_path("/usr/local"))
    expect_equal(
      path_common(c("/usr/local", "/usr/local")),
      fs_path("/usr/local")
    )
    expect_equal(
      path_common(c("/usr/local/", "/usr/local")),
      fs_path("/usr/local")
    )
    expect_equal(
      path_common(c("/usr/local/", "/usr/local/")),
      fs_path("/usr/local")
    )
    expect_equal(
      path_common(c("/usr//local/bin", "/usr/local//bin")),
      fs_path("/usr/local/bin")
    )
    expect_equal(
      path_common(c("/usr/./local", "/./usr/local")),
      fs_path("/usr/local")
    )
    expect_equal(
      path_common(c("/", "/dev")),
      fs_path("/")
    )
    expect_equal(
      path_common(c("/usr", "/dev")),
      fs_path("/")
    )
    expect_equal(
      path_common(c("/usr/lib/", "/usr/lib/python3")),
      fs_path("/usr/lib")
    )
    expect_equal(
      path_common(c("/usr/lib/", "/usr/lib64/")),
      fs_path("/usr")
    )

    expect_equal(
      path_common(c("/usr/lib", "/usr/lib64")),
      fs_path("/usr")
    )
    expect_equal(
      path_common(c("/usr/lib/", "/usr/lib64")),
      fs_path("/usr")
    )

    expect_equal(
      path_common(c("spam")),
      fs_path("spam")
    )
    expect_equal(
      path_common(c("spam", "spam")),
      fs_path("spam")
    )
    expect_equal(
      path_common(c("spam", "alot")),
      fs_path("")
    )
    expect_equal(
      path_common(c("and/jam", "and/spam")),
      fs_path("and")
    )
    expect_equal(
      path_common(c("and//jam", "and/spam//")),
      fs_path("and")
    )
    expect_equal(
      path_common(c("and/./jam", "./and/spam")),
      fs_path("and")
    )
    expect_equal(path_common(c("and/jam", "and/spam", "alot")), fs_path(""))
    expect_equal(path_common(c("and/jam", "and/spam", "and")), fs_path("and"))

    expect_equal(path_common(c("")), fs_path(""))
    expect_equal(path_common(c("", "spam/alot")), fs_path(""))

    expect_error(path_common(c("", "/spam/alot")), "Can't mix", class = "fs_error")
  })

  it("returns NA if any input is NA", {
    expect_equal(path_common(NA), fs_path(NA_character_))
    expect_equal(path_common(c("and/jam", NA)), fs_path(NA_character_))
    expect_equal(path_common(c("and/jam", NA, "and")), fs_path(NA_character_))
  })
})

describe("path_has_parent", {
  it("works on single paths", {
    expect_false(path_has_parent("foo", "bar"))
    expect_false(path_has_parent("foo", "foo/bar"))

    expect_false(path_has_parent("/usr/var2/log", "/usr/var"))

    expect_true(path_has_parent("foo/bar", "foo"))
    expect_true(path_has_parent("path/myfiles/myfile", "path/to/files/../../myfiles"))
  })
  it("works with multiple paths", {
    expect_equal(path_has_parent(c("/a/b/c", "x/y"), "/a/b"), c(TRUE, FALSE))

    expect_equal(path_has_parent("/a/b/c", c("/a/b", "/x/y")), c(TRUE, FALSE))

    expect_error(path_has_parent(c("/a/b/c", "x/y"), c("/a/b", "x/y", "foo/bar")), "consistent lengths", class = "invalid_argument")
  })
})

describe("path_join", {
    it("converts inputs to character if required", {
        expect_equal(path_join(c("a", NA)), path_join(c("a", NA_character_)))
        expect_equal(path_join(c("a", 1)), path_join(c("a", "1")))
        expect_equal(path_join(list(c("a", 1), c("a/b", 2))), fs_path(c(path_join(c("a", "1")), path_join(c("a/b", "2")))))
    })
    it("works with list inputs", {
        expect_equal(path_join(list(c("foo", "bar"), c("a/b", "c"))), fs_path(c(path_join(c("foo", "bar")), path_join(c("a/b", "c")))))
        expect_equal(path_join(list(NA, c("a", 1), c("a/b", 2))), fs_path(c(path_join(NA_character_), path_join(c("a", "1")), path_join(c("a/b", "2")))))
    })
})

# derived from https://github.com/python/cpython/blob/6f0eb93183519024cb360162bdd81b9faec97ba6/Lib/test/test_posixpath.py#L483
describe("path_rel", {
  it("works for POSIX paths", {
    cur_dir <- path_file(getwd())
    expect_equal(path_rel("a"), fs_path("a"))
    expect_equal(path_rel(path_abs("a")), fs_path("a"))
    expect_equal(path_rel("a/b"), fs_path("a/b"))
    expect_equal(path_rel("../a/b"), fs_path("../a/b"))
    expect_equal(path_rel("a", "../b"), path_join(c("..", cur_dir, "a")))
    expect_equal(path_rel("a/b", "../c"), path_join(c("..", cur_dir, "a", "b")))
    expect_equal(path_rel("a", "b/c"), fs_path("../../a"))
    expect_equal(path_rel("a", "a"), fs_path("."))
    expect_equal(
      path_rel("/foo/bar/bat", "/x/y/z"),
      fs_path("../../../foo/bar/bat")
    )
    expect_equal(path_rel("/foo/bar/bat", "/foo/bar"), fs_path("bat"))
    expect_equal(path_rel("/foo/bar/bat", "/"), fs_path("foo/bar/bat"))
    expect_equal(path_rel("/", "/foo/bar/bat"), fs_path("../../.."))
    expect_equal(path_rel("/foo/bar/bat", "/x"), fs_path("../foo/bar/bat"))
    expect_equal(path_rel("/x", "/foo/bar/bat"), fs_path("../../../x"))
    expect_equal(path_rel("/", "/"), fs_path("."))
    expect_equal(path_rel("/a", "/a"), fs_path("."))
    expect_equal(path_rel("/a/b", "/a/b"), fs_path("."))

    expect_equal(
      path_rel(c("a", "a/b", "a/b/c"), "a/b"),
      fs_path(c("..", ".", "c"))
    )
    expect_error(
      path_rel(c("a", "a/b", "a/b/c"), c("a/b", "a")),
      "`start` must be a single path to a starting directory",
      fixed = TRUE
    )
  })

  it("works for windows paths", {
    expect_equal(
      path_rel("c:/foo/bar/bat", "c:/x/y"),
      fs_path("../../foo/bar/bat")
    )
    expect_equal(
      path_rel("//conky/mountpoint/a", "//conky/mountpoint/b/c"),
      fs_path("../../a")
    )
    expect_equal(
      path_rel("d:/Users/a", "D:/Users/b"),
      fs_path("../a")
    )
  })

  it("expands path before computing relativity", {
    expect_equal(path_rel("/foo/bar/baz", "~"), path_rel("/foo/bar/baz", path_expand("~")))
    expect_equal(path_rel("~/foo/bar/baz", "~"), fs_path("foo/bar/baz"))
  })

  it("propagates NAs", {
    expect_equal(path_rel(NA_character_), fs_path(NA_character_))
    expect_equal(path_rel("/foo/bar/baz", NA_character_), fs_path(NA_character_))
  })

  it("can be reversed by path_abs", {
    skip_on_os("windows")

    f <- file_temp()
    expect_equal(path_abs(path_rel(f)), f)

    home <- path_home()
    home_f <- path_abs(path_home("../../foo/bar/baz"))

    expect_equal(path_abs(path_rel(home_f, start = home), start = home), home_f)
  })
})

describe("path_abs", {
  it("uses the start parameter", {
    expect_equal(path_abs("b", "a"), path_abs("a/b"))
  })
})

describe("path_home", {
  # The trailing slash is needed to ensure we get path expansion
  # on POSIX systems when readline support is not built in. (#60)
  it("is equivalent to path_expand(\"~/\")", {
    expect_equal(path_home(), path_tidy(path_expand("~/")))
  })
})

describe("path_dir", {
  it("works like dirname for normal paths", {
    expect_equal(path_dir("foo/bar"), "foo")
    expect_equal(path_dir("bar"), ".")
    expect_equal(path_dir(c("foo/bar", "baz")), c("foo", "."))
  })
  it("propagates NAs", {
    expect_equal(path_dir(NA_character_), NA_character_)
    expect_equal(path_dir(c("foo/bar", NA)), c("foo", NA_character_))
  })
  it("returns a normal character vector", {
    expect_equal(class(path_dir("foo/bar")), "character")
  })
})

describe("path_file", {
  it("works like dirname for normal paths", {
    expect_equal(path_file("foo/bar"), "bar")
    expect_equal(path_file("bar"), "bar")
    expect_equal(path_file(c("foo/bar", "baz")), c("bar", "baz"))
  })
  it("propagates NAs", {
    expect_equal(path_file(NA_character_), NA_character_)
    expect_equal(path_file(c("foo/bar", NA)), c("bar", NA_character_))
  })
  it("returns a normal character vector", {
    expect_equal(class(path_dir("foo/bar")), "character")
  })
})

# These tests were adapted from
# https://github.com/python/cpython/blob/48e8c82fc63d2ddcddce8aa637a892839b551619/Lib/test/test_ntpath.py,
# hence the flying circus names, but I like dead parrots too, so keeping them.
describe("path_expand", {
  it("works on windows", {
    withr::local_envvar(c(FS_IS_WINDOWS = "TRUE"))
    withr::with_envvar(c("USERPROFILE" = NA, "HOMEDRIVE" = NA, "HOMEPATH" = NA), {
      expect_equal(path_expand("~test"), fs_path("~test"))
    })
    withr::with_envvar(c("USERPROFILE" = NA, "HOMEDRIVE" = "C:\\", "HOMEPATH" = "eric\\idle"), {
      expect_equal(path_expand("~"), fs_path("C:/eric/idle"))
      expect_equal(path_expand("~test"), fs_path("C:/eric/test"))
    })
    withr::with_envvar(c("USERPROFILE" = NA, "HOMEDRIVE" = NA, "HOMEPATH" = "eric/idle"), {
      expect_equal(path_expand("~"), fs_path("eric/idle"))
      expect_equal(path_expand("~test"), fs_path("eric/test"))
    })
    withr::with_envvar(c("USERPROFILE" = "C:\\idle\\eric"), {
      expect_equal(path_expand("~"), fs_path("C:/idle/eric"))
      expect_equal(path_expand("~test"), fs_path("C:/idle/test"))


      expect_equal(
        path_expand("~test/foo/bar"),
        fs_path("C:/idle/test/foo/bar")
      )
      expect_equal(
        path_expand("~test/foo/bar/"),
        fs_path("C:/idle/test/foo/bar")
      )
      expect_equal(
        path_expand("~test\\foo\\bar"),
        fs_path("C:/idle/test/foo/bar")
      )
      expect_equal(
        path_expand("~test\\foo\\bar\\"),
        fs_path("C:/idle/test/foo/bar")
      )
      expect_equal(
        path_expand("~/foo/bar"),
        fs_path("C:/idle/eric/foo/bar")
      )
      expect_equal(
        path_expand("~\\foo\\bar"),
        fs_path("C:/idle/eric/foo/bar")
      )
    })
    withr::with_envvar(c("USERPROFILE" = "C:\\idle\\eric", "R_FS_HOME" = "C:\\john\\cleese"), {
      # R_FS_HOME overrides userprofile
      expect_equal(path_expand("~"), fs_path("C:/john/cleese"))
      expect_equal(path_expand("~test"), fs_path("C:/john/test"))
    })
  })
  it("repects R_FS_HOME", {
    withr::with_envvar(c("R_FS_HOME" = "/foo/bar"), {
      expect_equal(path_expand("~"), fs_path("/foo/bar"))
      expect_equal(path_expand("~test"), fs_path("/foo/test"))
    })
  })
})

test_that("valid names", {
  expect_equal(path_sanitize("the quick brown fox jumped over the lazy dog.mp3"), "the quick brown fox jumped over the lazy dog.mp3")
  expect_equal(path_sanitize("résumé"), "résumé")
})

test_that("valid names", {
  expect_equal(path_sanitize("the quick brown fox jumped over the lazy dog.mp3", "_"), "the quick brown fox jumped over the lazy dog.mp3")
  expect_equal(path_sanitize("résumé", "_"), "résumé")
})

test_that("control characters", {
  expect_equal(path_sanitize("hello\nworld"), "helloworld")
})

test_that("control characters", {
  expect_equal(path_sanitize("hello\nworld", "_"), "hello_world")
})

test_that("restricted codes", {
  expect_equal(path_sanitize("h?w"), "hw")
  expect_equal(path_sanitize("h/w"), "hw")
  expect_equal(path_sanitize("h*w"), "hw")
})

test_that("restricted codes", {
  expect_equal(path_sanitize("h?w", "_"), "h_w")
  expect_equal(path_sanitize("h/w", "_"), "h_w")
  expect_equal(path_sanitize("h*w", "_"), "h_w")
})

# https://msdn.microsoft.com/en-us/library/aa365247(v=vs.85).aspx
test_that("restricted suffixes", {
  expect_equal(path_sanitize("mr."), "mr")
  expect_equal(path_sanitize("mr.."), "mr")
  expect_equal(path_sanitize("mr "), "mr")
  expect_equal(path_sanitize("mr  "), "mr")
})

test_that("relative paths", {
  expect_equal(path_sanitize("."), "")
  expect_equal(path_sanitize(".."), "")
  expect_equal(path_sanitize("./"), "")
  expect_equal(path_sanitize("../"), "")
  expect_equal(path_sanitize("/.."), "")
  expect_equal(path_sanitize("/../"), "")
  expect_equal(path_sanitize("*.|."), "")
})

test_that("relative path with replacement", {
  expect_equal(path_sanitize("..", "_"), "_")
})

test_that("reserved filename in Windows", {
  expect_equal(path_sanitize("con"), "")
  expect_equal(path_sanitize("COM1"), "")
  expect_equal(path_sanitize("PRN."), "")
  expect_equal(path_sanitize("aux.txt"), "")
  expect_equal(path_sanitize("LPT9.asdfasdf"), "")
  expect_equal(path_sanitize("LPT10.txt"), "LPT10.txt")
})

test_that("reserved filename in Windows with replacement", {
  expect_equal(path_sanitize("con", "_"), "_")
  expect_equal(path_sanitize("COM1", "_"), "_")
  expect_equal(path_sanitize("PRN.", "_"), "_")
  expect_equal(path_sanitize("aux.txt", "_"), "_")
  expect_equal(path_sanitize("LPT9.asdfasdf", "_"), "_")
  expect_equal(path_sanitize("LPT10.txt", "_"), "LPT10.txt")
})

test_that("invalid replacement", {
  expect_equal(path_sanitize(".", "."), "")
  expect_equal(path_sanitize("foo?.txt", ">"), "foo.txt")
  expect_equal(path_sanitize("con.txt", "aux"), "")
  expect_equal(path_sanitize("valid.txt", "\\/:*?\"<>|"), "valid.txt")
})

test_string_fs <- function(str, tmpdir) {
  sanitized <- path_sanitize(str)
  if (sanitized == "") {
    sanitized <- "default"
  }
  filepath <- path(tmpdir, sanitized)

  # Should not contain any directories or relative paths
  expect_equal(path_dir(path("/abs/path", sanitized)), "/abs/path")

  # Should write and read file to disk
  expect_equal(path_dir(path_tidy(filepath)), as.character(tmpdir))
  expect_error_free(
    writeLines("foobar", filepath))
  expect_equal(readLines(filepath), "foobar")
  expect_error_free(file_delete(filepath))
}

test_that("filesystems can read, write and delete sanitized files", {
  skip_if_not_utf8()
  skip_on_os("solaris")

 strings <- c(
   # Windows R functions cannot handle the unicode filenames or long filenames + paths correctly.
   if (!is_windows()) {
     readLines(testthat::test_path("blns.txt.xz"), encoding = "UTF-8")
   },
   "the quick brown fox jumped over the lazy dog",
   "résumé",
   "hello\nworld",
   "semi;colon.js",
   ";leading-semi.js",
   "mode_tslash\\.js",
   "slash/.js",
   "col:on.js",
   "star*.js",
   "question?.js",
   "quote\".js",
   "singlequote'.js",
   "brack<e>ts.js",
   "p|pes.js",
   "plus+.js",
   "'five and six<seven'.js",
   " space at front",
   "space at end ",
   ".period",
   "period.",
   "relative/path/to/some/dir",
   "/abs/path/to/some/dir",
   "~/.notssh/authorized_keys",
   "",
   "h?w",
   "h/w",
   "h*w",
   ".",
   "..",
   "./",
   "../",
   "/..",
   "/../",
   "*.|.",
   "./",
   "./foobar",
   "../foobar",
   "../../foobar",
   "./././foobar",
   "|*.what",
   "LPT9.asdf")

 td <- dir_create(file_temp())
 on.exit(dir_delete(td))
 for (str in strings) {
   test_string_fs(str, td)
 }
})

describe("path", {
  it("returns paths UTF-8 encoded", {
    skip_on_os("solaris")
    expect_equal(Encoding(path("föö")), "UTF-8")
  })

  it("returns paths UTF-8 encoded", {
    skip_on_os("solaris")
    expect_equal(Encoding(path("\U4F60\U597D.R")), "UTF-8")
  })

  it("returns empty strings for empty inputs", {
    expect_equal(path(""), fs_path(""))
    expect_equal(path(character()), fs_path(character()))
    expect_equal(path("foo", character(), "bar"), fs_path(character()))
  })

  it("propagates NA strings", {
    expect_equal(path(NA_character_), fs_path(NA_character_))
    expect_equal(path("foo", NA_character_), fs_path(NA_character_))
    expect_equal(
      path(c("foo", "bar"), c("baz", NA_character_)),
      fs_path(c("foo/baz", NA_character_))
    )
  })

  it("appends a ext if provided", {
    expect_equal(path("foo", ext = "bar"), fs_path("foo.bar"))
    expect_equal(
      path(c("foo", "baz"), ext = "bar"),
      fs_path(c("foo.bar", "baz.bar"))
    )
    expect_equal(
      path(c("foo", "baz", NA_character_), ext = "bar"),
      fs_path(c("foo.bar", "baz.bar", NA_character_))
    )
  })

  it("does not double paths", {
    expect_equal(path("/", "foo"), fs_path("/foo"))
    expect_equal(path("\\", "foo"), fs_path("/foo"))
    expect_equal(path("", "foo"), fs_path("/foo"))
    expect_equal(path("foo/", "bar"), fs_path("foo/bar"))
    expect_equal(path("foo\\", "bar"), fs_path("foo/bar"))
    expect_equal(path("foo//", "bar"), fs_path("foo/bar"))

    # This could be a UNC path, so we keep the doubled path.
    expect_equal(path("//", "foo"), fs_path("//foo"))
  })

  it("errors on paths which are too long", {
    expect_error(path(paste(rep("a", 100000), collapse = "")), "less than PATH_MAX")

    expect_error(do.call(path, as.list(rep("a", 100000))), "less than PATH_MAX")
  })

  it("follows recycling rules", {
    expect_equal(path("foo", character()), fs_path(character()))
    expect_equal(path("foo", "bar"), fs_path("foo/bar"))
    expect_equal(
      path("foo", c("bar", "baz")),
      fs_path(c("foo/bar", "foo/baz"))
    )
    expect_equal(
      path(c("foo", "qux"), c("bar", "baz")),
      fs_path(c("foo/bar", "qux/baz"))
    )

    expect_error(path(c("foo", "qux", "foo2"), c("bar", "baz")), "Arguments must have consistent lengths", class = "invalid_argument")
  })
})

describe("path_wd", {
  it("returns an absolute path from the working directory", {
    x <- path_wd()
    expect_true(is_absolute_path(x))

    expect_equal(x, path(getwd()))
  })
})

describe("path_real", {
  it("returns the real path for symbolic links", {
    with_dir_tree(list("foo/bar" = "test"), {
      link_create(path_real("foo"), "foo2")
      expect_equal(path_real("foo2"), path_real("foo"))
    })
  })

  it ("works with indirect symlinks", {
    skip_on_os("windows")

    with_dir_tree("foo", {
      wd <- path_wd()
      link_create(path("..", path_file(wd)), "self")
      link_create("self", "link")
      expect_equal(path_real("link"), wd)
    })
  })

  it ("works with parent symlinks", {
    skip_on_os("windows")

    # If there are symlinks in the parents of relative paths we need to resolve
    # them.
    # e.g. working directory is /usr/doc with 'doc' being a symlink to
    # /usr/share/doc. If we call `path_real("a")` we should return
    # /usr/share/doc/a.
    with_dir_tree("y", {
      wd <- path_wd()
      link_create("y", "k")
      withr::with_dir("k", {
        file_create("a")
        expect_equal(path_real("a"), path(wd, "/y/a"))
        unlink("a")
      })
    })
  })

  it ("resolves paths before normalizing", {

    skip_on_os("windows")

    # if we have the following hierarchy: a/k/y
    # and a symbolic link 'link-y' pointing to 'y' in directory 'a',
    # then `path_real("link-y/..")` should return 'k', not 'a'.
    with_dir_tree("a/k/y", {
      link_create("a/k/y", "link-y")

      expect_equal(path_real("link-y/.."), path_real("a/k"))
    })
  })

  it ("resolves paths before normalizing", {

    skip_on_os("windows")

    # if we have the following hierarchy: a/k/y
    # and a symbolic link 'link-y' pointing to 'y' in directory 'a',
    # then `path_real("link-y/..")` should return 'k', not 'a'.
    with_dir_tree("k", {
      wd <- path_wd()
      link_create(wd, "link")

      withr::with_dir(path_dir(wd), {
        base <- path_file(wd)
        expect_equal(path_real(path(base, "link")), wd)
        expect_equal(path_real(path(base, "link/k")), path(wd, "k"))
      })
    })
  })

  it("propagates NAs", {
    with_dir_tree(list("foo/bar" = "test"), {
      link_create(path_real("foo"), fs_path("foo2"))
      expect_equal(path_real(NA_character_), fs_path(NA_character_))
      expect_equal(
        path_real(c("foo2", NA_character_)),
        fs_path(c(path_real("foo"), NA_character_))
      )
    })
  })

  it("converts inputs to character if required", {
    with_dir_tree(list("foo/bar" = "test"), {
      link_create(path_real("foo"), "2")
      expect_equal(path_real(NA), fs_path(NA_character_)) ## NA (logical) coerced to NA_character_
      expect_equal(path_real(2), path_real("2"))
    })
  })

})

describe("path_split", {
  it("returns the path split", {
    expect_equal(path_split("foo/bar")[[1]], c("foo", "bar"))
    expect_equal(path_split(c("foo/bar", "foo/baz")), list(c("foo", "bar"), c("foo", "baz")))
  })

  it("does not split the root path", {
    expect_equal(path_split("/usr/bin")[[1]], c("/", "usr", "bin"))
    expect_equal(path_split("c:/usr/bin")[[1]], c("C:", "usr", "bin"))
    expect_equal(path_split("X:/usr/bin")[[1]], c("X:", "usr", "bin"))
    expect_equal(path_split("//server/usr/bin")[[1]], c("//server", "usr", "bin"))
    expect_equal(path_split("\\\\server\\usr\\bin")[[1]], c("//server", "usr", "bin"))
  })
})

describe("path_tidy", {
  it("does not expand ~", {
    expect_equal(path_tidy("~/foo"), fs_path("~/foo"))
    expect_equal(path_tidy("~/foo/"), fs_path("~/foo"))
    expect_equal(path_tidy("~//foo"), fs_path("~/foo"))
    expect_equal(path_tidy("~//foo"), fs_path("~/foo"))
    expect_equal(path_tidy("~\\foo\\"), fs_path("~/foo"))
  })

  it("always uses / for delimiting, never multiple / or trailing /", {
    expect_equal(path_tidy("foo/bar/baz"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("foo/bar/baz/"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("foo//bar//baz"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("foo//bar//baz//"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("foo\\bar\\baz"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("foo\\\\bar\\\\baz"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("//foo\\\\bar\\\\baz"), fs_path("//foo/bar/baz"))
    expect_equal(path_tidy("foo\\\\bar\\\\baz\\"), fs_path("foo/bar/baz"))
    expect_equal(path_tidy("foo\\\\bar\\\\baz\\\\"), fs_path("foo/bar/baz"))
  })

  it("always capitalizes and appends windows root paths with /", {
    expect_equal(path_tidy("C:"), fs_path("C:/"))
    expect_equal(path_tidy("c:"), fs_path("C:/"))
    expect_equal(path_tidy("X:"), fs_path("X:/"))
    expect_equal(path_tidy("x:"), fs_path("X:/"))
    expect_equal(path_tidy("c:/"), fs_path("C:/"))
    expect_equal(path_tidy("c://"), fs_path("C:/"))
    expect_equal(path_tidy("c:\\"), fs_path("C:/"))
    expect_equal(path_tidy("c:\\\\"), fs_path("C:/"))
  })

  it("passes NA along", {
    expect_equal(path_tidy(NA_character_), fs_path(NA_character_))
    expect_equal(
      path_tidy(c("foo/bar", NA_character_)),
      fs_path(c("foo/bar", NA_character_))
    )
  })

  it("works for non-UTF8 path", {
    x <- "folder//fa\xE7ile.txt/"
    Encoding(x) <- "latin1"
    out <- fs::path_tidy(x)
    expect_equal(Encoding(out), "UTF-8")
    expect_equal(out, fs_path("folder/façile.txt"))
  })

  it("converts inputs to character if required", {
      expect_equal(path_tidy(c("foo/bar", NA)), fs_path(c("foo/bar", NA_character_)))
      expect_equal(path_tidy(1), fs_path("1"))
  })
})

describe("path_temp", {
  it("returned tidies tempdir()", {
    expect_equal(path_temp(), path_tidy(tempdir()))
  })
})

describe("path_ext", {
  it ("returns 0 length outputs for 0 length inputs", {
    expect_equal(path_ext(character()), character())
  })

  it ("returns the path extension, or \"\" if one does not exist", {
    expect_equal(path_ext("foo.bar"), "bar")
    expect_equal(path_ext("foo.boo.bar"), "bar")
    expect_equal(path_ext("foo.boo.biff.bar"), "bar")
    expect_equal(path_ext(".csh.rc"), "rc")
    expect_equal(path_ext("nodots"), "")
    expect_equal(path_ext(".cshrc"), "")
    expect_equal(path_ext("...manydots"), "")
    expect_equal(path_ext("...manydots.ext"), "ext")
    expect_equal(path_ext("."), "")
    expect_equal(path_ext(".."), "")
    expect_equal(path_ext("........"), "")
    expect_equal(path_ext(""), "")
    expect_equal(path_ext(".bar"), "")
    expect_equal(path_ext("foo/.bar"), "")
    expect_equal(path_ext(c("foo.bar", NA_character_)), c("bar", NA_character_))
    expect_equal(path_ext("foo.bar/baz"), "")
  })
  it ("works with non-ASCII inputs", {
    skip_if_not_utf8()

    expect_equal(path_ext("f\U00F6\U00F6.txt"), "txt")
    expect_equal(path_ext("f\U00F6\U00F6.t\U00E4r"), "t\U00E4r")
  })
  it("returns a normal character vector", {
    expect_equal(class(path_dir("foo.sh")), "character")
  })
})

describe("path_ext_remove", {
  it ("removes the path extension", {
    expect_equal(path_ext_remove("foo.bar"), "foo")
    expect_equal(path_ext_remove("foo.boo.bar"), "foo.boo")
    expect_equal(path_ext_remove("foo.boo.biff.bar"), "foo.boo.biff")
    expect_equal(path_ext_remove(".csh.rc"), ".csh")
    expect_equal(path_ext_remove("nodots"), "nodots")
    expect_equal(path_ext_remove(".cshrc"), ".cshrc")
    expect_equal(path_ext_remove("...manydots"), "...manydots")
    expect_equal(path_ext_remove("...manydots.ext"), "...manydots")
    expect_equal(path_ext_remove("."), ".")
    expect_equal(path_ext_remove(".."), "..")
    expect_equal(path_ext_remove("........"), "........")
    expect_equal(path_ext_remove(""), "")
    expect_equal(path_ext_remove(NA_character_), NA_character_)
    expect_equal(path_ext_remove(c("foo.bar", NA_character_)), c("foo", NA_character_))
    expect_equal(path_ext_remove(".bar"), ".bar")
    expect_equal(path_ext_remove("foo/.bar"), "foo/.bar")
    expect_equal(path_ext_remove("foo.bar/abc.123"), "foo.bar/abc")
    expect_equal(path_ext_remove("foo.bar/abc"), "foo.bar/abc")
  })
  it ("works with non-ASCII inputs", {
    skip_if_not_utf8()

    expect_equal(path_ext_remove("f\U00F6\U00F6.txt"), "f\U00F6\U00F6")
    expect_equal(path_ext_remove("f\U00F6\U00F6.t\U00E4r"), "f\U00F6\U00F6")
  })
})

describe("path_ext_set", {
  it ("replaces the path extension", {
    expect_equal(path_ext_set("foo.bar", "baz"), fs_path("foo.baz"))
    expect_equal(path_ext_set("foo.boo.bar", "baz"), fs_path("foo.boo.baz"))
    expect_equal(
      path_ext_set("foo.boo.biff.bar", "baz"),
      fs_path("foo.boo.biff.baz")
    )
    expect_equal(path_ext_set(".csh.rc", "gz"), fs_path(".csh.gz"))
    expect_equal(path_ext_set("nodots", "bar"), fs_path("nodots.bar"))
    expect_equal(path_ext_set(".cshrc", "bar"), fs_path(".cshrc.bar"))
    expect_equal(
      path_ext_set("...manydots", "bar"),
      fs_path("...manydots.bar")
    )
    expect_equal(
      path_ext_set("...manydots.ext", "bar"),
      fs_path("...manydots.bar")
    )
    expect_equal(path_ext_set(".", "bar"), fs_path("..bar"))
    expect_equal(path_ext_set("..", "bar"), fs_path("...bar"))
    expect_equal(path_ext_set("........", "bar"), fs_path(".........bar"))
    expect_equal(path_ext_set("", "bar"), fs_path(".bar"))
    expect_equal(path_ext_set(NA_character_, "bar"), fs_path(NA_character_))
    expect_equal(
      path_ext_set(c("foo", NA_character_), "bar"),
      fs_path(c("foo.bar", NA_character_))
    )
    expect_equal(path_ext_set(".bar", "baz"), fs_path(".bar.baz"))
    expect_equal(path_ext_set("foo/.bar", "baz"), fs_path("foo/.bar.baz"))
    expect_equal(path_ext_set("foo", ""), fs_path("foo"))
  })
  it ("works the same with and without a leading . for ext", {
    expect_equal(path_ext_set("foo", "bar"), fs_path("foo.bar"))
    expect_equal(path_ext_set("foo", ".bar"), fs_path("foo.bar"))
  })
  it ("only removes a leading . from the extension", {
    expect_equal(path_ext_set("foo", "b.ar"), fs_path("foo.b.ar"))
    expect_equal(path_ext_set("foo", ".b.ar"), fs_path("foo.b.ar"))
  })
  it ("works with multiple paths (#205)", {
    multiple_paths <- c("a", "b")
    expect_equal(
      path_ext_set(multiple_paths, "csv"),
      fs_path(c("a.csv", "b.csv"))
    )
  })
  it ("works with multiple extensions (#250)", {
    multiple_paths <- c("a", "b")
    multiple_exts <- c("csv", "tsv")
    expect_equal(
      path_ext_set(multiple_paths, multiple_exts),
      fs_path(c("a.csv", "b.tsv"))
    )

    expect_error(path_ext_set(multiple_paths, c(multiple_exts, "xls")), class = "fs_error", "consistent lengths")
  })
  it ("works with non-ASCII inputs", {
    skip_if_not_utf8()

    expect_equal(
      path_ext_set("f\U00F6\U00F6.txt", "bar"),
      fs_path("f\U00F6\U00F6.bar")
    )
    expect_equal(
      path_ext_set("f\U00F6\U00F6.t\U00E4r", "bar"),
      fs_path("f\U00F6\U00F6.bar")
    )
  })
})

describe("path_ext<-", {
  it ("replaces the path extension", {
    x <- "...manydots"
    path_ext(x) <- "bar"
    expect_equal(x, fs_path("...manydots.bar"))
  })
})

# test cases derived from https://github.com/python/cpython/blob/6f0eb93183519024cb360162bdd81b9faec97ba6/Lib/test/test_posixpath.py#L276

describe("path_norm", {
  it ("works with POSIX paths", {
    expect_equal(path_norm(""), fs_path("."))
    expect_equal(path_norm(""), fs_path("."))
    expect_equal(path_norm(".."), fs_path(".."))
    expect_equal(path_norm("/../../.."), fs_path("/"))
    expect_equal(path_norm("/../.././.."), fs_path("/"))
    expect_equal(path_norm("../.././.."), fs_path("../../.."))
    expect_equal(path_norm("/"), fs_path("/"))
    expect_equal(path_norm("//"), fs_path("/"))
    expect_equal(path_norm("///"), fs_path("/"))
    expect_equal(path_norm("//foo"), fs_path("//foo"))
    expect_equal(path_norm("//foo/.//bar//"), fs_path("//foo/bar"))
    expect_equal(path_norm("/foo/.//bar//.//..//.//baz"), fs_path("/foo/baz"))
    expect_equal(path_norm("/..//./foo/.//bar"), fs_path("/foo/bar"))
  })

  it ("works with POSIX paths", {
    expect_equal(path_norm("A//////././//.//B"), fs_path("A/B"))
    expect_equal(path_norm("A/./B"), fs_path("A/B"))
    expect_equal(path_norm("A/foo/../B"), fs_path("A/B"))
    expect_equal(path_norm("C:A//B"), fs_path("C:A/B"))
    expect_equal(path_norm("D:A/./B"), fs_path("D:A/B"))
    expect_equal(path_norm("e:A/foo/../B"), fs_path("E:A/B"))

    expect_equal(path_norm("C:///A//B"), fs_path("C:/A/B"))
    expect_equal(path_norm("D:///A/./B"), fs_path("D:/A/B"))
    expect_equal(path_norm("e:///A/foo/../B"), fs_path("E:/A/B"))

    expect_equal(path_norm(".."), fs_path(".."))
    expect_equal(path_norm("."), fs_path("."))
    expect_equal(path_norm(""), fs_path("."))
    expect_equal(path_norm("/"), fs_path("/"))
    expect_equal(path_norm("c:/"), fs_path("C:/"))
    expect_equal(path_norm("/../.././.."), fs_path("/"))
    expect_equal(path_norm("c:/../../.."), fs_path("C:/"))
    expect_equal(path_norm("../.././.."), fs_path("../../.."))
    expect_equal(path_norm("C:////a/b"), fs_path("C:/a/b"))
    expect_equal(
      path_norm("//machine/share//a/b"),
      fs_path("//machine/share/a/b")
    )

    expect_equal(path_norm("\\\\?\\D:/XY\\Z"), fs_path("//?/D:/XY/Z"))
  })

  it ("works with missing values", {
    expect_equal(path_norm(NA), fs_path(NA_character_))
    expect_equal(path_norm(c("foo", NA)), fs_path(c("foo", NA)))
    expect_equal(
      path_norm(c(NA, NA)),
      fs_path(c(NA_character_, NA_character_))
    )  })
})

# Test cases derived from https://github.com/python/cpython/blob/6f0eb93183519024cb360162bdd81b9faec97ba6/Lib/test/test_posixpath.py

describe("path_common", {
  it ("finds the common path", {
    expect_error(path_common(c("/usr", "usr")), "Can't mix", class = "fs_error")
    expect_error(path_common(c("usr", "/usr")), "Can't mix", class = "fs_error")

    expect_equal(path_common(c("/usr/local")), fs_path("/usr/local"))
    expect_equal(
      path_common(c("/usr/local", "/usr/local")),
      fs_path("/usr/local")
    )
    expect_equal(
      path_common(c("/usr/local/", "/usr/local")),
      fs_path("/usr/local")
    )
    expect_equal(
      path_common(c("/usr/local/", "/usr/local/")),
      fs_path("/usr/local")
    )
    expect_equal(
      path_common(c("/usr//local/bin", "/usr/local//bin")),
      fs_path("/usr/local/bin")
    )
    expect_equal(
      path_common(c("/usr/./local", "/./usr/local")),
      fs_path("/usr/local")
    )
    expect_equal(
      path_common(c("/", "/dev")),
      fs_path("/")
    )
    expect_equal(
      path_common(c("/usr", "/dev")),
      fs_path("/")
    )
    expect_equal(
      path_common(c("/usr/lib/", "/usr/lib/python3")),
      fs_path("/usr/lib")
    )
    expect_equal(
      path_common(c("/usr/lib/", "/usr/lib64/")),
      fs_path("/usr")
    )

    expect_equal(
      path_common(c("/usr/lib", "/usr/lib64")),
      fs_path("/usr")
    )
    expect_equal(
      path_common(c("/usr/lib/", "/usr/lib64")),
      fs_path("/usr")
    )

    expect_equal(
      path_common(c("spam")),
      fs_path("spam")
    )
    expect_equal(
      path_common(c("spam", "spam")),
      fs_path("spam")
    )
    expect_equal(
      path_common(c("spam", "alot")),
      fs_path("")
    )
    expect_equal(
      path_common(c("and/jam", "and/spam")),
      fs_path("and")
    )
    expect_equal(
      path_common(c("and//jam", "and/spam//")),
      fs_path("and")
    )
    expect_equal(
      path_common(c("and/./jam", "./and/spam")),
      fs_path("and")
    )
    expect_equal(path_common(c("and/jam", "and/spam", "alot")), fs_path(""))
    expect_equal(path_common(c("and/jam", "and/spam", "and")), fs_path("and"))

    expect_equal(path_common(c("")), fs_path(""))
    expect_equal(path_common(c("", "spam/alot")), fs_path(""))

    expect_error(path_common(c("", "/spam/alot")), "Can't mix", class = "fs_error")
  })

  it("returns NA if any input is NA", {
    expect_equal(path_common(NA), fs_path(NA_character_))
    expect_equal(path_common(c("and/jam", NA)), fs_path(NA_character_))
    expect_equal(path_common(c("and/jam", NA, "and")), fs_path(NA_character_))
  })
})

describe("path_has_parent", {
  it("works on single paths", {
    expect_false(path_has_parent("foo", "bar"))
    expect_false(path_has_parent("foo", "foo/bar"))

    expect_false(path_has_parent("/usr/var2/log", "/usr/var"))

    expect_true(path_has_parent("foo/bar", "foo"))
    expect_true(path_has_parent("path/myfiles/myfile", "path/to/files/../../myfiles"))
  })
  it("works with multiple paths", {
    expect_equal(path_has_parent(c("/a/b/c", "x/y"), "/a/b"), c(TRUE, FALSE))

    expect_equal(path_has_parent("/a/b/c", c("/a/b", "/x/y")), c(TRUE, FALSE))

    expect_error(path_has_parent(c("/a/b/c", "x/y"), c("/a/b", "x/y", "foo/bar")), "consistent lengths", class = "invalid_argument")
  })
})

describe("path_join", {
    it("converts inputs to character if required", {
        expect_equal(path_join(c("a", NA)), path_join(c("a", NA_character_)))
        expect_equal(path_join(c("a", 1)), path_join(c("a", "1")))
        expect_equal(path_join(list(c("a", 1), c("a/b", 2))), fs_path(c(path_join(c("a", "1")), path_join(c("a/b", "2")))))
    })
    it("works with list inputs", {
        expect_equal(path_join(list(c("foo", "bar"), c("a/b", "c"))), fs_path(c(path_join(c("foo", "bar")), path_join(c("a/b", "c")))))
        expect_equal(path_join(list(NA, c("a", 1), c("a/b", 2))), fs_path(c(path_join(NA_character_), path_join(c("a", "1")), path_join(c("a/b", "2")))))
    })
})

# derived from https://github.com/python/cpython/blob/6f0eb93183519024cb360162bdd81b9faec97ba6/Lib/test/test_posixpath.py#L483
describe("path_rel", {
  it("works for POSIX paths", {
    cur_dir <- path_file(getwd())
    expect_equal(path_rel("a"), fs_path("a"))
    expect_equal(path_rel(path_abs("a")), fs_path("a"))
    expect_equal(path_rel("a/b"), fs_path("a/b"))
    expect_equal(path_rel("../a/b"), fs_path("../a/b"))
    expect_equal(path_rel("a", "../b"), path_join(c("..", cur_dir, "a")))
    expect_equal(path_rel("a/b", "../c"), path_join(c("..", cur_dir, "a", "b")))
    expect_equal(path_rel("a", "b/c"), fs_path("../../a"))
    expect_equal(path_rel("a", "a"), fs_path("."))
    expect_equal(
      path_rel("/foo/bar/bat", "/x/y/z"),
      fs_path("../../../foo/bar/bat")
    )
    expect_equal(path_rel("/foo/bar/bat", "/foo/bar"), fs_path("bat"))
    expect_equal(path_rel("/foo/bar/bat", "/"), fs_path("foo/bar/bat"))
    expect_equal(path_rel("/", "/foo/bar/bat"), fs_path("../../.."))
    expect_equal(path_rel("/foo/bar/bat", "/x"), fs_path("../foo/bar/bat"))
    expect_equal(path_rel("/x", "/foo/bar/bat"), fs_path("../../../x"))
    expect_equal(path_rel("/", "/"), fs_path("."))
    expect_equal(path_rel("/a", "/a"), fs_path("."))
    expect_equal(path_rel("/a/b", "/a/b"), fs_path("."))

    expect_equal(
      path_rel(c("a", "a/b", "a/b/c"), "a/b"),
      fs_path(c("..", ".", "c"))
    )
  })

  it("works for windows paths", {
    expect_equal(
      path_rel("c:/foo/bar/bat", "c:/x/y"),
      fs_path("../../foo/bar/bat")
    )
    expect_equal(
      path_rel("//conky/mountpoint/a", "//conky/mountpoint/b/c"),
      fs_path("../../a")
    )
    expect_equal(
      path_rel("d:/Users/a", "D:/Users/b"),
      fs_path("../a")
    )
  })

  it("expands path before computing relativity", {
    expect_equal(path_rel("/foo/bar/baz", "~"), path_rel("/foo/bar/baz", path_expand("~")))
    expect_equal(path_rel("~/foo/bar/baz", "~"), fs_path("foo/bar/baz"))
  })

  it("propagates NAs", {
    expect_equal(path_rel(NA_character_), fs_path(NA_character_))
    expect_equal(path_rel("/foo/bar/baz", NA_character_), fs_path(NA_character_))
  })

  it("can be reversed by path_abs", {
    skip_on_os("windows")

    f <- file_temp()
    expect_equal(path_abs(path_rel(f)), f)

    home <- path_home()
    home_f <- path_abs(path_home("../../foo/bar/baz"))

    expect_equal(path_abs(path_rel(home_f, start = home), start = home), home_f)
  })
})

describe("path_abs", {
  it("uses the start parameter", {
    expect_equal(path_abs("b", "a"), path_abs("a/b"))
  })
})

describe("path_home", {
  # The trailing slash is needed to ensure we get path expansion
  # on POSIX systems when readline support is not built in. (#60)
  it("is equivalent to path_expand(\"~/\")", {
    expect_equal(path_home(), path_tidy(path_expand("~/")))
  })
})

describe("path_dir", {
  it("works like dirname for normal paths", {
    expect_equal(path_dir("foo/bar"), "foo")
    expect_equal(path_dir("bar"), ".")
    expect_equal(path_dir(c("foo/bar", "baz")), c("foo", "."))
  })
  it("propagates NAs", {
    expect_equal(path_dir(NA_character_), NA_character_)
    expect_equal(path_dir(c("foo/bar", NA)), c("foo", NA_character_))
  })
  it("returns a normal character vector", {
    expect_equal(class(path_dir("foo/bar")), "character")
  })
})

describe("path_file", {
  it("works like dirname for normal paths", {
    expect_equal(path_file("foo/bar"), "bar")
    expect_equal(path_file("bar"), "bar")
    expect_equal(path_file(c("foo/bar", "baz")), c("bar", "baz"))
  })
  it("propagates NAs", {
    expect_equal(path_file(NA_character_), NA_character_)
    expect_equal(path_file(c("foo/bar", NA)), c("bar", NA_character_))
  })
  it("returns a normal character vector", {
    expect_equal(class(path_dir("foo/bar")), "character")
  })
})

# These tests were adapted from
# https://github.com/python/cpython/blob/48e8c82fc63d2ddcddce8aa637a892839b551619/Lib/test/test_ntpath.py,
# hence the flying circus names, but I like dead parrots too, so keeping them.
describe("path_expand", {
  it("works on windows", {
    withr::local_envvar(c(FS_IS_WINDOWS = "TRUE"))
    withr::with_envvar(c("USERPROFILE" = NA, "HOMEDRIVE" = NA, "HOMEPATH" = NA), {
      expect_equal(path_expand("~test"), fs_path("~test"))
    })
    withr::with_envvar(c("USERPROFILE" = NA, "HOMEDRIVE" = "C:\\", "HOMEPATH" = "eric\\idle"), {
      expect_equal(path_expand("~"), fs_path("C:/eric/idle"))
      expect_equal(path_expand("~test"), fs_path("C:/eric/test"))
    })
    withr::with_envvar(c("USERPROFILE" = NA, "HOMEDRIVE" = NA, "HOMEPATH" = "eric/idle"), {
      expect_equal(path_expand("~"), fs_path("eric/idle"))
      expect_equal(path_expand("~test"), fs_path("eric/test"))
    })
    withr::with_envvar(c("USERPROFILE" = "C:\\idle\\eric"), {
      expect_equal(path_expand("~"), fs_path("C:/idle/eric"))
      expect_equal(path_expand("~test"), fs_path("C:/idle/test"))


      expect_equal(
        path_expand("~test/foo/bar"),
        fs_path("C:/idle/test/foo/bar")
      )
      expect_equal(
        path_expand("~test/foo/bar/"),
        fs_path("C:/idle/test/foo/bar")
      )
      expect_equal(
        path_expand("~test\\foo\\bar"),
        fs_path("C:/idle/test/foo/bar")
      )
      expect_equal(
        path_expand("~test\\foo\\bar\\"),
        fs_path("C:/idle/test/foo/bar")
      )
      expect_equal(
        path_expand("~/foo/bar"),
        fs_path("C:/idle/eric/foo/bar")
      )
      expect_equal(
        path_expand("~\\foo\\bar"),
        fs_path("C:/idle/eric/foo/bar")
      )
    })
    withr::with_envvar(c("USERPROFILE" = "C:\\idle\\eric", "R_FS_HOME" = "C:\\john\\cleese"), {
      # R_FS_HOME overrides userprofile
      expect_equal(path_expand("~"), fs_path("C:/john/cleese"))
      expect_equal(path_expand("~test"), fs_path("C:/john/test"))
    })
  })
  it("repects R_FS_HOME", {
    withr::with_envvar(c("R_FS_HOME" = "/foo/bar"), {
      expect_equal(path_expand("~"), fs_path("/foo/bar"))
      expect_equal(path_expand("~test"), fs_path("/foo/test"))
    })
  })
})

test_that("examples", {
  path <- fs::path("some", "simple", "path", "to", "a", "file.txt")
  expect_equal(
    path_select_components(path, 1:3),
    path("some", "simple", "path")
  )
  expect_equal(
    path_select_components(path, 1:3, "end"),
    path("to", "a", "file.txt")
  )
  expect_equal(
    path_select_components(path, -1, "end"),
    fs::path("some", "simple", "path", "to", "a")
  )
  expect_equal(
    path_select_components(path, 6),
    path("file.txt")
  )
})

test_that("vectorized", {
  path <- c(
    fs::path("some", "simple", "path", "to", "a", "file.txt"),
    fs::path("another", "path", "to", "file2.txt")
  )

  expect_snapshot({
    path_select_components(path, 1:3)
    path_select_components(path, 1:3, "end")
    path_select_components(path, integer())
    unclass(path_select_components(path, integer()))

    path_select_components(fs_path(character()), 1:3)
    path_select_components(fs_path(character()), 1:3, "end")
    class(path_select_components(fs_path(character()), 1:3))
    class(path_select_components(fs_path(character()), 1:3, "end"))
  })
})

#' Path computations
#'
#' All functions apart from `path_real()` are purely path computations, so the
#' files in question do not need to exist on the filesystem.
#' @template fs
#' @name path_math
#' @return The new path(s) in an `fs_path` object, which is a character vector
#'   that also has class `fs_path`. Except `path_split()`, which returns a list
#'   of character vectors of path components.
#' @seealso [path_expand()] for expansion of user's home directory.
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' dir_create("a")
#' file_create("a/b")
#' link_create(path_abs("a"), "c")
#'
#' # Realize the path
#' path_real("c/b")
#'
#' # Split a path
#' parts <- path_split("a/b")
#' parts
#'
#' # Join it together
#' path_join(parts)
#'
#' # Find the absolute path
#' path_abs("..")
#'
#' # Normalize a path
#' path_norm("a/../b\\c/.")
#'
#' # Compute a relative path
#' path_rel("/foo/abc", "/foo/bar/baz")
#'
#' # Find the common path between multiple paths
#' path_common(c("/foo/bar/baz", "/foo/bar/abc", "/foo/xyz/123"))
#'
#' # Cleanup
#' dir_delete("a")
#' link_delete("c")
#' \dontshow{setwd(.old_wd)}
NULL

#' Construct path to a file or directory
#'
#' `path()` constructs a relative path, `path_wd()` constructs an absolute path
#' from the current working directory.
#'
#' @param ... character vectors, if any values are NA, the result will also be
#'   NA. The paths follow the recycling rules used in the tibble package,
#'   namely that only length 1 arguments are recycled.
#' @param ext An optional extension to append to the generated path.
#' @export
#' @seealso [path_home()], [path_package()] for functions to construct paths
#'   relative to the home and package directories respectively.
#' @examples
#' path("foo", "bar", "baz", ext = "zip")
#'
#' path("foo", letters[1:3], ext = "txt")
path <- function(..., ext = "") {
  args <- list(...)
  assert_recyclable(args)

  path_tidy(.Call(fs_path_, lapply(args, function(x) enc2utf8(as.character(x))), ext))
}

assert_recyclable <- function(x) {
  if (length(x) == 0) {
    return()
  }
  len <- vapply(x, length, integer(1))
  max_len <- max(len)
  different <- which(len != 0 & len != max_len)
  assert(
    "Arguments must have consistent lengths, only values of length one are recycled.",
    all(len[different] == 1)
  )
}

#' @rdname path
#' @export
path_wd <- function(..., ext = "") {
  path(getwd(), ..., ext = ext)
}

#' @describeIn path_math returns the canonical path, eliminating any symbolic
#'   links and the special references `~`, `~user`, `.`, and `..`, , i.e. it
#'   calls `path_expand()` (literally) and `path_norm()` (effectively).
#' @export
path_real <- function(path) {
  path <- enc2utf8(as.character(path))
  old <- path_expand(path)

  is_missing <- is.na(path)
  old[!is_missing] <- .Call(fs_realize_, old[!is_missing])

  path_tidy(old)
}

#' Tidy paths
#'
#' untidy paths are all different, tidy paths are all the same.
#' Tidy paths always use `/` to delimit directories, never have
#' multiple `/` or trailing `/` and have colourised output based on the file
#' type.
#'
#' @return An `fs_path` object, which is a character vector that also has class
#'   `fs_path`
#' @template fs
#' @export
path_tidy <- function(path) {
  path <- enc2utf8(as.character(path))
  new_fs_path(.Call(fs_tidy_, path))
}


#' @describeIn path_math splits paths into parts.
#' @export
# TODO: examples
path_split <- function(path) {
  path <- path_tidy(path)

  # Split drive / UNC parts
  # split keep unc paths together, but keep root paths as first part.
  # //foo => '//foo' 'bar'
  # /foo/bar => '/' 'foo' 'bar'
  strsplit(path, "^(?=/)(?!//)|(?<!^)(?<!^/)/", perl = TRUE)
}

#' @describeIn path_math joins parts together. The inverse of [path_split()].
#' See [path()] to concatenate vectorized strings into a path.
#' @param parts A character vector or a list of character vectors, corresponding
#'   to split paths.
#' @export
path_join <- function(parts) {
  if (length(parts) == 0) {
    return(path_tidy(""))
  }
  if (!is.list(parts)) {
    return(path_tidy(.Call(fs_path_, as.list(enc2utf8(as.character(parts))), "")))
  }
  path_tidy(vapply(parts, path_join, character(1)))
}

#' @describeIn path_math returns a normalized, absolute version of a path.
#' @export
path_abs <- function(path, start = ".") {
  if (!is_absolute_path(start)) {
    start <- path_norm(path(getwd(), start))
  }
  is_abs <- is_absolute_path(path)
  path[is_abs] <- path_norm(path[is_abs])
  path[!is_abs] <- path_norm(path(start, path[!is_abs]))
  path_tidy(path)
}


#' @describeIn path_math eliminates `.` references and rationalizes up-level
#'   `..` references, so `A/./B` and `A/foo/../B` both become `A/B`, but `../B`
#'   is not changed. If one of the paths is a symbolic link, this may change the
#'   meaning of the path, so consider using `path_real()` instead.
#' @export
path_norm <- function(path) {
  non_missing <- !is.na(path)

  parts <- path_split(path)
  path_norm_one <- function(p) {
    p <- p[p != "."]
    double_dots <- p == ".."
    if (any(double_dots)) {
      res <- character(length(p))
      size <- 0
      is_abs <- is_absolute_path(p[[1]])
      for (i in seq_along(p)) {
        if (p[[i]] != ".." || (! is_abs && size == 0) || (size > 0 && res[[size]] == "..")) {
          res[[size <- size + 1]] <- p[[i]]
        } else if (size > 0) {
          size <- size - 1
        }
      }
      res <- res[seq(1, size)]
      if (is_abs && res[[1]] != p[[1]]) {
        res <- c(p[[1]], res)
      }
      p <- res
    }
    if (length(p) == 0) {
      return(path_tidy("."))
    }
    path_join(p)
  }
  parts[non_missing] <- vapply(parts[non_missing], path_norm_one, character(1))
  path_tidy(parts)
}

#' @describeIn path_math computes the path relative to the `start` path,
#'   which can be either an absolute or relative path.
#' @export
#' @param start A starting directory to compute the path relative to.
# This implementation is partially derived from
# https://github.com/python/cpython/blob/9c99fd163d5ca9bcc0b7ddd0d1e3b8717a63237c/Lib/posixpath.py#L446
path_rel <- function(path, start = ".") {
  start <- path_abs(path_expand(start))
  path <- path_abs(path_expand(path))

  path_rel_one <- function(p) {
    common <- path_common(c(start, p))
    starts <- path_split(start)[[1]]
    paths <- path_split(p)[[1]]

    i <- length(path_split(common)[[1]])
    double_dot_part <- rep("..", (length(starts) - i))
    if (i + 1 <= length(paths)) {
      path_part <- paths[seq(i + 1, length(paths))]
    } else {
      path_part <- character()
    }
    rels <- c(double_dot_part, path_part)
    if (length(rels) == 0) {
      return(path_tidy("."))
    }

    path_join(rels)
  }
  if (is.na(start)) {
    return(path_tidy(NA_character_))
  }

  is_missing <- is.na(path)

  path[!is_missing] <- vapply(path[!is_missing], path_rel_one, character(1))

  path_tidy(path)
}

#' Finding the User Home Directory
#'
#' * `path_expand()` performs tilde expansion on a path, replacing instances of
#' `~` or `~user` with the user's home directory.
#' * `path_home()` constructs a path within the expanded users home directory,
#' calling it with _no_ arguments can be useful to verify what fs considers the
#' home directory.
#' * `path_expand_r()` and `path_home_r()` are equivalents which always use R's
#' definition of the home directory.
#' @details
#' `path_expand()` differs from [base::path.expand()] in the interpretation of
#' the home directory of Windows. In particular `path_expand()` uses the path
#' set in the `USERPROFILE` environment variable and, if unset, then uses
#' `HOMEDRIVE`/`HOMEPATH`.
#'
#' In contrast [base::path.expand()] first checks for `R_USER` then `HOME`,
#' which in the default configuration of R on Windows are both set to the user's
#' document directory, e.g. `C:\\Users\\username\\Documents`.
#' [base::path.expand()] also does not support `~otheruser` syntax on Windows,
#' whereas `path_expand()` does support this syntax on all systems.
#'
#' This definition makes fs more consistent with the definition of home
#' directory used on Windows in other languages, such as
#' [python](https://docs.python.org/3/library/os.path.html#os.path.expanduser)
#' and [rust](https://doc.rust-lang.org/std/env/fn.home_dir.html#windows). This
#' is also more compatible with external tools such as git and ssh, both of
#' which put user-level files in `USERPROFILE` by default. It also allows you to
#' write portable paths, such as `~/Desktop` that points to the Desktop location
#' on Windows, macOS and (most) Linux systems.
#'
#' Users can set the `R_FS_HOME` environment variable to override the
#' definitions on any platform.
#' @seealso [R for Windows FAQ - 2.14](https://cran.r-project.org/bin/windows/base/rw-FAQ.html#What-are-HOME-and-working-directories_003f)
#' for behavior of [base::path.expand()].
#' @param ... Additional paths appended to the home directory by [path()].
#' @inheritParams path_math
#' @export
#' @examples
#' # Expand a path
#' path_expand("~/bin")
#'
#' # You can use `path_home()` without arguments to see what is being used as
#' # the home diretory.
#' path_home()
#' path_home("R")
#'
#' # This will likely differ from the above on Windows
#' path_home_r()
path_expand <- function(path) {
  path <- enc2utf8(path)

  # We use the windows implementation if R_FS_HOME is set or if on windows
  path_tidy(.Call(fs_expand_, path, Sys.getenv("R_FS_HOME") != "" || is_windows()))
}

#' @rdname path_expand
#' @export
path_expand_r <- function(path) {
  path <- enc2utf8(path)

  # Unconditionally use R_ExpandFileName
  path_tidy(.Call(fs_expand_, path, FALSE))
}

#' @rdname path_expand
#' @export
path_home <- function(...) {
  path(path_expand("~/"), ...)
}

#' @rdname path_expand
#' @export
path_home_r <- function(...) {
  path(path_expand_r("~/"), ...)
}

#' Manipulate file paths
#'
#' `path_file()` returns the filename portion of the path, `path_dir()` returns
#' the directory portion. `path_ext()` returns the last extension (if any) for a
#' path. `path_ext_remove()` removes the last extension and returns the rest of
#' the path. `path_ext_set()` replaces the extension with a new extension. If
#' there is no existing extension the new extension is appended.
#'
#' Note because these are not full file paths they return regular character
#' vectors, not `fs_path()` objects.
#' @template fs
#' @param ext,value The new file extension.
#' @seealso [base::basename()], [base::dirname()]
#' @export
#' @examples
#' path_file("dir/file.zip")
#'
#' path_dir("dir/file.zip")
#'
#' path_ext("dir/file.zip")
#'
#' path_ext("file.tar.gz")
#'
#' path_ext_remove("file.tar.gz")
#'
#' # Only one level of extension is removed
#' path_ext_set(path_ext_remove("file.tar.gz"), "zip")
#' @export
path_file <- function(path) {
  is_missing <- is.na(path)
  path[!is_missing] <- basename(path[!is_missing])
  as.character(path)
}

#' @rdname path_file
#' @export
path_dir <- function(path) {
  is_missing <- is.na(path)
  path[!is_missing] <- dirname(path[!is_missing])
  as.character(path_tidy(path))
}

#' @rdname path_file
#' @export
path_ext <- function(path) {
  if (length(path) == 0) {
    return(character())
  }

  res <- captures(path_file(path), regexpr("(?<!^|[.]|/)[.]([^.]+)$", path_file(path), perl = TRUE))[[1]]
  res[!is.na(path) & is.na(res)] <- ""
  res
}

#' @rdname path_file
#' @export
path_ext_remove <- function(path) {
  dir <- path_dir(path)
  file <- sub("(?<!^|[.]|/)[.][^.]+$", "", path_file(path), perl = TRUE)

  na <- is.na(path)
  no_dir <- dir == "." | dir == ""

  path[!na & no_dir] <- path_tidy(file[!na & no_dir])

  path[!na & !no_dir] <- path(dir[!na & !no_dir], file[!na & !no_dir])

  path
}

#' @rdname path_file
#' @export
path_ext_set <- function(path, ext) {

  if (!(length(ext) == length(path) || length(ext) == 1)) {
    assert_recyclable(list(path, ext))
  }

  # Remove a leading . if present
  ext <- sub("[.]", "", ext)

  has_ext <- nzchar(ext)
  to_set <- !is.na(path) & has_ext

  if (length(ext) == 1) {
    ext <- rep(ext, sum(to_set))
  }

  path[to_set] <- paste0(
    path_ext_remove(path[to_set]), ".", ext
  )

  path_tidy(path)
}

#' @rdname path_file
#' @export
`path_ext<-` <- function(path, value) {
  path_ext_set(path, value)
}

#' @describeIn path_math finds the common parts of two (or more) paths.
#' @export
path_common <- function(path) {

  is_missing <- is.na(path)

  if (any(is_missing)) {
    return(path_tidy(NA))
  }

  is_abs <- is_absolute_path(path)

  # We must either have all absolute paths, or all relative paths.
  if (!(all(is_abs) || all(!is_abs))) {
    stop(fs_error("Can't mix absolute and relative paths"))
  }

  path <- path_norm(path)
  path <- sort(path)

  # remove . entries from the split paths
  parts <- lapply(path_split(path), function(x) x[x != "."])
  s1 <- parts[[1]]
  s2 <- parts[[length(parts)]]
  common <- s1
  for (i in seq_along(s1)) {
    if (s1[[i]] != s2[[i]]) {
      if (i == 1) {
        common <- ""
      } else {
        common <- s1[seq(1, i - 1)]
      }
      break;
    }
  }
  path_join(common)
}

#' Filter paths
#'
#' @template fs
#' @param glob A wildcard aka globbing pattern (e.g. `*.csv`) passed on to [grep()] to filter paths.
#' @param regexp A regular expression (e.g. `[.]csv$`) passed on to [grep()] to filter paths.
#' @param invert If `TRUE` return files which do _not_ match
#' @param ... Additional arguments passed to [grep].
#' @export
#' @examples
#' path_filter(c("foo", "boo", "bar"), glob = "*oo")
#' path_filter(c("foo", "boo", "bar"), glob = "*oo", invert = TRUE)
#'
#' path_filter(c("foo", "boo", "bar"), regexp = "b.r")
path_filter <- function(path, glob = NULL, regexp = NULL, invert = FALSE, ...) {
  if (!is.null(glob)) {
    if (!is.null(regexp)) {
      stop(fs_error("`glob` and `regexp` cannot both be set."))
    }
    regexp <- utils::glob2rx(glob)
  }
  if (!is.null(regexp)) {
    path <- grep(x = path, pattern = regexp, value = TRUE, invert = isTRUE(invert), ...)
  }
  setNames(path_tidy(path), path)
}

#' @describeIn path_math determine if a path has a given parent.
#' @param parent The parent path.
#' @export
path_has_parent <- function(path, parent) {
  path <- path_abs(path)
  parent <- path_abs(parent)

  res <- logical(length(path))

  assert_recyclable(list(path, parent))
  if (length(path) == 1) {
    path <- rep(path, length(parent))
  }

  if (length(parent) == 1) {
    parent <- rep(parent, length(path))
  }

  for (i in seq_along(path)) {
    res[[i]] <- identical(as.character(path_common(c(path[[i]], parent[[i]]))), as.character(parent[[i]]))
  }
  res
}

#' List files
#'
#' @description
#' `dir_ls()` is equivalent to the `ls` command. It returns filenames as a
#' named `fs_path` character vector. The names are equivalent to the values,
#' which is useful for passing onto functions like `purrr::map_dfr()`.
#'
#' `dir_info()` is equivalent to `ls -l` and a shortcut for
#' `file_info(dir_ls())`.
#'
#' `dir_map()` applies a function `fun()` to each entry in the path and returns
#' the result in a list.
#'
#' `dir_walk()` calls `fun` for its side-effect and returns the input `path`.
#'
#' @param type File type(s) to return, one or more of "any", "file", "directory",
#'   "symlink", "FIFO", "socket", "character_device" or "block_device".
#' @param recurse If `TRUE` recurse fully, if a positive number the number of levels
#'   to recurse.
#' @param recursive (Deprecated) If `TRUE` recurse fully.
#' @inheritParams path_filter
#' @param all If `TRUE` hidden files are also returned.
#' @param fail Should the call fail (the default) or warn if a file cannot be
#'   accessed.
#' @template fs
#' @export
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' dir_ls(R.home("share"), type = "directory")
#'
#' # Create a shorter link
#' link_create(system.file(package = "base"), "base")
#'
#' dir_ls("base", recurse = TRUE, glob = "*.R")
#'
#' # If you need the full paths input an absolute path
#' dir_ls(path_abs("base"))
#'
#' dir_map("base", identity)
#'
#' dir_walk("base", str)
#'
#' dir_info("base")
#'
#' # Cleanup
#' link_delete("base")
#' \dontshow{setwd(.old_wd)}
dir_ls <- function(path = ".", all = FALSE, recurse = FALSE, type = "any",
                   glob = NULL, regexp = NULL, invert = FALSE, fail = TRUE,
                   ..., recursive) {
  assert_no_missing(path)

  if (!missing(recursive)) {
    recurse <- recursive
    warning("`recursive` is deprecated, please use `recurse` instead", immediate. = TRUE, call. = FALSE)
  }

  old <- path_expand(path)

  files <- as.character(dir_map(old, identity, all, recurse, type, fail))

  path_filter(files, glob, regexp, invert = invert, ...)
}

directory_entry_types <- c(
  "any" = -1L,
  "unknown" = 1L,
  "file" = 2L,
  "directory" = 4L,
  "symlink" = 8L,
  "FIFO" = 16L,
  "socket" = 32L,
  "character_device" = 64L,
  "block_device" = 128L)

#' @rdname dir_ls
#' @param fun A function, taking one parameter, the current path entry.
#' @export
dir_map <- function(path = ".", fun, all = FALSE, recurse = FALSE, type =
  "any", fail = TRUE) {
  assert_no_missing(path)

  if (is.logical(recurse)) {
    if (isTRUE(recurse)) {
      recurse <- -1
    } else {
      recurse <- 0
    }
  }

  type <- match.arg(type, names(directory_entry_types), several.ok = TRUE)

  old <- path_expand(path)

  .Call(fs_dir_map_, old, fun, all, sum(directory_entry_types[type]), as.integer(recurse), fail)
}

#' @rdname dir_ls
#' @export
dir_walk <- function(path = ".", fun, all = FALSE, recurse = FALSE, type = "any", fail = TRUE) {
  assert_no_missing(path)

  old <- path_expand(path)

  dir_map(old, fun, all, recurse, type, fail)
  invisible(path_tidy(path))
}

#' @rdname dir_ls
#' @export
dir_info <- function(path = ".", all = FALSE, recurse = FALSE,
                     type = "any", regexp = NULL, glob = NULL, fail = TRUE, ...) {
  assert_no_missing(path)

  file_info(dir_ls(path = path, all = all, recurse = recurse, type = type,
    regexp = regexp, glob = glob, fail = fail, ...), fail = fail)
}

describe("dir_ls", {
  it("Does not include '.' or double '/' in results", {
    with_dir_tree(list("foo" = "test"), {
      expect_equal(dir_ls(), new_fs_path(c(foo = "foo")))
      expect_equal(dir_ls("."), new_fs_path(c(foo = "foo")))
    })

    with_dir_tree(list("foo/bar" = "test"), {
      expect_equal(dir_ls(recurse = TRUE), named_fs_path(c("foo", "foo/bar")))
      expect_equal(dir_ls(recurse = TRUE, type = "file"), named_fs_path("foo/bar"))
      expect_equal(dir_ls("./", recurse = TRUE), named_fs_path(c("foo", "foo/bar")))
      expect_equal(dir_ls("foo"), named_fs_path("foo/bar"))
      expect_equal(dir_ls("foo/"), named_fs_path("foo/bar"))
    })
  })

  it("Does not follow symbolic links", {
    with_dir_tree(list("foo/bar/baz" = "test"), {
      link_create(path_abs("foo"), named_fs_path("foo/bar/qux"))
      expect_equal(
        dir_ls(recurse = TRUE),
        named_fs_path(c("foo", "foo/bar", "foo/bar/baz", "foo/bar/qux"))
      )
      expect_equal(
        dir_ls(recurse = TRUE, type = "symlink"),
        named_fs_path("foo/bar/qux")
      )
    })
  })

  it("Uses grep to filter output", {
    with_dir_tree(list(
        "foo/bar/baz" = "test",
        "foo/bar/test2" = "",
        "foo/bar/test3" = ""), {
      expect_equal(
        dir_ls(recurse = TRUE, glob = "*baz"),
        named_fs_path("foo/bar/baz")
      )
      expect_equal(
        dir_ls(recurse = TRUE, regexp = "baz"),
        named_fs_path("foo/bar/baz")
      )
      expect_equal(
        dir_ls(recurse = TRUE, regexp = "[23]"),
        named_fs_path(c("foo/bar/test2", "foo/bar/test3"))
      )
      expect_equal(
        dir_ls(recurse = TRUE, regexp = "(?<=a)z", perl = TRUE),
        named_fs_path("foo/bar/baz")
      )
    })
  })

  it("Does not print hidden files by default", {
    with_dir_tree(list(
        ".foo" = "foo",
        "bar" = "bar"), {
      expect_equal(dir_ls(), named_fs_path("bar"))
      expect_equal(dir_ls(all = TRUE), named_fs_path(c(".foo", "bar")))
    })
  })

  it("can find multiple types", {
    with_dir_tree(list(
        "file" = "foo",
        "dir"), {
      link_create(path_abs("dir"), "link")
      expect_equal(dir_ls(type = "file"), named_fs_path("file"))
      expect_equal(dir_ls(type = "directory"), named_fs_path("dir"))
      expect_equal(dir_ls(type = "symlink"), named_fs_path("link"))
      expect_equal(
        dir_ls(type = c("directory", "symlink")),
        named_fs_path(c("dir", "link"))
      )
      expect_equal(
        dir_ls(type = c("file", "directory", "symlink")),
        named_fs_path(c("dir", "file", "link"))
      )
    })
  })
  it("works with UTF-8 encoded filenames", {
    skip_if_not_utf8()
    skip_on_os("solaris")
    with_dir_tree("\U7684\U6D4B\U8BD5\U6587\U4EF6", {
      file_create("fs\U7684\U6D4B\U8BD5\U6587\U4EF6.docx")
      link_create(path_abs("\U7684\U6D4B\U8BD5\U6587\U4EF6"), "\U7684\U6D4B")

      expect_equal(
        dir_ls(type = "file"),
        named_fs_path("fs\U7684\U6D4B\U8BD5\U6587\U4EF6.docx")
      )
      expect_equal(
        dir_ls(type = "directory"),
        named_fs_path("\U7684\U6D4B\U8BD5\U6587\U4EF6")
      )
      expect_equal(
        dir_ls(type = "symlink"),
        named_fs_path("\U7684\U6D4B")
      )
      expect_equal(path_file(link_path("\U7684\U6D4B")), "\U7684\U6D4B\U8BD5\U6587\U4EF6")
    })
  })
  it("errors on missing input", {
    expect_error(dir_ls(NA), class = "invalid_argument")
  })
  it("warns if fail == FALSE", {
    skip_on_os("windows")
    if (Sys.info()[["effective_user"]] == "root") skip("root user")
    with_dir_tree(list(
        "foo",
        "foo2/bar/baz"), {

      file_chmod("foo", "a-r")
      expect_error(dir_ls(".", recurse = TRUE), class = "EACCES")
      expect_warning(dir_ls(fail = FALSE, recurse = TRUE), class = "EACCES")
      file_chmod("foo", "a+r")

      file_chmod("foo2/bar", "a-r")
      expect_warning(dir_ls("foo2", fail = FALSE, recurse = TRUE), class = "EACCES")
      file_chmod("foo2/bar", "a+r")
    })
  })
})

describe("dir_map", {
  it("can find multiple types", {
    with_dir_tree(list(
        "file" = "foo",
        "dir"), {
      nc <- function(x) nchar(x, keepNA = FALSE)
      expect_equal(dir_map(type = "file", fun = nc), list(4))
      expect_equal(dir_map(type = "directory", fun = nc), list(3))
      expect_equal(dir_map(type = c("file", "directory"), fun = nc), list(3, 4))
    })
  })
  it("errors on missing input", {
    expect_error(dir_map(NA, fun = identity), class = "invalid_argument")
  })

  it("warns if fail == FALSE", {
    skip_on_os("windows")
    if (Sys.info()[["effective_user"]] == "root") skip("root user")
    with_dir_tree(list(
        "foo",
        "foo2/bar/baz"), {

      file_chmod("foo", "a-r")
      expect_error(dir_map(".", fun = identity, recurse = TRUE), class = "EACCES")
      expect_warning(dir_map(fail = FALSE, fun = identity, recurse = TRUE), class = "EACCES")
      file_chmod("foo", "a+r")

      file_chmod("foo2/bar", "a-r")
      expect_warning(dir_map("foo2", fail = FALSE, fun = identity, recurse = TRUE), class = "EACCES")
      file_chmod("foo2/bar", "a+r")
    })
  })
})

describe("dir_walk", {
  it("can find multiple types", {
    x <- character()
    f <- function(p) x <<- p

    with_dir_tree(list(
        "file" = "foo",
        "dir"), {
      link_create(path_abs("dir"), "link")

      dir_walk(type = "file", fun = f)
      expect_equal(x, "file")

      dir_walk(type = "directory", fun = f)
      expect_equal(x, "dir")

      dir_walk(type = "symlink", fun = f)
      expect_equal(x, "link")

      x <- character()
      dir_walk(type = c("directory", "symlink"), fun = function(p) x <<- c(x, p))
      expect_equal(x, c("dir", "link"))

      x <- character()
      dir_walk(type = c("file", "directory", "symlink"), fun = function(p) x <<- c(x, p))
      expect_equal(x, c("dir", "file", "link"))
    })
  })
  it("errors on missing input", {
    expect_error(dir_walk(NA, fun = identity), class = "invalid_argument")
  })

  it("warns if fail == FALSE", {
    skip_on_os("windows")
    if (Sys.info()[["effective_user"]] == "root") skip("root user")
    with_dir_tree(list(
        "foo",
        "foo2/bar/baz"), {

      file_chmod("foo", "a-r")
      expect_error(dir_walk(".", fun = identity, recurse = TRUE), class = "EACCES")
      expect_warning(dir_walk(fail = FALSE, fun = identity, recurse = TRUE), class = "EACCES")
      file_chmod("foo", "a+r")

      file_chmod("foo2/bar", "a-r")
      expect_warning(dir_walk("foo2", fail = FALSE, fun = identity, recurse = TRUE), class = "EACCES")
      file_chmod("foo2/bar", "a+r")
    })
  })
})

describe("dir_info", {
  it("is identical to file_info(dir_ls())", {
    with_dir_tree(list(
        "file" = "foo",
        "dir"), {
      link_create(path_abs("dir"), "link")
      expect_identical(dir_info(), file_info(dir_ls()))
    })
  })
  it("errors on missing input", {
    expect_error(dir_info(NA), class = "invalid_argument")
  })

  it("warns if fail == FALSE", {
    skip_on_os("windows")
    if (Sys.info()[["effective_user"]] == "root") skip("root user")
    with_dir_tree(list(
        "foo",
        "foo2/bar/baz"), {

      file_chmod("foo", "a-r")
      expect_error(dir_info(".", fun = identity, recurse = TRUE), class = "EACCES")
      expect_warning(dir_info(fail = FALSE, fun = identity, recurse = TRUE), class = "EACCES")
      file_chmod("foo", "a+r")

      file_chmod("foo2/bar", "a-r")
      expect_warning(dir_info("foo2", fail = FALSE, fun = identity, recurse = TRUE), class = "EACCES")
      file_chmod("foo2/bar", "a+r")
    })
  })
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

mkdir_ <- function(path, mode) {
    invisible(.Call(`_fs_mkdir_`, path, mode))
}

rmdir_ <- function(path) {
    invisible(.Call(`_fs_rmdir_`, path))
}

dir_map_ <- function(path, fun, all, type, recurse, fail) {
    .Call(`_fs_dir_map_`, path, fun, all, type, recurse, fail)
}

move_ <- function(path, new_path) {
    invisible(.Call(`_fs_move_`, path, new_path))
}

create_ <- function(path, mode) {
    invisible(.Call(`_fs_create_`, path, mode))
}

stat_ <- function(path, fail) {
    .Call(`_fs_stat_`, path, fail)
}

access_ <- function(path, mode) {
    .Call(`_fs_access_`, path, mode)
}

chmod_ <- function(path, mode) {
    invisible(.Call(`_fs_chmod_`, path, mode))
}

unlink_ <- function(path) {
    invisible(.Call(`_fs_unlink_`, path))
}

copyfile_ <- function(path, new_path, overwrite) {
    invisible(.Call(`_fs_copyfile_`, path, new_path, overwrite))
}

chown_ <- function(path, uid, gid) {
    invisible(.Call(`_fs_chown_`, path, uid, gid))
}

touch_ <- function(path, atime, mtime) {
    invisible(.Call(`_fs_touch_`, path, atime, mtime))
}

cleanup_ <- function() {
    invisible(.Call(`_fs_cleanup_`))
}

getmode_ <- function(mode_str, mode) {
    .Call(`_fs_getmode_`, mode_str, mode)
}

strmode_ <- function(mode) {
    .Call(`_fs_strmode_`, mode)
}

file_code_ <- function(path, mode) {
    .Call(`_fs_file_code_`, path, mode)
}

getpwnam_ <- function(name) {
    .Call(`_fs_getpwnam_`, name)
}

getgrnam_ <- function(name) {
    .Call(`_fs_getgrnam_`, name)
}

groups_ <- function() {
    .Call(`_fs_groups_`)
}

users_ <- function() {
    .Call(`_fs_users_`)
}

link_create_hard_ <- function(path, new_path) {
    invisible(.Call(`_fs_link_create_hard_`, path, new_path))
}

link_create_symbolic_ <- function(path, new_path) {
    invisible(.Call(`_fs_link_create_symbolic_`, path, new_path))
}

readlink_ <- function(path) {
    .Call(`_fs_readlink_`, path)
}

realize_ <- function(path) {
    .Call(`_fs_realize_`, path)
}

path_ <- function(paths, ext) {
    .Call(`_fs_path_`, paths, ext)
}

expand_ <- function(path, windows) {
    .Call(`_fs_expand_`, path, windows)
}

tidy_ <- function(path) {
    .Call(`_fs_tidy_`, path)
}

#' Copy files, directories or links
#'
#' @description
#' `file_copy()` copies files.
#'
#' `link_copy()` creates a new link pointing to the same location as the previous link.
#'
#' `dir_copy()` copies the directory recursively at the new location.
#' @details
#' The behavior of `dir_copy()` differs slightly than that of `file.copy()` when
#' `overwrite = TRUE`. The directory will always be copied to `new_path`, even
#' if the name differs from the basename of `path`.
#' @param new_path A character vector of paths to the new locations.
#' @param overwrite Overwrite files if they exist. If this is `FALSE` and the
#'   file exists an error will be thrown.
#' @template fs
#' @return The new path (invisibly).
#' @name copy
#' @export
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' file_create("foo")
#' file_copy("foo", "bar")
#' try(file_copy("foo", "bar"))
#' file_copy("foo", "bar", overwrite = TRUE)
#' file_delete(c("foo", "bar"))
#'
#' dir_create("foo")
#' # Create a directory and put a few files in it
#' files <- file_create(c("foo/bar", "foo/baz"))
#' file_exists(files)
#'
#' # Copy the directory
#' dir_copy("foo", "foo2")
#' file_exists(path("foo2", path_file(files)))
#'
#' # Create a link to the directory
#' link_create(path_abs("foo"), "loo")
#' link_path("loo")
#' link_copy("loo", "loo2")
#' link_path("loo2")
#'
#' # Cleanup
#' dir_delete(c("foo", "foo2"))
#' link_delete(c("loo", "loo2"))
#' \dontshow{setwd(.old_wd)}
file_copy <- function(path, new_path, overwrite = FALSE) {
  # TODO: copy attributes, e.g. cp -p?
  assert_no_missing(path)
  assert_no_missing(new_path)

  old <- path_expand(path)
  new <- path_expand(new_path)

  is_directory <- file_exists(new) & is_dir(new)

  if (length(new) == 1 && is_directory[[1]]) {
    new <- rep(new, length(path))
  }
  assert("Length of `path` must equal length of `new_path`", length(old) == length(new))

  new[is_directory] <- path(new[is_directory], basename(old))

  .Call(fs_copyfile_, old, new, isTRUE(overwrite))

  invisible(path_tidy(new))
}

#' @rdname copy
#' @export
dir_copy <- function(path, new_path, overwrite = FALSE) {
  assert_no_missing(path)
  assert_no_missing(new_path)
  assert("`path` must be a directory", all(is_dir(path)))
  assert("Length of `path` must equal length of `new_path`", length(path) == length(new_path))

  for (i in seq_along(path)) {
    if (!isTRUE(overwrite) && isTRUE(unname(is_dir(new_path[[i]])))) {
      new_path[[i]] <- path(new_path[[i]], path_file(path))
    }
    dir_create(new_path[[i]])

    dirs <- dir_ls(path[[i]], type = "directory", recurse = TRUE, all = TRUE)
    dir_create(path(new_path[[i]], path_rel(dirs, path[[i]])))

    files <- dir_ls(path[[i]], recurse = TRUE,
      type = c("unknown", "file", "FIFO", "socket", "character_device", "block_device"), all = TRUE)
    file_copy(files, path(new_path[[i]], path_rel(files, path[[i]])), overwrite = overwrite)

    links <- dir_ls(path[[i]], recurse = TRUE,
      type = "symlink", all = TRUE)
    link_copy(links, path(new_path[[i]], path_rel(links, path[[i]])), overwrite = overwrite)
  }

  invisible(path_tidy(new_path))
}

#' @rdname copy
#' @export
link_copy <- function(path, new_path, overwrite = FALSE) {
  assert_no_missing(path)
  assert_no_missing(new_path)
  stopifnot(all(is_link(path)))

  old <- path_expand(path)
  new <- path_expand(new_path)

  to_delete <- isTRUE(overwrite) & link_exists(new)
  if (any(to_delete)) {
    link_delete(new[to_delete])
  }

  invisible(link_create(link_path(old), new))
}

#' Select path components by their position/index.
#'
#' `path_select_components()` allows to select individual components from
#' an `fs_path` object via their index.
#'
#' `path_select_components()` is vectorized.
#'
#' @param path A path of class `fs_path`.
#' @param index An integer vector of path positions. (A negative index will work
#'  according to R's usual subsetting rules.)
#' @param from A character of either `"start"` or `"end"` to choose if indexing
#'  should start from the first or last component of the `path`.
#'
#' @return An `fs_path` object, which is a character vector that also has class
#'  `fs_path`.
#'
#' @export
#' @examples
#' path <- fs::path("some", "simple", "path", "to", "a", "file.txt")
#'
#' path_select_components(path, 1:3)
#'
#' path_select_components(path, 1:3, "end")
#'
#' path_select_components(path, -1, "end")
#'
#' path_select_components(path, 6)

path_select_components <- function(path, index, from = c("start", "end")) {
  from <- match.arg(from)

  res <- fs_path(character(length(path)))
  ps <- fs::path_split(path)
  for (idx in seq_along(ps)) {
    path <- ps[[idx]]
    path_seq <- seq_along(path)

    if (length(index) > 0 && max(abs(index)) > length(path)) {
      stop("`seq` contains a higher number than the path has components.")
    }

    if (from == "start") {
      path <- path[index]
    }
    if (from == "end") {
      path_seq <- rev(rev(path_seq)[index])
      path <- path[path_seq]
    }

    res[idx] <- fs::path_join(path)
  }

  res
}

#' Delete files, directories, or links
#'
#' @description
#' `file_delete()` and `link_delete()` delete file and links. Compared to
#' [file.remove] they always fail if they cannot delete the object rather than
#' changing return value or signalling a warning. If any inputs are
#' directories, they are passed to `dir_delete()`, so `file_delete()` can
#' therefore be used to delete any filesystem object.
#'
#' `dir_delete()` will first delete the contents of the directory, then remove
#' the directory. Compared to [unlink] it will always throw an error if the
#' directory cannot be deleted rather than being silent or signalling a warning.
#' @template fs
#' @export
#' @return The deleted paths (invisibly).
#' @name delete
#' @examples
#' \dontshow{.old_wd <- setwd(tempdir())}
#' # create a directory, with some files and a link to it
#' dir_create("dir")
#' files <- file_create(path("dir", letters[1:5]))
#' link <- link_create(path_abs("dir"), "link")
#'
#' # All files created
#' dir_exists("dir")
#' file_exists(files)
#' link_exists("link")
#' file_exists(link_path("link"))
#'
#' # Delete a file
#' file_delete(files[1])
#' file_exists(files[1])
#'
#' # Delete the directory (which deletes the files as well)
#' dir_delete("dir")
#' file_exists(files)
#' dir_exists("dir")
#'
#' # The link still exists, but what it points to does not.
#' link_exists("link")
#' dir_exists(link_path("link"))
#'
#' # Delete the link
#' link_delete("link")
#' link_exists("link")
#' \dontshow{setwd(.old_wd)}
file_delete <- function(path) {
  assert_no_missing(path)

  old <- path_expand(path)

  dirs <- is_dir(old)
  dir_delete(old[dirs])

  .Call(fs_unlink_, old[!dirs])

  invisible(path_tidy(path))
}

#' @rdname delete
#' @export
dir_delete <- function(path) {
  assert_no_missing(path)

  old <- path_expand(path)

  dirs <- dir_ls(old, type = "directory", recurse = TRUE, all = TRUE)
  files <- dir_ls(old,
    type = c("unknown", "file", "symlink", "FIFO", "socket", "character_device", "block_device"),
    recurse = TRUE,
    all = TRUE)
  .Call(fs_unlink_, files)
  .Call(fs_rmdir_, rev(c(old, dirs)))

  invisible(path_tidy(path))
}

#' @rdname delete
#' @export
link_delete <- function(path) {
  assert_no_missing(path)

  assert("`path` must be a link",
    all(is_link(path))
  )

  old <- path_expand(path)

  .Call(fs_unlink_, old)

  invisible(path_tidy(path))
}

#' Sanitize a filename by removing directory paths and invalid characters
#'
#' `path_sanitize()` removes the following:
#' - [Control characters](https://en.wikipedia.org/wiki/C0_and_C1_control_codes)
#' - [Reserved characters](https://kb.acronis.com/content/39790)
#' - Unix reserved filenames (`.` and `..`)
#' - Trailing periods and spaces (invalid on Windows)
#' - Windows reserved filenames (`CON`, `PRN`, `AUX`, `NUL`, `COM1`, `COM2`,
#'   `COM3`, COM4, `COM5`, `COM6`, `COM7`, `COM8`, `COM9`, `LPT1`, `LPT2`,
#'   `LPT3`, `LPT4`, `LPT5`, `LPT6`, LPT7, `LPT8`, and `LPT9`)
#' The resulting string is then truncated to [255 bytes in length](https://en.wikipedia.org/wiki/Comparison_of_file_systems#Limits)
#' @param filename A character vector to be sanitized.
#' @param replacement A character vector used to replace invalid characters.
#' @seealso <https://www.npmjs.com/package/sanitize-filename>, upon which this
#'   function is based.
#' @export
#' @examples
#' # potentially unsafe string
#' str <- "~/.\u0001ssh/authorized_keys"
#' path_sanitize(str)
#'
#' path_sanitize("..")
path_sanitize <- function(filename, replacement = "") {
  illegal <- "[/\\?<>\\:*|\":]"
  control <- "[[:cntrl:]]"
  reserved <- "^[.]+$"
  windows_reserved <- "^(con|prn|aux|nul|com[0-9]|lpt[0-9])([.].*)?$"
  windows_trailing <- "[. ]+$"

  filename <- gsub(illegal, replacement, filename)
  filename <- gsub(control, replacement, filename)
  filename <- gsub(reserved, replacement, filename)
  filename <- gsub(windows_reserved, replacement, filename, ignore.case = TRUE)
  filename <- gsub(windows_trailing, replacement, filename)

  # TODO: this substr should really be unicode aware, so it doesn't chop a
  # multibyte code point in half.
  filename <- substr(filename, 1, 255)
  if (replacement == "") {
    return(filename)
  }
  path_sanitize(filename, "")
}

#' Sanitize a filename by removing directory paths and invalid characters
#'
#' `path_sanitize()` removes the following:
#' - [Control characters](https://en.wikipedia.org/wiki/C0_and_C1_control_codes)
#' - [Reserved characters](https://web.archive.org/web/20230126161942/https://kb.acronis.com/content/39790)
#' - Unix reserved filenames (`.` and `..`)
#' - Trailing periods and spaces (invalid on Windows)
#' - Windows reserved filenames (`CON`, `PRN`, `AUX`, `NUL`, `COM1`, `COM2`,
#'   `COM3`, COM4, `COM5`, `COM6`, `COM7`, `COM8`, `COM9`, `LPT1`, `LPT2`,
#'   `LPT3`, `LPT4`, `LPT5`, `LPT6`, LPT7, `LPT8`, and `LPT9`)
#' The resulting string is then truncated to [255 bytes in length](https://en.wikipedia.org/wiki/Comparison_of_file_systems#Limits)
#' @param filename A character vector to be sanitized.
#' @param replacement A character vector used to replace invalid characters.
#' @seealso <https://www.npmjs.com/package/sanitize-filename>, upon which this
#'   function is based.
#' @export
#' @examples
#' # potentially unsafe string
#' str <- "~/.\u0001ssh/authorized_keys"
#' path_sanitize(str)
#'
#' path_sanitize("..")
path_sanitize <- function(filename, replacement = "") {
  illegal <- "[/\\?<>\\:*|\":]"
  control <- "[[:cntrl:]]"
  reserved <- "^[.]+$"
  windows_reserved <- "^(con|prn|aux|nul|com[0-9]|lpt[0-9])([.].*)?$"
  windows_trailing <- "[. ]+$"

  filename <- gsub(illegal, replacement, filename)
  filename <- gsub(control, replacement, filename)
  filename <- gsub(reserved, replacement, filename)
  filename <- gsub(windows_reserved, replacement, filename, ignore.case = TRUE)
  filename <- gsub(windows_trailing, replacement, filename)

  # TODO: this substr should really be unicode aware, so it doesn't chop a
  # multibyte code point in half.
  filename <- substr(filename, 1, 255)
  if (replacement == "") {
    return(filename)
  }
  path_sanitize(filename, "")
}

#' Construct a path to a location within an installed or development package
#'
#' `path_package` differs from [system.file()] in that it always returns an
#' error if the package does not exist. It also returns a different error if
#' the file within the package does not exist.
#'
#'`path_package()` also automatically works with packages loaded with devtools
#'even if the `path_package()` call comes from a different package.
#'
#' @param package Name of the package to in which to search
#' @param ... Additional paths appended to the package path by [path()].
#' @examples
#' path_package("base")
#' path_package("stats")
#' path_package("base", "INDEX")
#' path_package("splines", "help", "AnIndex")
#' @export
path_package <- function(package, ...) {
  if (!is.character(package) || length(package) != 1L) {
    stop(fs_error(sprintf("`package` must be a character vector of length 1")))
  }

  pkg_path <-
    tryCatch(
      find.package(package, quiet = FALSE),

      error = function(error) {
        is_not_found_error <- grepl(
          gettextf("there is no package called %s", sQuote(package)),
          conditionMessage(error), fixed = TRUE)
        if (!is_not_found_error) {
          stop(error)
        }

        msg <- sprintf(
"Can't find package `%s` in library locations:
%s",
    package,
    paste0("  - '", path_tidy(.libPaths()), "'", collapse = "\n"))

    stop(fs_error(msg, class = "EEXIST"))
    })


  files_inst <- path(pkg_path, "inst", ...)
  present_inst <- file_exists(files_inst)
  files_top <- path(pkg_path, ...)
  present_top <- file_exists(files_top)
  files <- files_top
  files[present_inst] <- files_inst[present_inst]
  files <- files[present_inst | present_top]

  if (length(files) == 0) {
    msg <- sprintf(
      "File(s) %s do not exist",
      paste0("'", files_top, "'", collapse = ", ")
    )
    stop(fs_error(msg, class = "EEXIST"))
  }
  files
}

context("test-path_package.R")

describe("path_package", {
  it("errors if the package does not exist", {
    expect_error(path_package("arst"), class = "EEXIST", msg = "Can't find package")
  })
  it("errors if the package location is not found", {
    expect_error(path_package("base", "foo"), class = "EEXIST", msg = "does not exist")
  })
  it("works on installed files", {
    expect_error_free(p <- path_package("base", "INDEX"))
    expect_is(p, "fs_path")
  })
})

---
title: "Comparison of fs functions, base R, and shell commands"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{base vs fs}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| label: setup
#| include: false
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Most of the functionality is **fs** can be approximated with functions in **base R**
or in a command line shell. The table at the end of this vignette can be used
as a translation aid between these three methods.

**fs** functions smooth over some of the idiosyncrasies of file handling with
base R functions:

* Vectorization. All **fs** functions are vectorized, accepting multiple paths
  as input. Base functions are inconsistently vectorized.

* Predictable return values that always convey a path. All **fs** functions
  return a character vector of paths, a named integer or a logical vector, where
  the names give the paths. Base return values are more varied: they are often
  logical or contain error codes which require downstream processing.

* Explicit failure. If **fs** operations fail, they throw an error. Base
  functions tend to generate a warning and a system dependent error code. This
  makes it easy to miss a failure.

* UTF-8 all the things. **fs** functions always convert input paths to UTF-8 and
  return results as UTF-8. This gives you path encoding consistency across OSes.
  Base functions rely on the native system encoding.

* Naming convention. **fs** functions use a consistent naming convention.
  Because base R's functions were gradually added over time there are a number
  of different conventions used (e.g. `path.expand()` vs `normalizePath()`;
  `Sys.chmod()` vs `file.access()`).

## Directory functions

| fs                                 | base                                                                    | shell                                |
| ---                                | ---                                                                     | ---                                  |
| `dir_ls("path")`                   | `list.files("path")`                                                    | `ls path`                            |
| `dir_info("path")`                 | `file.info(list.files("path", full.names = TRUE))`                      | `ls -al path`                        |
| `dir_copy("path", "new-path")`     | `dir.create("new-path"); file.copy("path", "new-path", recursive=TRUE)` | `cp path new-path`                   |
| `dir_create("path")`               | `dir.create("path")`                                                    | `mkdir path`                         |
| `dir_delete("path")`               | `unlink("path", recursive = TRUE)`                                      | `rm -rf path`                        |
| `dir_exists("path")`               | `dir.exists("path")`                                                    | `if [ -d "path" ]; then ... ; fi`    |
| ~~`dir_move()`~~ (see `file_move`) | `file.rename("path", "new-path")`                                       | `mv path new-path`                   |
| `dir_map("path", fun)`             | *No direct equivalent*                                                  | `for file in $(ls path); do ...; done` |
| `dir_tree("path")`                 | *No direct equivalent*                                                  | `tree path`                          |


## File functions

| fs                                          | base                              | shell                             |
| ---                                         | ---                               | ---                               |
| `file_chmod("path", "mode")`                | `Sys.chmod("path", "mode")`       | `chmod mode path`                 |
| `file_chown("path", "user_id", "group_id")` | *No direct equivalent*            | `chown options path `             |
| `file_copy("path", "new-path")`             | `file.copy("path", "new-path")`   | `cp path new-path`                |
| `file_create("new-path")`                   | `file.create("new-path")`         | `touch new-path`                  |
| `file_delete("path")`                       | `unlink("path")`                  | `rm path`                         |
| `file_exists("path")`                       | `file.exists("path")`             | `if [ -f "path" ]; then ... ; fi` |
| `file_info("path")`                         | `file.info("path")`               | `ls -al path`                     |
| `file_move("path", "new-path")`             | `file.rename("path", "new-path")` | `mv path new-path`                |
| `file_show("path")`                         | `browseURL("path")`               | `open path`                       |
| `file_touch()`                              | *No direct equivalent*            | `touch path`                      |
| `file_temp()`                               | `tempfile()`                      | `mktemp`                          |
|  *No direct equivalent*                     | `file_test()`                     | `if [ -d "path" ]; then ...; fi`  |

## Path functions

| fs                                                   | base                                              | shell                         |
| ---                                                  | ---                                               | ---                           |
| `path("top_dir", "nested_dir", "file", ext = "ext")` | `file.path("top_dir", "nested_dir", "file.ext")`  | `top_dir/nested_dir/file.ext` |
| `path_temp()`, `path_temp("path")`                   | `tempdir()`, `file.path(tempdir(), "path")`       | `mktemp -d`                   |
| `path_expand("~/path")`                              | `path.expand()`                                   | `realpath -m -s ~/path`       |
| `path_dir("path")`                                   | `dirname("path")`                                 | `dirname path`                |
| `path_file("path")`                                  | `basename("path")`                                | `basename path`               |
| `path_home()`                                        | `path.expand("~")`                                | `$HOME`                       |
| `path_package("pkgname", "dir", "file")`             | `system.file("dir", "file", package = "pkgname")` | *No direct equivalent*        |
| `path_norm("path")`                                  | `normalizePath()`                                 | `realpath`                    |
| `path_real("path")`                                  | `normalizePath(mustWork = TRUE)`                  | `realpath`                    |
| `path_rel("path/foo", "path/bar")`                   | *No direct equivalent*                            | *No direct equivalent*        |
| `path_common(c("path/foo", "path/bar", "path/baz"))` | *No direct equivalent*                            | *No direct equivalent*        |
| `path_ext_remove("path")`                            | `sub("\\.[a-zA-Z0-9]*$", "", "path")`             | *No direct equivalent*        |
| `path_ext_set("path", "new_ext")`                    | `sub("\\.[a-zA-Z0-9]*$", "new_ext", "path")`      | *No direct equivalent*        |
| `path_sanitize("path")`                              | *No direct equivalent*                            | *No direct equivalent*        |

# dir_tree() works

    Code
      dir_tree("foo")
    Output
      foo
      \-- 1

---

    Code
      dir_tree("foo")
    Output
      foo
      +-- 1
      +-- 2
      +-- 3
      \-- 4

---

    Code
      dir_tree("foo")
    Output
      foo
      +-- 4
      +-- bar
      |   +-- 1
      |   \-- baz
      |       \-- 2
      \-- qux
          \-- 3

# Contributing to PyFilesystem

Pull Requests are very welcome for this project!

For bug fixes or new features, please file an issue before submitting a pull
request. If the change isn't trivial, it may be best to wait for feedback.
For a quicker response, contact [Will McGugan](mailto:willmcgugan+pyfs@gmail.com)
directly.


## `tox`

Most of the guidelines that follow can be checked with a particular
[`tox`](https://pypi.org/project/tox/) environment. Having it installed will
help you develop and verify your code locally without having to wait for
our Continuous Integration pipeline to finish.


## Tests

New code should have unit tests. We strive to have near 100% coverage.
Get in touch, if you need assistance with the tests. You shouldn't refrain
from opening a Pull Request even if all the tests were not added yet, or if
not all of them are passing yet.

### Dependencies

The dependency for running the tests can be found in the `tests/requirements.txt` file.
If you're using `tox`, you won't have to install them manually. Otherwise,
they can be installed with `pip`:
```console
$ pip install -r tests/requirements.txt
```

### Running (with `tox`)

Simply run in the repository folder to execute the tests for all available
environments:
```console
$ tox
```

Since this can take some time, you can use a single environment to run
tests only once, for instance to run tests only with Python 3.9:
```console
$ tox -e py39
```

### Running (without `tox`)

Tests are written using the standard [`unittest`](https://docs.python.org/3/library/unittest.html)
framework. You should be able to run them using the standard library runner:
```console
$ python -m unittest discover -vv
```


## Coding Guidelines

This project runs on Python2.7 and Python3.X. Python2.7 will be dropped at
some point, but for now, please maintain compatibility. PyFilesystem2 uses
the [`six`](https://pypi.org/project/six/) library to write version-agnostic
Python code.

### Style

The code (including the tests) should follow PEP8. You can check for the
code style with:
```console
$ tox -e codestyle
```

This will invoke [`flake8`](https://pypi.org/project/flake8/) with some common
plugins such as [`flake8-comprehensions`](https://pypi.org/project/flake8-comprehensions/).

### Format

Please format new code with [black](https://github.com/ambv/black), using the
default settings. You can check whether the code is well-formatted with:
```console
$ tox -e codeformat
```

### Type annotations

The code is typechecked with [`mypy`](https://pypi.org/project/mypy/), and
type annotations written as comments, to stay compatible with Python2. Run
the typechecking with:
```console
$ tox -e typecheck
```


## Documentation

### Dependencies

The documentation is built with [Sphinx](https://pypi.org/project/Sphinx/),
using the [ReadTheDocs](https://pypi.org/project/sphinx-rtd-theme/) theme.
The dependencies are listed in `docs/requirements.txt` and can be installed with
`pip`:
```console
$ pip install -r docs/requirements.txt
```

### Building

Run the following command to build the HTML documentation:
```console
$ python setup.py build_sphinx
```

The documentation index will be written to the `build/sphinx/html/`
directory.

### Style

The API reference is written in the Python source, using docstrings in
[Google format](https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html).
The documentation style can be checked with:
```console
$ tox -e docstyle
```

# Contributors (sorted alphabetically)

Many thanks to the following developers for contributing to this project:

- [Adrian Garcia Badaracco](https://github.com/adriangb)
- [Alex Povel](https://github.com/alexpovel)
- [Andreas Tollkötter](https://github.com/atollk)
- [Andrew Scheller](https://github.com/lurch)
- [Andrey Serov](https://github.com/zmej-serow)
- [Ben Lindsay](https://github.com/benlindsay)
- [Bernhard M. Wiedemann](https://github.com/bmwiedemann)
- [@chfw](https://github.com/chfw)
- [Dafna Hirschfeld](https://github.com/kamomil)
- [Diego Argueta](https://github.com/dargueta)
- [Eelke van den Bos](https://github.com/eelkevdbos)
- [Egor Namakonov](https://github.com/fresheed)
- [@FooBarQuaxx](https://github.com/FooBarQuaxx)
- [Geoff Jukes](https://github.com/geoffjukes)
- [George Macon](https://github.com/gmacon)
- [Giampaolo Cimino](https://github.com/gpcimino)
- [@Hoboneer](https://github.com/Hoboneer)
- [Jon Hagg](https://github.com/jon-hagg)
- [Joseph Atkins-Turkish](https://github.com/Spacerat)
- [Joshua Tauberer](https://github.com/JoshData)
- [Justin Charlong](https://github.com/jcharlong)
- [Louis Sautier](https://github.com/sbraz)
- [Martin Durant](https://github.com/martindurant)
- [Martin Larralde](https://github.com/althonos)
- [Masaya Nakamura](https://github.com/mashabow)
- [Morten Engelhardt Olsen](https://github.com/xoriath)
- [@mrg0029](https://github.com/mrg0029)
- [Nathan Goldbaum](https://github.com/ngoldbaum)
- [Nick Henderson](https://github.com/nwh)
- [Oliver Galvin](https://github.com/odgalvin)
- [Philipp Wiesner](https://github.com/birnbaum)
- [Philippe Ombredanne](https://github.com/pombredanne)
- [Rehan Khwaja](https://github.com/rkhwaja)
- [Silvan Spross](https://github.com/sspross)
- [@sqwishy](https://github.com/sqwishy)
- [Sven Schliesing](https://github.com/muffl0n)
- [Tim Gates](https://github.com/timgates42/)
- [@tkossak](https://github.com/tkossak)
- [Todd Levi](https://github.com/televi)
- [Vilius Grigaliūnas](https://github.com/vilius-g)
- [Will McGugan](https://github.com/willmcgugan)
- [Zmej Serow](https://github.com/zmej-serow)

---
title: "Comparison of fs functions, base R, and shell commands"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{base vs fs}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Most of the functionality is **fs** can be approximated with functions in **base R**
or in a command line shell. The table at the end of this vignette can be used
as a translation aid between these three methods.

**fs** functions smooth over some of the idiosyncrasies of file handling with
base R functions:

* Vectorization. All **fs** functions are vectorized, accepting multiple paths
  as input. Base functions are inconsistently vectorized.

* Predictable return values that always convey a path. All **fs** functions
  return a character vector of paths, a named integer or a logical vector, where
  the names give the paths. Base return values are more varied: they are often
  logical or contain error codes which require downstream processing.

* Explicit failure. If **fs** operations fail, they throw an error. Base
  functions tend to generate a warning and a system dependent error code. This
  makes it easy to miss a failure.

* UTF-8 all the things. **fs** functions always convert input paths to UTF-8 and
  return results as UTF-8. This gives you path encoding consistency across OSes.
  Base functions rely on the native system encoding.

* Naming convention. **fs** functions use a consistent naming convention.
  Because base R's functions were gradually added over time there are a number
  of different conventions used (e.g. `path.expand()` vs `normalizePath()`;
  `Sys.chmod()` vs `file.access()`).

## Directory functions

| fs                                 | base                                                                    | shell                                |
| ---                                | ---                                                                     | ---                                  |
| `dir_ls("path")`                   | `list.files("path")`                                                    | `ls path`                            |
| `dir_info("path")`                 | `do.call(rbind, lapply(list.files("path"), file.info))`                 | `ls -al path`                        |
| `dir_copy("path", "new-path")`     | `dir.create("new-path"); file.copy("path", "new-path", recursive=TRUE)` | `cp path new-path`                   |
| `dir_create("path")`               | `dir.create("path")`                                                    | `mkdir path`                         |
| `dir_delete("path")`               | `unlink("path", recursive = TRUE)`                                      | `rm -rf path`                        |
| `dir_exists("path")`               | `dir.exists("path")`                                                    | `if [ -d "path" ]; then ... ; fi`    |
| ~~`dir_move()`~~ (see `file_move`) | `file.rename("path", "new-path")`                                       | `mv path new-path`                   |
| `dir_map("path", fun)`             | *No direct equivalent*                                                  | `for file in $(ls path); do ...; done` |
| `dir_tree("path")`                 | *No direct equivalent*                                                  | `tree path`                          |


## File functions

| fs                                          | base                              | shell                             |
| ---                                         | ---                               | ---                               |
| `file_chmod("path", "mode")`                | `Sys.chmod("path", "mode")`       | `chmod mode path`                 |
| `file_chown("path", "user_id", "group_id")` | *No direct equivalent*            | `chown options path `             |
| `file_copy("path", "new-path")`             | `file.copy("path", "new-path")`   | `cp path new-path`                |
| `file_create("new-path")`                   | `file.create("new-path")`         | `touch new-path`                  |
| `file_delete("path")`                       | `unlink("path")`                  | `rm path`                         |
| `file_exists("path")`                       | `file.exists("path")`             | `if [ -f "path" ]; then ... ; fi` |
| `file_info("path")`                         | `file.info("path")`               | `ls -al path`                     |
| `file_move("path", "new-path")`             | `file.rename("path", "new-path")` | `mv path new-path`                |
| `file_show("path")`                         | `browseURL("path")`               | `open path`                       |
| `file_touch()`                              | *No direct equivalent*            | `touch path`                      |
| `file_temp()`                               | `tempfile()`                      | `mktemp`                          |
| `file_test()`                               | *No direct equivalent*            | `if [ -d "path" ]; then ...; fi`  |

## Path functions

| fs                                                   | base                                              | shell                         |
| ---                                                  | ---                                               | ---                           |
| `path("top_dir", "nested_dir", "file", ext = "ext")` | `file.path("top_dir", "nested_dir", "file.ext")`  | `top_dir/nested_dir/file.ext` |
| `path_temp()`, `path_temp("path")`                   | `tempdir()`, `file.path(tempdir(), "path")`       | `mktemp -d`                   |
| `path_expand("~/path")`                              | `path.expand()`                                   | `realpath -m -s ~/path`       |
| `path_dir("path")`                                   | `dirname("path")`                                 | `dirname path`                |
| `path_file("path")`                                  | `basename("path")`                                | `basename path`               |
| `path_home()`                                        | `path.expand("~")`                                | `$HOME`                       |
| `path_package("pkgname", "dir", "file")`             | `system.file("dir", "file", package = "pkgname")` | *No direct equivalent*        |
| `path_norm("path")`                                  | `normalizePath()`                                 | `realpath`                    |
| `path_real("path")`                                  | `normalizePath(mustWork = TRUE)`                  | `realpath`                    |
| `path_rel("path/foo", "path/bar")`                   | *No direct equivalent*                            | *No direct equivalent*        |
| `path_common(c("path/foo", "path/bar", "path/baz"))` | *No direct equivalent*                            | *No direct equivalent*        |
| `path_ext_remove("path")`                            | `sub("\\.[a-zA-Z0-9]*$", "", "path")`             | *No direct equivalent*        |
| `path_ext_set("path", "new_ext")`                    | `sub("\\.[a-zA-Z0-9]*$", "new_ext", "path")`      | *No direct equivalent*        |
| `path_sanitize("path")`                              | *No direct equivalent*                            | *No direct equivalent*        |

# as_fs_path / fails with non-character inputs

    Code
      as_fs_path(1)
    Condition
      Error in `UseMethod()`:
      ! no applicable method for 'as_fs_path' applied to an object of class "c('double', 'numeric')"
    Code
      as_fs_path(TRUE)
    Condition
      Error in `UseMethod()`:
      ! no applicable method for 'as_fs_path' applied to an object of class "logical"

# path / errors on paths which are too long

    Code
      path(paste(rep("a", 1e+05), collapse = ""))
    Condition
      Error in `path_tidy()`:
      ! Total path length must be less than PATH_MAX: <path-max>
    Code
      do.call(path, as.list(rep("a", 1e+05)))
    Condition
      Error in `path_tidy()`:
      ! Total path length must be less than PATH_MAX: <path-max>

# path_rel / works for POSIX paths

    Code
      path_rel(c("a", "a/b", "a/b/c"), c("a/b", "a"))
    Condition
      Error:
      ! `start` must be a single path to a starting directory.

# vectorized

    Code
      path_select_components(path, 1:3)
    Output
      some/simple/path another/path/to  
    Code
      path_select_components(path, 1:3, "end")
    Output
      to/a/file.txt     path/to/file2.txt 
    Code
      path_select_components(path, integer())
    Output
        
    Code
      unclass(path_select_components(path, integer()))
    Output
      [1] "" ""
    Code
      path_select_components(fs_path(character()), 1:3)
    Output
      character(0)
    Code
      path_select_components(fs_path(character()), 1:3, "end")
    Output
      character(0)
    Code
      class(path_select_components(fs_path(character()), 1:3))
    Output
      [1] "fs_path"   "character"
    Code
      class(path_select_components(fs_path(character()), 1:3, "end"))
    Output
      [1] "fs_path"   "character"

# Ops.fs_bytes / errors for unary operators

    Code
      !x
    Condition
      Error:
      ! unary '!' not defined for "fs_bytes" objects
    Code
      +x
    Condition
      Error:
      ! unary '+' not defined for "fs_bytes" objects
    Code
      -x
    Condition
      Error:
      ! unary '-' not defined for "fs_bytes" objects

# Ops.fs_bytes / errors for other binary operators

    Code
      x %% 2
    Condition
      Error:
      ! '%%' not defined for "fs_bytes" objects
    Code
      x %/% 2
    Condition
      Error:
      ! '%/%' not defined for "fs_bytes" objects
    Code
      x & TRUE
    Condition
      Error:
      ! '&' not defined for "fs_bytes" objects
    Code
      x | TRUE
    Condition
      Error:
      ! '|' not defined for "fs_bytes" objects

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path.R
\name{path_expand}
\alias{path_expand}
\alias{path_expand_r}
\alias{path_home}
\alias{path_home_r}
\title{Finding the User Home Directory}
\usage{
path_expand(path)

path_expand_r(path)

path_home(...)

path_home_r(...)
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{...}{Additional paths appended to the home directory by \code{\link[=path]{path()}}.}
}
\description{
\itemize{
\item \code{path_expand()} performs tilde expansion on a path, replacing instances of
\code{~} or \code{~user} with the user's home directory.
\item \code{path_home()} constructs a path within the expanded users home directory,
calling it with \emph{no} arguments can be useful to verify what fs considers the
home directory.
\item \code{path_expand_r()} and \code{path_home_r()} are equivalents which always use R's
definition of the home directory.
}
}
\details{
\code{path_expand()} differs from \code{\link[base:path.expand]{base::path.expand()}} in the interpretation of
the home directory of Windows. In particular \code{path_expand()} uses the path
set in the \code{USERPROFILE} environment variable and, if unset, then uses
\code{HOMEDRIVE}/\code{HOMEPATH}.

In contrast \code{\link[base:path.expand]{base::path.expand()}} first checks for \code{R_USER} then \code{HOME},
which in the default configuration of R on Windows are both set to the user's
document directory, e.g. \verb{C:\\\\Users\\\\username\\\\Documents}.
\code{\link[base:path.expand]{base::path.expand()}} also does not support \code{~otheruser} syntax on Windows,
whereas \code{path_expand()} does support this syntax on all systems.

This definition makes fs more consistent with the definition of home
directory used on Windows in other languages, such as
\href{https://docs.python.org/3/library/os.path.html#os.path.expanduser}{python}
and \href{https://doc.rust-lang.org/std/env/fn.home_dir.html#windows}{rust}. This
is also more compatible with external tools such as git and ssh, both of
which put user-level files in \code{USERPROFILE} by default. It also allows you to
write portable paths, such as \verb{~/Desktop} that points to the Desktop location
on Windows, macOS and (most) Linux systems.

Users can set the \code{R_FS_HOME} environment variable to override the
definitions on any platform.
}
\examples{
# Expand a path
path_expand("~/bin")

# You can use `path_home()` without arguments to see what is being used as
# the home diretory.
path_home()
path_home("R")

# This will likely differ from the above on Windows
path_home_r()
}
\seealso{
\href{https://cran.r-project.org/bin/windows/base/rw-FAQ.html#What-are-HOME-and-working-directories_003f}{R for Windows FAQ - 2.14}
for behavior of \code{\link[base:path.expand]{base::path.expand()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.R
\name{is_file}
\alias{is_file}
\alias{is_dir}
\alias{is_link}
\alias{is_file_empty}
\title{Functions to test for file types}
\usage{
is_file(path, follow = TRUE)

is_dir(path, follow = TRUE)

is_link(path)

is_file_empty(path, follow = TRUE)
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{follow}{If \code{TRUE}, symbolic links will be followed (recursively) and
the results will be that of the final file rather than the link.}
}
\value{
A named logical vector, where the names give the paths. If the given
object does not exist, \code{NA} is returned.
}
\description{
Functions to test for file types
}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
dir_create("d")

file_create("d/file.txt")
dir_create("d/dir")
link_create(path(path_abs("d"), "file.txt"), "d/link")

paths <- dir_ls("d")
is_file(paths)
is_dir(paths)
is_link(paths)

# Cleanup
dir_delete("d")
\dontshow{setwd(.old_wd)}
}
\seealso{
\code{\link[=file_exists]{file_exists()}}, \code{\link[=dir_exists]{dir_exists()}} and \code{\link[=link_exists]{link_exists()}} if you want
to ensure that the path also exists.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fs_bytes.R
\name{fs_bytes}
\alias{fs_bytes}
\alias{as_fs_bytes}
\title{Human readable file sizes}
\usage{
as_fs_bytes(x)

fs_bytes(x)
}
\arguments{
\item{x}{A numeric or character vector. Character representations can use
shorthand sizes (see examples).}
}
\description{
Construct, manipulate and display vectors of file sizes. These are numeric
vectors, so you can compare them numerically, but they can also be compared
to human readable values such as '10MB'.
}
\examples{
fs_bytes("1")
fs_bytes("1K")
fs_bytes("1Kb")
fs_bytes("1Kib")
fs_bytes("1MB")

fs_bytes("1KB") < "1MB"

sum(fs_bytes(c("1MB", "5MB", "500KB")))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/id.R
\name{id}
\alias{id}
\alias{group_ids}
\alias{user_ids}
\title{Lookup Users and Groups on a system}
\usage{
group_ids()

user_ids()
}
\value{
They return their results in a \code{data.frame}. On windows both
functions return an empty \code{data.frame} because windows does not have user
or group ids.
}
\description{
These functions use the GETPWENT(3) and GETGRENT(3) system calls to query
users and groups respectively.
}
\examples{
# list first 6 groups
head(group_ids())

# list first 6 users
head(user_ids())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path_select.R
\name{path_select_components}
\alias{path_select_components}
\title{Select path components by their position/index.}
\usage{
path_select_components(path, index, from = c("start", "end"))
}
\arguments{
\item{path}{A path of class \code{fs_path}.}

\item{index}{An integer vector of path positions. (A negative index will work
according to R's usual subsetting rules.)}

\item{from}{A character of either \code{"start"} or \code{"end"} to choose if indexing
should start from the first or last component of the \code{path}.}
}
\value{
An \code{fs_path} object, which is a character vector that also has class
\code{fs_path}.
}
\description{
\code{path_select_components()} allows to select individual components from
an \code{fs_path} object via their index.
}
\details{
\code{path_select_components()} is vectorized.
}
\examples{
path <- fs::path("some", "simple", "path", "to", "a", "file.txt")

path_select_components(path, 1:3)

path_select_components(path, 1:3, "end")

path_select_components(path, -1, "end")

path_select_components(path, 6)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path.R
\name{path_math}
\alias{path_math}
\alias{path_real}
\alias{path_split}
\alias{path_join}
\alias{path_abs}
\alias{path_norm}
\alias{path_rel}
\alias{path_common}
\alias{path_has_parent}
\title{Path computations}
\usage{
path_real(path)

path_split(path)

path_join(parts)

path_abs(path, start = ".")

path_norm(path)

path_rel(path, start = ".")

path_common(path)

path_has_parent(path, parent)
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{parts}{A character vector or a list of character vectors, corresponding
to split paths.}

\item{start}{A starting directory to compute the path relative to.}

\item{parent}{The parent path.}
}
\value{
The new path(s) in an \code{fs_path} object, which is a character vector
that also has class \code{fs_path}. Except \code{path_split()}, which returns a list
of character vectors of path components.
}
\description{
All functions apart from \code{path_real()} are purely path computations, so the
files in question do not need to exist on the filesystem.
}
\section{Functions}{
\itemize{
\item \code{path_real}: returns the canonical path, eliminating any symbolic
links and the special references \code{~}, \code{~user}, \code{.}, and \code{..}, , i.e. it
calls \code{path_expand()} (literally) and \code{path_norm()} (effectively).

\item \code{path_split}: splits paths into parts.

\item \code{path_join}: joins parts together. The inverse of \code{\link[=path_split]{path_split()}}.
See \code{\link[=path]{path()}} to concatenate vectorized strings into a path.

\item \code{path_abs}: returns a normalized, absolute version of a path.

\item \code{path_norm}: eliminates \code{.} references and rationalizes up-level
\code{..} references, so \code{A/./B} and \code{A/foo/../B} both become \code{A/B}, but \code{../B}
is not changed. If one of the paths is a symbolic link, this may change the
meaning of the path, so consider using \code{path_real()} instead.

\item \code{path_rel}: computes the path relative to the \code{start} path,
which can be either an absolute or relative path.

\item \code{path_common}: finds the common parts of two (or more) paths.

\item \code{path_has_parent}: determine if a path has a given parent.
}}

\examples{
\dontshow{.old_wd <- setwd(tempdir())}
dir_create("a")
file_create("a/b")
link_create(path_abs("a"), "c")

# Realize the path
path_real("c/b")

# Split a path
parts <- path_split("a/b")
parts

# Join it together
path_join(parts)

# Find the absolute path
path_abs("..")

# Normalize a path
path_norm("a/../b\\\\c/.")

# Compute a relative path
path_rel("/foo/abc", "/foo/bar/baz")

# Find the common path between multiple paths
path_common(c("/foo/bar/baz", "/foo/bar/abc", "/foo/xyz/123"))

# Cleanup
dir_delete("a")
link_delete("c")
\dontshow{setwd(.old_wd)}
}
\seealso{
\code{\link[=path_expand]{path_expand()}} for expansion of user's home directory.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path.R
\name{path_math}
\alias{path_math}
\alias{path_real}
\alias{path_split}
\alias{path_join}
\alias{path_abs}
\alias{path_norm}
\alias{path_rel}
\alias{path_common}
\alias{path_has_parent}
\title{Path computations}
\usage{
path_real(path)

path_split(path)

path_join(parts)

path_abs(path, start = ".")

path_norm(path)

path_rel(path, start = ".")

path_common(path)

path_has_parent(path, parent)
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{parts}{A character vector or a list of character vectors, corresponding
to split paths.}

\item{start}{A starting directory to compute the path relative to.}

\item{parent}{The parent path.}
}
\value{
The new path(s) in an \code{fs_path} object, which is a character vector
that also has class \code{fs_path}. Except \code{path_split()}, which returns a list
of character vectors of path components.
}
\description{
All functions apart from \code{path_real()} are purely path computations, so the
files in question do not need to exist on the filesystem.
}
\section{Functions}{
\itemize{
\item \code{path_real()}: returns the canonical path, eliminating any symbolic
links and the special references \code{~}, \code{~user}, \code{.}, and \code{..}, , i.e. it
calls \code{path_expand()} (literally) and \code{path_norm()} (effectively).

\item \code{path_split()}: splits paths into parts.

\item \code{path_join()}: joins parts together. The inverse of \code{\link[=path_split]{path_split()}}.
See \code{\link[=path]{path()}} to concatenate vectorized strings into a path.

\item \code{path_abs()}: returns a normalized, absolute version of a path.

\item \code{path_norm()}: eliminates \code{.} references and rationalizes up-level
\code{..} references, so \code{A/./B} and \code{A/foo/../B} both become \code{A/B}, but \code{../B}
is not changed. If one of the paths is a symbolic link, this may change the
meaning of the path, so consider using \code{path_real()} instead.

\item \code{path_rel()}: computes the path relative to the \code{start} path,
which can be either an absolute or relative path.

\item \code{path_common()}: finds the common parts of two (or more) paths.

\item \code{path_has_parent()}: determine if a path has a given parent.

}}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
dir_create("a")
file_create("a/b")
link_create(path_abs("a"), "c")

# Realize the path
path_real("c/b")

# Split a path
parts <- path_split("a/b")
parts

# Join it together
path_join(parts)

# Find the absolute path
path_abs("..")

# Normalize a path
path_norm("a/../b\\\\c/.")

# Compute a relative path
path_rel("/foo/abc", "/foo/bar/baz")

# Find the common path between multiple paths
path_common(c("/foo/bar/baz", "/foo/bar/abc", "/foo/xyz/123"))

# Cleanup
dir_delete("a")
link_delete("c")
\dontshow{setwd(.old_wd)}
}
\seealso{
\code{\link[=path_expand]{path_expand()}} for expansion of user's home directory.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/file.R
\name{file_move}
\alias{file_move}
\title{Move or rename files}
\usage{
file_move(path, new_path)
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{new_path}{New file path. If \code{new_path} is existing directory, the file
will be moved into that directory; otherwise it will be moved/renamed to
the full path.

Should either be the same length as \code{path}, or a single directory.}
}
\value{
The new path (invisibly).
}
\description{
Compared to \link{file.rename} \code{file_move()} always fails if it is unable to move
a file, rather than signaling a Warning and returning an error code.
}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
file_create("foo")
file_move("foo", "bar")
file_exists(c("foo", "bar"))
file_delete("bar")
\dontshow{setwd(.old_wd)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/file.R
\name{file_chmod}
\alias{file_chmod}
\title{Change file permissions}
\usage{
file_chmod(path, mode)
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{mode}{A character representation of the mode, in either hexidecimal or symbolic format.}
}
\description{
Change file permissions
}
\details{
\strong{Cross-compatibility warning:} File permissions differ on Windows
from POSIX systems. Windows does not use an executable bit, so attempting
to change this will have no effect. Windows also does not have user
groups, so only the user permissions (\code{u}) are relevant.
}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
file_create("foo", mode = "000")
file_chmod("foo", "777")
file_info("foo")$permissions

file_chmod("foo", "u-x")
file_info("foo")$permissions

file_chmod("foo", "a-wrx")
file_info("foo")$permissions

file_chmod("foo", "u+wr")
file_info("foo")$permissions

# It is also vectorized
files <- c("foo", file_create("bar", mode = "000"))
file_chmod(files, "a+rwx")
file_info(files)$permissions

file_chmod(files, c("644", "600"))
file_info(files)$permissions
\dontshow{setwd(.old_wd)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.R
\name{is_absolute_path}
\alias{is_absolute_path}
\title{Test if a path is an absolute path}
\usage{
is_absolute_path(path)
}
\arguments{
\item{path}{A character vector of one or more paths.}
}
\description{
Test if a path is an absolute path
}
\examples{
is_absolute_path("/foo")
is_absolute_path("C:\\\\foo")
is_absolute_path("\\\\\\\\myserver\\\\foo\\\\bar")

is_absolute_path("foo/bar")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/file.R
\name{file_chown}
\alias{file_chown}
\title{Change owner or group of a file}
\usage{
file_chown(path, user_id = NULL, group_id = NULL)
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{user_id}{The user id of the new owner, specified as a numeric ID or
name. The R process must be privileged to change this.}

\item{group_id}{The group id of the new owner, specified as a numeric ID or
name.}
}
\description{
Change owner or group of a file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sanitize.R
\name{path_sanitize}
\alias{path_sanitize}
\title{Sanitize a filename by removing directory paths and invalid characters}
\usage{
path_sanitize(filename, replacement = "")
}
\arguments{
\item{filename}{A character vector to be sanitized.}

\item{replacement}{A character vector used to replace invalid characters.}
}
\description{
\code{path_sanitize()} removes the following:
\itemize{
\item \href{https://en.wikipedia.org/wiki/C0_and_C1_control_codes}{Control characters}
\item \href{https://web.archive.org/web/20230126161942/https://kb.acronis.com/content/39790}{Reserved characters}
\item Unix reserved filenames (\code{.} and \code{..})
\item Trailing periods and spaces (invalid on Windows)
\item Windows reserved filenames (\code{CON}, \code{PRN}, \code{AUX}, \code{NUL}, \code{COM1}, \code{COM2},
\code{COM3}, COM4, \code{COM5}, \code{COM6}, \code{COM7}, \code{COM8}, \code{COM9}, \code{LPT1}, \code{LPT2},
\code{LPT3}, \code{LPT4}, \code{LPT5}, \code{LPT6}, LPT7, \code{LPT8}, and \code{LPT9})
The resulting string is then truncated to \href{https://en.wikipedia.org/wiki/Comparison_of_file_systems#Limits}{255 bytes in length}
}
}
\examples{
# potentially unsafe string
str <- "~/.\u0001ssh/authorized_keys"
path_sanitize(str)

path_sanitize("..")
}
\seealso{
\url{https://www.npmjs.com/package/sanitize-filename}, upon which this
function is based.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sanitize.R
\name{path_sanitize}
\alias{path_sanitize}
\title{Sanitize a filename by removing directory paths and invalid characters}
\usage{
path_sanitize(filename, replacement = "")
}
\arguments{
\item{filename}{A character vector to be sanitized.}

\item{replacement}{A character vector used to replace invalid characters.}
}
\description{
\code{path_sanitize()} removes the following:
\itemize{
\item \href{https://en.wikipedia.org/wiki/C0_and_C1_control_codes}{Control characters}
\item \href{https://kb.acronis.com/content/39790}{Reserved characters}
\item Unix reserved filenames (\code{.} and \code{..})
\item Trailing periods and spaces (invalid on Windows)
\item Windows reserved filenames (\code{CON}, \code{PRN}, \code{AUX}, \code{NUL}, \code{COM1}, \code{COM2},
\code{COM3}, COM4, \code{COM5}, \code{COM6}, \code{COM7}, \code{COM8}, \code{COM9}, \code{LPT1}, \code{LPT2},
\code{LPT3}, \code{LPT4}, \code{LPT5}, \code{LPT6}, LPT7, \code{LPT8}, and \code{LPT9})
The resulting string is then truncated to \href{https://en.wikipedia.org/wiki/Comparison_of_file_systems#Limits}{255 bytes in length}
}
}
\examples{
# potentially unsafe string
str <- "~/.\u0001ssh/authorized_keys"
path_sanitize(str)

path_sanitize("..")
}
\seealso{
\url{https://www.npmjs.com/package/sanitize-filename}, upon which this
function is based.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/link.R
\name{link_path}
\alias{link_path}
\title{Read the value of a symbolic link}
\usage{
link_path(path)
}
\arguments{
\item{path}{A character vector of one or more paths.}
}
\value{
A tidy path to the object the link points to.
}
\description{
Read the value of a symbolic link
}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
file_create("foo")
link_create(path_abs("foo"), "bar")
link_path("bar")

# Cleanup
file_delete(c("foo", "bar"))
\dontshow{setwd(.old_wd)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fs_path.R
\name{fs_path}
\alias{fs_path}
\alias{as_fs_path}
\title{File paths}
\usage{
as_fs_path(x)

fs_path(x)
}
\arguments{
\item{x}{vector to be coerced to a fs_path object.}
}
\description{
Tidy file paths, character vectors which are coloured by file type on
capable terminals.

Colouring can be customized by setting the \code{LS_COLORS} environment variable,
the format is the same as that read by GNU ls / dircolors.

Colouring of file paths can be disabled by setting \code{LS_COLORS} to an empty
string e.g. \code{Sys.setenv(LS_COLORS = "")}.
}
\seealso{
\url{https://geoff.greer.fm/lscolors},
\url{https://github.com/trapd00r/LS_COLORS},
\url{https://github.com/seebi/dircolors-solarized} for some example colour
settings.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/file.R
\name{file_info}
\alias{file_info}
\alias{file_size}
\title{Query file metadata}
\usage{
file_info(path, fail = TRUE, follow = FALSE)

file_size(path, fail = TRUE)
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{fail}{Should the call fail (the default) or warn if a file cannot be
accessed.}

\item{follow}{If \code{TRUE}, symbolic links will be followed (recursively) and
the results will be that of the final file rather than the link.}
}
\value{
A data.frame with metadata for each file. Columns returned are as follows.
\item{path}{The input path, as a \code{\link[=fs_path]{fs_path()}} character vector.}
\item{type}{The file type, as a factor of file types.}
\item{size}{The file size, as a \code{\link[=fs_bytes]{fs_bytes()}} numeric vector.}
\item{permissions}{The file permissions, as a \code{\link[=fs_perms]{fs_perms()}} integer vector.}
\item{modification_time}{The time of last data modification, as a \link{POSIXct} datetime.}
\item{user}{The file owner name - as a character vector.}
\item{group}{The file group name - as a character vector.}
\item{device_id}{The file device id - as a numeric vector.}
\item{hard_links}{The number of hard links to the file - as a numeric vector.}
\item{special_device_id}{The special device id of the file - as a numeric vector.}
\item{inode}{The inode of the file - as a numeric vector.}
\item{block_size}{The optimal block for the file - as a numeric vector.}
\item{blocks}{The number of blocks allocated for the file - as a numeric vector.}
\item{flags}{The user defined flags for the file - as an integer vector.}
\item{generation}{The generation number for the file - as a numeric vector.}
\item{access_time}{The time of last access - as a \link{POSIXct} datetime.}
\item{change_time}{The time of last file status change - as a \link{POSIXct} datetime.}
\item{birth_time}{The time when the inode was created - as a \link{POSIXct} datetime.}
}
\description{
Compared to \verb{[file.info]} the full results of a \code{stat(2)} system call are
returned and some columns are returned as S3 classes to make manipulation
more natural. On systems which do not support all metadata (such as Windows)
default values are used.
}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
write.csv(mtcars, "mtcars.csv")
file_info("mtcars.csv")

# Files in the working directory modified more than 20 days ago
files <- file_info(dir_ls())
files$path[difftime(Sys.time(), files$modification_time, units = "days") > 20]

# Cleanup
file_delete("mtcars.csv")
\dontshow{setwd(.old_wd)}
}
\seealso{
\code{\link[=dir_info]{dir_info()}} to display file information for files in a given
directory.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path_package.R
\name{path_package}
\alias{path_package}
\title{Construct a path to a location within an installed or development package}
\usage{
path_package(package, ...)
}
\arguments{
\item{package}{Name of the package to in which to search}

\item{...}{Additional paths appended to the package path by \code{\link[=path]{path()}}.}
}
\description{
\code{path_package} differs from \code{\link[=system.file]{system.file()}} in that it always returns an
error if the package does not exist. It also returns a different error if
the file within the package does not exist.
}
\details{
\code{path_package()} also automatically works with packages loaded with devtools
even if the \code{path_package()} call comes from a different package.
}
\examples{
path_package("base")
path_package("stats")
path_package("base", "INDEX")
path_package("splines", "help", "AnIndex")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fs_perms.R
\name{fs_perms}
\alias{fs_perms}
\alias{as_fs_perms}
\title{Create, modify and view file permissions}
\usage{
as_fs_perms(x, ...)

fs_perms(x, ...)
}
\arguments{
\item{x}{An object which is to be coerced to a fs_perms object. Can be an
number or octal character representation, including symbolic
representations.}

\item{...}{Additional arguments passed to methods.}
}
\description{
\code{fs_perms()} objects help one create and modify file permissions easily.
They support both numeric input, octal and symbolic character
representations. Compared to \link{octmode} they support symbolic representations
and display the mode the same format as \code{ls} on POSIX systems.
}
\details{
On POSIX systems the permissions are displayed as a 9 character string with
three sets of three characters. Each set corresponds to the permissions for
the user, the group and other (or default) users.

If the first character of each set is a "r", the file is readable for those
users, if a "-", it is not readable.

If the second character of each set is a "w", the file is writable for those
users, if a "-", it is not writable.

The third character is more complex, and is the first of the following
characters which apply.
\itemize{
\item 'S' If the character is part of the owner permissions and the file is not
executable or the directory is not searchable by the owner, and the
set-user-id bit is set.
\item 'S' If the character is part of the group permissions and the file is not
executable or the directory is not searchable by the group, and the
set-group-id bit is set.
\item 'T' If the character is part of the other permissions and the file is not
executable or the directory is not searchable by others, and the 'sticky'
(S_ISVTX) bit is set.
\item 's' If the character is part of the owner permissions and the file is
executable or the directory searchable by the owner, and the set-user-id bit
is set.
\item 's' If the character is part of the group permissions and the file is
executable or the directory searchable by the group, and the set-group-id
bit is set.
\item 't' If the character is part of the other permissions and the file is
executable or the directory searchable by others, and the ''sticky''
(S_ISVTX) bit is set.
\item 'x' The file is executable or the directory is searchable.
\item '-' If none of the above apply.
Most commonly the third character is either 'x' or \code{-}.
}

On Windows the permissions are displayed as a 3 character string where the
third character is only \code{-} or \code{x}.
}
\examples{
# Integer and numeric
fs_perms(420L)
fs_perms(c(511, 420))

# Octal
fs_perms("777")
fs_perms(c("777", "644"))

# Symbolic
fs_perms("a+rwx")
fs_perms(c("a+rwx", "u+rw,go+r"))

# Use the `&` and `|`operators to check for certain permissions
(fs_perms("777") & "u+r") == "u+r"
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create.R
\name{create}
\alias{create}
\alias{file_create}
\alias{dir_create}
\alias{link_create}
\title{Create files, directories, or links}
\usage{
file_create(path, ..., mode = "u=rw,go=r")

dir_create(path, ..., mode = "u=rwx,go=rx", recurse = TRUE, recursive)

link_create(path, new_path, symbolic = TRUE)
}
\arguments{
\item{path}{A character vector of one or more paths. For \code{link_create()},
this is the target.}

\item{...}{Additional arguments passed to \code{\link[=path]{path()}}}

\item{mode}{If file/directory is created, what mode should it have?

Links do not have mode; they inherit the mode of the file they link to.}

\item{recurse}{should intermediate directories be created if they do not
exist?}

\item{recursive}{(Deprecated) If \code{TRUE} recurse fully.}

\item{new_path}{The path where the link should be created.}

\item{symbolic}{Boolean value determining if the link should be a symbolic
(the default) or hard link.}
}
\value{
The path to the created object (invisibly).
}
\description{
The functions \code{file_create()} and \code{dir_create()} ensure that \code{path} exists;
if it already exists it will be left unchanged. That means that compared to
\code{\link[=file.create]{file.create()}}, \code{file_create()} will not truncate an existing file, and
compared to \code{\link[=dir.create]{dir.create()}}, \code{dir_create()} will silently ignore existing
directories.
}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
file_create("foo")
is_file("foo")
# dir_create applied to the same path will fail
try(dir_create("foo"))

dir_create("bar")
is_dir("bar")
# file_create applied to the same path will fail
try(file_create("bar"))

# Cleanup
file_delete("foo")
dir_delete("bar")
\dontshow{setwd(.old_wd)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path.R
\name{path_filter}
\alias{path_filter}
\title{Filter paths}
\usage{
path_filter(path, glob = NULL, regexp = NULL, invert = FALSE, ...)
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{glob}{A wildcard aka globbing pattern (e.g. \verb{*.csv}) passed on to \code{\link[=grep]{grep()}} to filter paths.}

\item{regexp}{A regular expression (e.g. \verb{[.]csv$}) passed on to \code{\link[=grep]{grep()}} to filter paths.}

\item{invert}{If \code{TRUE} return files which do \emph{not} match}

\item{...}{Additional arguments passed to \link{grep}.}
}
\description{
Filter paths
}
\examples{
path_filter(c("foo", "boo", "bar"), glob = "*oo")
path_filter(c("foo", "boo", "bar"), glob = "*oo", invert = TRUE)

path_filter(c("foo", "boo", "bar"), regexp = "b.r")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path.R
\name{path_tidy}
\alias{path_tidy}
\title{Tidy paths}
\usage{
path_tidy(path)
}
\arguments{
\item{path}{A character vector of one or more paths.}
}
\value{
An \code{fs_path} object, which is a character vector that also has class
\code{fs_path}
}
\description{
untidy paths are all different, tidy paths are all the same.
Tidy paths always use \code{/} to delimit directories, never have
multiple \code{/} or trailing \code{/} and have colourised output based on the file
type.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/file.R
\name{file_touch}
\alias{file_touch}
\title{Change file access and modification times}
\usage{
file_touch(path, access_time = Sys.time(), modification_time = access_time)
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{access_time, modification_time}{The times to set, inputs will be
coerced to \link{POSIXct} objects.}
}
\description{
Unlike the touch POSIX utility this does not create the file if it does not
exist. Use \code{\link[=file_create]{file_create()}} to do this if needed.
}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
file_create("foo")
file_touch("foo", "2018-01-01")
file_info("foo")[c("access_time", "modification_time", "change_time", "birth_time")]
\dontshow{setwd(.old_wd)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delete.R
\name{delete}
\alias{delete}
\alias{file_delete}
\alias{dir_delete}
\alias{link_delete}
\title{Delete files, directories, or links}
\usage{
file_delete(path)

dir_delete(path)

link_delete(path)
}
\arguments{
\item{path}{A character vector of one or more paths.}
}
\value{
The deleted paths (invisibly).
}
\description{
\code{file_delete()} and \code{link_delete()} delete file and links. Compared to
\link{file.remove} they always fail if they cannot delete the object rather than
changing return value or signalling a warning. If any inputs are
directories, they are passed to \code{dir_delete()}, so \code{file_delete()} can
therefore be used to delete any filesystem object.

\code{dir_delete()} will first delete the contents of the directory, then remove
the directory. Compared to \link{unlink} it will always throw an error if the
directory cannot be deleted rather than being silent or signalling a warning.
}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
# create a directory, with some files and a link to it
dir_create("dir")
files <- file_create(path("dir", letters[1:5]))
link <- link_create(path_abs("dir"), "link")

# All files created
dir_exists("dir")
file_exists(files)
link_exists("link")
file_exists(link_path("link"))

# Delete a file
file_delete(files[1])
file_exists(files[1])

# Delete the directory (which deletes the files as well)
dir_delete("dir")
file_exists(files)
dir_exists("dir")

# The link still exists, but what it points to does not.
link_exists("link")
dir_exists(link_path("link"))

# Delete the link
link_delete("link")
link_exists("link")
\dontshow{setwd(.old_wd)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list.R
\name{dir_ls}
\alias{dir_ls}
\alias{dir_map}
\alias{dir_walk}
\alias{dir_info}
\title{List files}
\usage{
dir_ls(
  path = ".",
  all = FALSE,
  recurse = FALSE,
  type = "any",
  glob = NULL,
  regexp = NULL,
  invert = FALSE,
  fail = TRUE,
  ...,
  recursive
)

dir_map(
  path = ".",
  fun,
  all = FALSE,
  recurse = FALSE,
  type = "any",
  fail = TRUE
)

dir_walk(
  path = ".",
  fun,
  all = FALSE,
  recurse = FALSE,
  type = "any",
  fail = TRUE
)

dir_info(
  path = ".",
  all = FALSE,
  recurse = FALSE,
  type = "any",
  regexp = NULL,
  glob = NULL,
  fail = TRUE,
  ...
)
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{all}{If \code{TRUE} hidden files are also returned.}

\item{recurse}{If \code{TRUE} recurse fully, if a positive number the number of levels
to recurse.}

\item{type}{File type(s) to return, one or more of "any", "file", "directory",
"symlink", "FIFO", "socket", "character_device" or "block_device".}

\item{glob}{A wildcard aka globbing pattern (e.g. \verb{*.csv}) passed on to \code{\link[=grep]{grep()}} to filter paths.}

\item{regexp}{A regular expression (e.g. \verb{[.]csv$}) passed on to \code{\link[=grep]{grep()}} to filter paths.}

\item{invert}{If \code{TRUE} return files which do \emph{not} match}

\item{fail}{Should the call fail (the default) or warn if a file cannot be
accessed.}

\item{...}{Additional arguments passed to \link{grep}.}

\item{recursive}{(Deprecated) If \code{TRUE} recurse fully.}

\item{fun}{A function, taking one parameter, the current path entry.}
}
\description{
\code{dir_ls()} is equivalent to the \code{ls} command. It returns filenames as a
named \code{fs_path} character vector. The names are equivalent to the values,
which is useful for passing onto functions like \code{purrr::map_dfr()}.

\code{dir_info()} is equivalent to \code{ls -l} and a shortcut for
\code{file_info(dir_ls())}.

\code{dir_map()} applies a function \code{fun()} to each entry in the path and returns
the result in a list.

\code{dir_walk()} calls \code{fun} for its side-effect and returns the input \code{path}.
}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
dir_ls(R.home("share"), type = "directory")

# Create a shorter link
link_create(system.file(package = "base"), "base")

dir_ls("base", recurse = TRUE, glob = "*.R")

# If you need the full paths input an absolute path
dir_ls(path_abs("base"))

dir_map("base", identity)

dir_walk("base", str)

dir_info("base")

# Cleanup
link_delete("base")
\dontshow{setwd(.old_wd)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/access.R
\name{file_access}
\alias{file_access}
\alias{file_exists}
\alias{dir_exists}
\alias{link_exists}
\title{Query for existence and access permissions}
\usage{
file_access(path, mode = "exists")

file_exists(path)

dir_exists(path)

link_exists(path)
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{mode}{A character vector containing one or more of 'exists', 'read',
'write', 'execute'.}
}
\value{
A logical vector, with names corresponding to the input \code{path}.
}
\description{
\code{file_exists(path)} is a shortcut for \code{file_access(x, "exists")};
\code{dir_exists(path)} and \code{link_exists(path)} are similar but also check that
the path is a directory or link, respectively. (\code{file_exists(path)} returns
\code{TRUE} if \code{path} exists and it is a directory.)
}
\details{
\strong{Cross-compatibility warning:} There is no executable bit on
Windows. Checking a file for mode 'execute' on Windows, e.g.
\code{file_access(x, "execute")} will always return \code{TRUE}.
}
\examples{
file_access("/")
file_access("/", "read")
file_access("/", "write")

file_exists("WOMBATS")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path.R
\name{path}
\alias{path}
\alias{path_wd}
\title{Construct path to a file or directory}
\usage{
path(..., ext = "")

path_wd(..., ext = "")
}
\arguments{
\item{...}{character vectors, if any values are NA, the result will also be
NA. The paths follow the recycling rules used in the tibble package,
namely that only length 1 arguments are recycled.}

\item{ext}{An optional extension to append to the generated path.}
}
\description{
\code{path()} constructs a relative path, \code{path_wd()} constructs an absolute path
from the current working directory.
}
\examples{
path("foo", "bar", "baz", ext = "zip")

path("foo", letters[1:3], ext = "txt")
}
\seealso{
\code{\link[=path_home]{path_home()}}, \code{\link[=path_package]{path_package()}} for functions to construct paths
relative to the home and package directories respectively.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/copy.R
\name{copy}
\alias{copy}
\alias{file_copy}
\alias{dir_copy}
\alias{link_copy}
\title{Copy files, directories or links}
\usage{
file_copy(path, new_path, overwrite = FALSE)

dir_copy(path, new_path, overwrite = FALSE)

link_copy(path, new_path, overwrite = FALSE)
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{new_path}{A character vector of paths to the new locations.}

\item{overwrite}{Overwrite files if they exist. If this is \code{FALSE} and the
file exists an error will be thrown.}
}
\value{
The new path (invisibly).
}
\description{
\code{file_copy()} copies files.

\code{link_copy()} creates a new link pointing to the same location as the previous link.

\code{dir_copy()} copies the directory recursively at the new location.
}
\details{
The behavior of \code{dir_copy()} differs slightly than that of \code{file.copy()} when
\code{overwrite = TRUE}. The directory will always be copied to \code{new_path}, even
if the name differs from the basename of \code{path}.
}
\examples{
\dontshow{.old_wd <- setwd(tempdir())}
file_create("foo")
file_copy("foo", "bar")
try(file_copy("foo", "bar"))
file_copy("foo", "bar", overwrite = TRUE)
file_delete(c("foo", "bar"))

dir_create("foo")
# Create a directory and put a few files in it
files <- file_create(c("foo/bar", "foo/baz"))
file_exists(files)

# Copy the directory
dir_copy("foo", "foo2")
file_exists(path("foo2", path_file(files)))

# Create a link to the directory
link_create(path_abs("foo"), "loo")
link_path("loo")
link_copy("loo", "loo2")
link_path("loo2")

# Cleanup
dir_delete(c("foo", "foo2"))
link_delete(c("loo", "loo2"))
\dontshow{setwd(.old_wd)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fs-package.R
\docType{package}
\name{fs-package}
\alias{fs}
\alias{fs-package}
\title{fs: Cross-Platform File System Operations Based on 'libuv'}
\description{
A cross-platform interface to file system operations, built on top of the 'libuv' C library.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://fs.r-lib.org}
  \item \url{https://github.com/r-lib/fs}
  \item Report bugs at \url{https://github.com/r-lib/fs/issues}
}

}
\author{
\strong{Maintainer}: Gábor Csárdi \email{csardi.gabor@gmail.com}

Authors:
\itemize{
  \item Jim Hester
  \item Hadley Wickham \email{hadley@rstudio.com}
}

Other contributors:
\itemize{
  \item libuv project contributors (libuv library) [copyright holder]
  \item Joyent, Inc. and other Node contributors (libuv library) [copyright holder]
  \item RStudio [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path.R
\name{path_file}
\alias{path_file}
\alias{path_dir}
\alias{path_ext}
\alias{path_ext_remove}
\alias{path_ext_set}
\alias{path_ext<-}
\title{Manipulate file paths}
\usage{
path_file(path)

path_dir(path)

path_ext(path)

path_ext_remove(path)

path_ext_set(path, ext)

path_ext(path) <- value
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{ext, value}{The new file extension.}
}
\description{
\code{path_file()} returns the filename portion of the path, \code{path_dir()} returns
the directory portion. \code{path_ext()} returns the last extension (if any) for a
path. \code{path_ext_remove()} removes the last extension and returns the rest of
the path. \code{path_ext_set()} replaces the extension with a new extension. If
there is no existing extension the new extension is appended.
}
\details{
Note because these are not full file paths they return regular character
vectors, not \code{fs_path()} objects.
}
\examples{
path_file("dir/file.zip")

path_dir("dir/file.zip")

path_ext("dir/file.zip")

path_ext("file.tar.gz")

path_ext_remove("file.tar.gz")

# Only one level of extension is removed
path_ext_set(path_ext_remove("file.tar.gz"), "zip")
}
\seealso{
\code{\link[base:basename]{base::basename()}}, \code{\link[base:basename]{base::dirname()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fs-package.R
\docType{package}
\name{fs-package}
\alias{fs}
\alias{fs-package}
\title{fs: Cross-Platform File System Operations Based on 'libuv'}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A cross-platform interface to file system operations, built on top of the 'libuv' C library.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://fs.r-lib.org}
  \item \url{https://github.com/r-lib/fs}
  \item Report bugs at \url{https://github.com/r-lib/fs/issues}
}

}
\author{
\strong{Maintainer}: Jeroen Ooms \email{jeroenooms@gmail.com}

Authors:
\itemize{
  \item Jim Hester
  \item Hadley Wickham
  \item Gábor Csárdi
}

Other contributors:
\itemize{
  \item libuv project contributors (libuv library) [copyright holder]
  \item Joyent, Inc. and other Node contributors (libuv library) [copyright holder]
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{dir_tree}
\alias{dir_tree}
\title{Print contents of directories in a tree-like format}
\usage{
dir_tree(path = ".", recurse = TRUE, ...)
}
\arguments{
\item{path}{A path to print the tree from}

\item{recurse}{If \code{TRUE} recurse fully, if a positive number the number of levels
to recurse.}

\item{...}{Additional arguments passed to \link{dir_ls}.}
}
\description{
Print contents of directories in a tree-like format
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{dir_tree}
\alias{dir_tree}
\title{Print contents of directories in a tree-like format}
\usage{
dir_tree(path = ".", recurse = TRUE, ...)
}
\arguments{
\item{path}{A path to print the tree from}

\item{recurse}{If \code{TRUE} recurse fully, if a positive number the number of levels
to recurse.}

\item{...}{
  Arguments passed on to \code{\link[=dir_ls]{dir_ls}}
  \describe{
    \item{\code{type}}{File type(s) to return, one or more of "any", "file", "directory",
"symlink", "FIFO", "socket", "character_device" or "block_device".}
    \item{\code{recursive}}{(Deprecated) If \code{TRUE} recurse fully.}
    \item{\code{all}}{If \code{TRUE} hidden files are also returned.}
    \item{\code{fail}}{Should the call fail (the default) or warn if a file cannot be
accessed.}
    \item{\code{glob}}{A wildcard aka globbing pattern (e.g. \verb{*.csv}) passed on to \code{\link[=grep]{grep()}} to filter paths.}
    \item{\code{regexp}}{A regular expression (e.g. \verb{[.]csv$}) passed on to \code{\link[=grep]{grep()}} to filter paths.}
    \item{\code{invert}}{If \code{TRUE} return files which do \emph{not} match}
  }}
}
\description{
Print contents of directories in a tree-like format
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/temp.R
\name{file_temp}
\alias{file_temp}
\alias{file_temp_push}
\alias{file_temp_pop}
\alias{path_temp}
\title{Create names for temporary files}
\usage{
file_temp(pattern = "file", tmp_dir = tempdir(), ext = "")

file_temp_push(path)

file_temp_pop()

path_temp(...)
}
\arguments{
\item{pattern}{A character vector with the non-random portion of the name.}

\item{tmp_dir}{The directory the file will be created in.}

\item{ext}{The file extension of the temporary file.}

\item{path}{A character vector of one or more paths.}

\item{...}{Additional paths appended to the temporary directory by \code{path()}.}
}
\description{
\code{file_temp()} returns the name which can be used as a temporary file.
}
\details{
\code{file_temp_push()} can be used to supply deterministic entries in the
temporary file stack. This can be useful for reproducibility in like example
documentation and vignettes.

\code{file_temp_pop()} can be used to explicitly remove an entry from the
internal stack, however generally this is done instead by calling
\code{file_temp()}.

\code{path_temp()} constructs a path within the session temporary directory.
}
\examples{
\dontshow{file_temp_push("/tmp/filedd461c46df20")}

path_temp()
path_temp("does-not-exist")

file_temp()
file_temp(ext = "png")
file_temp("image", ext = "png")


# You can make the temp file paths deterministic
file_temp_push(letters)
file_temp()
file_temp()

# Or explicitly remove values
while (!is.null(file_temp_pop())) next
file_temp_pop()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/file.R
\name{file_show}
\alias{file_show}
\title{Open files or directories}
\usage{
file_show(path = ".", browser = getOption("browser"))
}
\arguments{
\item{path}{A character vector of one or more paths.}

\item{browser}{a non-empty character string giving the name of the
    program to be used as the HTML browser.  It should be in the PATH,
    or a full path specified.  Alternatively, an \R function to be
    called to invoke the browser.

    Under Windows \code{NULL} is also allowed (and is the default), and
    implies that the file association mechanism will be used.
  }
}
\value{
The directories that were opened (invisibly).
}
\description{
Open files or directories
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.R
\name{is_dir_empty}
\alias{is_dir_empty}
\title{Check if a directory is empty}
\usage{
is_dir_empty(path)
}
\arguments{
\item{path}{A character string specifying the path to the directory to check.}
}
\value{
A logical value. Returns \code{TRUE} if the directory is empty,
\code{FALSE} otherwise.
}
\description{
This function checks whether a given directory is empty or not.
}
\examples{
\dontrun{
is_dir_empty("path/to/empty/directory")  # Returns TRUE
is_dir_empty("path/to/non-empty/directory")  # Returns FALSE
}

}
